#ifndef __RLL_ACTION__
#define __RLL_ACTION__

#include <deque>
#include <nlp_string.h>
#include "UIObject.h"
using namespace std;

typedef deque <bool>						bool_dq_t;


// ----------------------------------------------
class Command
{
	private:
		static String_dq_t		dq_Commands;
		static bool_dq_t		dq_IsTyping;
		static bool_dq_t		dq_NeedsObject;
		static bool_dq_t		dq_NeedsParam;

	public:
		static int i_Count;

		static void Load (void);
		static void Clear (void);

		static String Get (int _iIndex)
			{ return dq_Commands [_iIndex]; };
		static bool IsTypingCommand (int _iIndex)
			{ return dq_IsTyping [_iIndex]; };
		static bool NeedsObject (int _iIndex)
			{ return dq_NeedsObject [_iIndex]; };
		static bool NeedsParam (int _iIndex)
			{ return dq_NeedsParam [_iIndex]; };
};



// ----------------------------------------------
class EnvironmentAction
{
	public:
		int				i_SelectedObjectIndex;
		UIObject*		p_SelectedObject;
		char			i_SelectedCommandIndex;
		String			s_SelectedParameter;
};



// ----------------------------------------------
class TextAction
{
	public:
		char			i_SelectedObjectWordLocation;
		char			i_SelectedCommandWordLocation;
		char			i_SelectedParameterWordLocation;
};



// ----------------------------------------------
struct ActionInfo
{
	int				i_SelectedObjectIndex;
	char			i_SelectedCommandIndex;
	char			i_SelectedObjectWordLocation;
	char			i_SelectedCommandWordLocation;
	char			i_SelectedParameterWordLocation;
};


union ActionInfo_u
{
	ActionInfo	o_ActionInfo;
	long		l_ActionInfo;
};


// ----------------------------------------------
class RLAction : public EnvironmentAction, public TextAction
{
	public:
		RLAction (void);

		long GetAsLong (void)
		{
			assert (sizeof (long) >= sizeof (ActionInfo));

			ActionInfo_u uAction;
			uAction.o_ActionInfo.i_SelectedObjectIndex = i_SelectedObjectIndex;
			uAction.o_ActionInfo.i_SelectedCommandIndex = i_SelectedCommandIndex;
			uAction.o_ActionInfo.i_SelectedObjectWordLocation 
				= i_SelectedObjectWordLocation;
			uAction.o_ActionInfo.i_SelectedCommandWordLocation
				= i_SelectedCommandWordLocation;
			uAction.o_ActionInfo.i_SelectedParameterWordLocation
				= i_SelectedParameterWordLocation;

			assert (*this == LongToAction (uAction.l_ActionInfo));
			return uAction.l_ActionInfo;
		};

		static long ActionToLong (char _iWU, char _iWC, char _iWP, char _iC, int _iU)
		{
			ActionInfo_u uAction;
			uAction.o_ActionInfo.i_SelectedObjectIndex = _iU;
			uAction.o_ActionInfo.i_SelectedCommandIndex = _iC;
			uAction.o_ActionInfo.i_SelectedObjectWordLocation = _iWU;
			uAction.o_ActionInfo.i_SelectedCommandWordLocation = _iWC;
			uAction.o_ActionInfo.i_SelectedParameterWordLocation = _iWP;
			return uAction.l_ActionInfo;
		};

		static RLAction LongToAction (long _lAction)
		{
			ActionInfo_u uAction;
			uAction.l_ActionInfo = _lAction;

			RLAction oAction;
			oAction.i_SelectedObjectIndex 
				= uAction.o_ActionInfo.i_SelectedObjectIndex;
			oAction.i_SelectedCommandIndex 
				= uAction.o_ActionInfo.i_SelectedCommandIndex;
			oAction.i_SelectedObjectWordLocation 
				= uAction.o_ActionInfo.i_SelectedObjectWordLocation;
			oAction.i_SelectedCommandWordLocation 
				= uAction.o_ActionInfo.i_SelectedCommandWordLocation;
			oAction.i_SelectedParameterWordLocation 
				= uAction.o_ActionInfo.i_SelectedParameterWordLocation;

			return oAction;
		};

		bool operator== (const RLAction& _rOther);
};

std::ostream& operator<< (std::ostream& _rStream, const RLAction& _rAction);



#endif
