#include "Annotation.h"
#include <nlp_macros.h>
#include <nlp_config.h>
#include <nlp_filesystem.h>
#include <assert.h>


IdToAnnotation_hmp_t	Annotation::hmp_IdToAnnotation;


//														
AnnotatedAction::AnnotatedAction (void)
{
	i_CommandWord = -1;
	i_ObjectWord = -1;
	i_Parameter = -1;
	b_Optional = false;
}


//														
AnnotatedSentence::~AnnotatedSentence (void)
{
	ITERATE (AnnotatedAction_dq_t, dq_Action, ite)
		delete *ite;
	dq_Action.clear ();
}


//														
Annotation::Annotation (String& _rID, char _cType)
{
	s_ID = _rID;
	c_Type = _cType;
	hmp_IdToAnnotation.insert (make_pair (s_ID, this));
}


//														
Annotation::~Annotation (void)
{
	assert (ANNOTATION_DELETED != c_Type);

	ITERATE (AnnotatedSentence_dq_t, dq_Sentence, ite)
		delete *ite;
	dq_Sentence.clear ();

	c_Type = ANNOTATION_DELETED;
}


//														
Annotation* Annotation::Find (String& _rID)
{
	IdToAnnotation_hmp_t::iterator	ite;
	ite = hmp_IdToAnnotation.find (_rID);
	if (hmp_IdToAnnotation.end () == ite)
		return NULL;

	return ite->second;
}


//														
bool Annotation::Load (void)
{
	if (1 == (int)(config)"use_full_annotations")
	{
		cout << "Loading full annotations ..." << endl;
		return LoadAnnotation (ANNOTATION_FULL);
	}
	else if (1 == (int)(config)"use_partial_annotations")
	{
		cout << "Loading partial annotations." << endl;
		return LoadAnnotation (ANNOTATION_PARTIAL);
	}
	return true;
}


//														
bool Annotation::LoadAnnotation (char _cType)
{
	cout << "Loading annotations" << endl;
	bool bFilterById = (1 == (int)(config)"filter_annotation_by_document_id");

	// load the ids of the documents for which full	
	// annotations should be loaded.				
	String_set_t	setAnnotatedDocs;
	if (true == bFilterById)
	{
		File file;
		if (false == file.Open ((config)"fully_annotated_document_id_file"))
			return false;

		String sLine;
		while (true == file.ReadLine (sLine))
		{
			sLine.Strip ();
			setAnnotatedDocs.insert (sLine);
		}
	}


	// load the full annotations ...				
	{
		File file;
		if (false == file.Open ((config)"full_annotation_file"))
			return false;

		bool bSkipDoc = false;
		Annotation* pAnnotation = NULL;
		AnnotatedSentence* pSentence = NULL;
		AnnotatedAction* pAction = NULL;

		String sLine;
		while (true == file.ReadLine (sLine))
		{
			String sCheck (sLine);
			sCheck.Replace ("-", "");
			sCheck.Strip ();
			if ("" == sCheck)
				continue;

			// document id ...						
			if (true == sLine.StartsWith ("    article ["))
			{
				size_t iStart = sLine.Find ("[") + 1;
				String sID (sLine.substr (iStart, sLine.Find ("]") - iStart));
				bSkipDoc = ((true == bFilterById) && 
							(setAnnotatedDocs.end () == setAnnotatedDocs.find (sID)));
				if (true == bSkipDoc)
					continue;

				pAnnotation = new Annotation (sID, _cType);
				hmp_IdToAnnotation.insert (make_pair (sID, pAnnotation));	
				pSentence = NULL;
				pAction = NULL;
				continue;
			}

			if (true == bSkipDoc)
				continue;


			// command words ...					
			if (true == sLine.StartsWith ("         c:"))
			{
				// only ANNOTATION_FULL uses the word annotations 
				if (ANNOTATION_PARTIAL == _cType)
					continue;

				String_dq_t dqWords;
				sLine.Split (dqWords);

				pAction = new AnnotatedAction;
				String_dq_t dqValues;

				dqWords [0].Split (dqValues, ':');
				pAction->i_CommandWord = dqValues [1];

				dqWords [1].Split (dqValues, ':');
				pAction->i_ObjectWord = dqValues [1];

				dqWords [2].Split (dqValues, ':');
				pAction->i_Parameter = dqValues [1];
				continue;
			}

			// environment command ...				
			if ((true == sLine.StartsWith ("         - ")) ||
				(true == sLine.StartsWith ("         ~ ")))
			{
				// this is the first line of an action annotation 
				// in the partial annotation case.  In full 	  
				// annotation, we would already have encountered  
				// a line with the word annotations, so pAction   
				// would already have been created.               
				if (ANNOTATION_PARTIAL == _cType)
					pAction = new AnnotatedAction;
				else
					assert (NULL != pAction);


				pAction->b_Optional = (true == sLine.StartsWith ("         ~ "));

				sLine.Replace ("         - ", "");
				sLine.Replace ("         ~ " , "");
				sLine.Strip ();
				sLine.ReplacePattern ("[ ]+", " ");

				size_t iStart = 0;
				size_t iEnd = sLine.Find ("(");
				pAction->s_Command = sLine.substr (0, iEnd - iStart - 1);

				iStart = sLine.Find ("[", iEnd) + 1;
				iEnd = sLine.Find ("]", iStart);
				pAction->s_ObjectType = sLine.substr (iStart, iEnd - iStart);
				pAction->s_ObjectType.LowerCase ();

				iStart = iEnd + 2;
				iEnd = sLine.Find ("\"", iStart);
				pAction->s_ObjectName = sLine.substr (iStart, iEnd - iStart - 1);
				continue;
			}


			if (true == sLine.StartsWith ("         uid:"))
			{
				if (NULL == pAction)
					continue;
				sLine.Replace ("         uid:", "");
				sLine.Strip ();
				pAction->s_TemporallyUniqueObjectId = sLine.substr (sLine.Find ("\x04"));
				pSentence->dq_Action.push_back (pAction);
				pAction = NULL;
				continue;
			}

			// sentence text ...					
			if (true == sLine.StartsWith ("      "))
			{
				sLine.Strip ();

				pSentence = new AnnotatedSentence;
				pSentence->s_Text = sLine;
				pAnnotation->dq_Sentence.push_back (pSentence);
				continue;
			}
		}
	}

	cout << "   - annotated docs : " << hmp_IdToAnnotation.size () << endl;
	return true;
}


