#ifndef __BIT_FEATURE_MATRIX__
#define __BIT_FEATURE_MATRIX__
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <hash_map>
#include "BitVector.h"
using namespace std;


typedef unsigned long					FeatureKey_t;
typedef hash_map <FeatureKey_t, void*>	KeyToData_hmp_t;

//													
template <size_t FEATURES, int TMPL_DIM>
class BitFeatureMatrix
{

	friend class BitFeatureMatrix <FEATURES, 5>;
	friend class BitFeatureMatrix <FEATURES, 4>;
	friend class BitFeatureMatrix <FEATURES, 3>;
	friend class BitFeatureMatrix <FEATURES, 2>;
	friend class BitFeatureMatrix <FEATURES, 1>;


	private:
		KeyToData_hmp_t		hmp_KeyToData;
		short int			i_BitOffsets [TMPL_DIM];
	
		BitVector<FEATURES>* GetData (FeatureKey_t _oKey)
		{
			BitVector<FEATURES>* pData = NULL;

			KeyToData_hmp_t::iterator	ite;
			ite = hmp_KeyToData.find (_oKey);
			if (hmp_KeyToData.end () == ite)
			{
				pData = new BitVector<FEATURES>;
				hmp_KeyToData.insert (make_pair (_oKey, pData));
			}
			else
				pData = (BitVector<FEATURES>*) ite->second;

			return pData;
		};


	public:
		BitFeatureMatrix (void);
		BitFeatureMatrix (int _i0);
		BitFeatureMatrix (int _i0, int _i1);
		BitFeatureMatrix (int _i0, int _i1, int _i2);
		BitFeatureMatrix (int _i0, int _i1, int _i2, int _i3);

		~BitFeatureMatrix (void);
		void Destroy (void);

		unsigned short int MinRequiredBits (int _iNumber)
			{ return (unsigned short int) floor (1 + log(_iNumber) / (double) log(2)); };

		void Create (int _i0);
		void Create (int _i0, int _i1);
		void Create (int _i0, int _i1, int _i2);
		void Create (int _i0, int _i1, int _i2, int _i3);


		void GetSlice (BitVector <FEATURES>** _ppBitFeatureMatrix, int _i0);
		void GetSlice (BitVector <FEATURES>** _ppBitFeatureMatrix, int _i0, int _i1);
		void GetSlice (BitVector <FEATURES>** _ppBitFeatureMatrix, int _i0, int _i1, int _i2);
		void GetSlice (BitVector <FEATURES>** _ppBitFeatureMatrix, int _i0, int _i1, int _i2, int _i3);

		void CopySlice (BitVector <FEATURES>** _ppBitFeatureMatrix, int _i0);
		void CopySlice (BitVector <FEATURES>** _ppBitFeatureMatrix, int _i0, int _i1);
		void CopySlice (BitVector <FEATURES>** _ppBitFeatureMatrix, int _i0, int _i1, int _i2);
		void CopySlice (BitVector <FEATURES>** _ppBitFeatureMatrix, int _i0, int _i1, int _i2, int _i3);
};


//													
template <size_t FEATURES, int TMPL_DIM>
BitFeatureMatrix<FEATURES, TMPL_DIM>::BitFeatureMatrix (void)
{
	memset (i_BitOffsets, 0, sizeof (short int) * TMPL_DIM);
}


//													
template <size_t FEATURES, int TMPL_DIM>
BitFeatureMatrix<FEATURES, TMPL_DIM>::~BitFeatureMatrix (void)
{
	ITERATE (KeyToData_hmp_t, hmp_KeyToData, ite)
		delete (BitVector<FEATURES>*) ite->second;

	hmp_KeyToData.clear ();
}



//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::Destroy (void)
{
	ITERATE (KeyToData_hmp_t, hmp_KeyToData, ite)
		delete (BitVector<FEATURES>*) ite->second;

	hmp_KeyToData.clear ();
}


//													
template <size_t FEATURES, int TMPL_DIM>
BitFeatureMatrix<FEATURES, TMPL_DIM>::BitFeatureMatrix (int _i0)
{
	Create (_i0);
}


//													
template <size_t FEATURES, int TMPL_DIM>
BitFeatureMatrix<FEATURES, TMPL_DIM>::BitFeatureMatrix (int _i0, int _i1)
{
	Create (_i0, _i1);
}


//													
template <size_t FEATURES, int TMPL_DIM>
BitFeatureMatrix<FEATURES, TMPL_DIM>::BitFeatureMatrix (int _i0, int _i1, int _i2)
{
	Create (_i0, _i1, _i2);
}


//													
template <size_t FEATURES, int TMPL_DIM>
BitFeatureMatrix<FEATURES, TMPL_DIM>::BitFeatureMatrix (int _i0, int _i1, int _i2, int _i3)
{
	Create (_i0, _i1, _i2, _i3);
}






//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::Create (int _i0)
{
	assert (1 == TMPL_DIM);
	assert (_i0 <= 8 * sizeof (FeatureKey_t));
	assert (true == hmp_KeyToData.empty ());

	i_BitOffsets [0] = -1;
}


//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::Create (int _i0, int _i1)
{
	assert (2 == TMPL_DIM);
	assert (8 * sizeof (FeatureKey_t) >= (size_t)(MinRequiredBits (_i0) 
												+ MinRequiredBits (_i1)));
	assert (true == hmp_KeyToData.empty ());

	i_BitOffsets [0] = MinRequiredBits (_i1);
	i_BitOffsets [1] = -1;
}


//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::Create (int _i0, int _i1, int _i2)
{
	assert (3 == TMPL_DIM);
	assert (true == hmp_KeyToData.empty ());
	assert (8 * sizeof (FeatureKey_t) >= (size_t)(MinRequiredBits (_i0) 
												+ MinRequiredBits (_i1)
												+ MinRequiredBits (_i2)));

	i_BitOffsets [0] = MinRequiredBits (_i1) + MinRequiredBits (_i2);
	i_BitOffsets [1] = MinRequiredBits (_i2);
	i_BitOffsets [2] = -1;
}


//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::Create (int _i0, int _i1, int _i2, int _i3)
{
	assert (4 == TMPL_DIM);
	assert (8 * sizeof (FeatureKey_t) >= (size_t)(MinRequiredBits (_i0) 
												+ MinRequiredBits (_i1)
												+ MinRequiredBits (_i2)
												+ MinRequiredBits (_i3)));
	assert (true == hmp_KeyToData.empty ());

	i_BitOffsets [0] = MinRequiredBits (_i1) + MinRequiredBits (_i2) + MinRequiredBits (_i3);
	i_BitOffsets [1] = MinRequiredBits (_i2) + MinRequiredBits (_i3);
	i_BitOffsets [2] = MinRequiredBits (_i3);
	i_BitOffsets [3] = -1;
}






//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::GetSlice (BitVector <FEATURES>** _ppVector,
													 int _i0)
{
	assert (1 == TMPL_DIM);
	FeatureKey_t oKey = _i0;

	*_ppVector = GetData (oKey);
}


//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::GetSlice (BitVector <FEATURES>** _ppVector,
													 int _i0, int _i1)
{
	assert (2 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						_i1;

	*_ppVector = GetData (oKey);
}


//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::GetSlice (BitVector <FEATURES>** _ppVector,
													 int _i0, int _i1, int _i2)
{
	assert (3 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						(_i1 << i_BitOffsets [1]) |
						_i2;

	*_ppVector = GetData (oKey);
}


//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::GetSlice (BitVector <FEATURES>** _ppVector,
													 int _i0, int _i1, int _i2, int _i3)
{
	assert (4 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						(_i1 << i_BitOffsets [1]) |
						(_i2 << i_BitOffsets [2]) |
						_i3;

	*_ppVector = GetData (oKey);
}






//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::CopySlice (BitVector <FEATURES>** _ppVector,
													 int _i0)
{
	assert (1 == TMPL_DIM);
	FeatureKey_t oKey = _i0;

	*_ppVector = GetData (oKey);
}


//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::CopySlice (BitVector <FEATURES>** _ppVector,
													 int _i0, int _i1)
{
	assert (2 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						_i1;

	*_ppVector = GetData (oKey);
}


//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::CopySlice (BitVector <FEATURES>** _ppVector, 
													 int _i0, int _i1, int _i2)
{
	assert (3 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						(_i1 << i_BitOffsets [1]) |
						_i2;

	*_ppVector = GetData (oKey);
}


//													
template <size_t FEATURES, int TMPL_DIM>
void BitFeatureMatrix<FEATURES, TMPL_DIM>::CopySlice (BitVector <FEATURES>** _ppVector,
													 int _i0, int _i1, int _i2, int _i3)
{
	assert (4 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						(_i1 << i_BitOffsets [1]) |
						(_i2 << i_BitOffsets [2]) |
						_i3;

	*_ppVector = GetData (oKey);
}








#endif

