#ifndef __BIT_VECTOR__
#define __BIT_VECTOR__
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <nlp_string.h>
using namespace std;


// -----------------------------------------------------
template <size_t FEATURES>
class BitVector
{
	protected:
		char	c_Features [1 + FEATURES / (8 * sizeof(char))];

	public:
		BitVector (void);

		void Clear (void);
		void Memset (int _iValue);
		void SetFeature (int _iIndex, int _iValue);
		void CopyVector (const BitVector<FEATURES>& _rSource);
		const int operator[] (int _iIndex) const;

		bool operator== (const BitVector& _rOther);
		bool operator!= (const BitVector& _rOther)
			{ return (false == operator== (_rOther)); };
};



//															
template <size_t FEATURES>
BitVector<FEATURES>::BitVector (void)
{
	Clear ();
}


//															
template <size_t FEATURES>
void BitVector<FEATURES>::Clear (void)
{
	memset (c_Features, 0, 1 + FEATURES / (8 * sizeof(char)));
}


//															
template <size_t FEATURES>
void BitVector<FEATURES>::Memset (int _iValue)
{
	assert (0 == _iValue);
	Clear ();
}


//															
template <size_t FEATURES>
void BitVector<FEATURES>::SetFeature (int _iIndex, int _iValue)
{
	int iByteIndex = _iIndex / (8 * sizeof(char));
	int iBitIndex = 1 << (_iIndex % (8 * sizeof(char)));

	if (0 == _iValue)
		c_Features [iByteIndex] &= ~iBitIndex;
	else if (1 == _iValue)
		c_Features [iByteIndex] |= iBitIndex;
	else
		assert (false);
}



//															
template <size_t FEATURES>
void BitVector<FEATURES>::CopyVector (const BitVector& _rSource)
{
	memcpy (c_Features, _rSource.c_Features, 1 + FEATURES / (8 * sizeof(char)));
}



//															
template <size_t FEATURES>
const int BitVector<FEATURES>::operator[] (int _iIndex) const
{
	int iByteIndex = _iIndex / (8 * sizeof(char));
	int iBitIndex = 1 << (_iIndex % (8 * sizeof(char)));

	return (0 == (c_Features [iByteIndex] & iBitIndex))? 0 : 1;
}



//															
template <size_t FEATURES>
bool BitVector<FEATURES>::operator== (const BitVector<FEATURES>& _rOther)
{
	return (0 == memcmp (c_Features, 
						 _rOther.c_Features,
						 1 + FEATURES / (8 * sizeof(char))));
}


//															
template <size_t FEATURES>
ostream& operator<< (ostream& _rStream, const BitVector<FEATURES>& _rFeatures)
{
	for (size_t i = 0; i < FEATURES; ++ i)
		_rStream << _rFeatures [i] << ' ';
	return _rStream;
}


//															
template <size_t FEATURES>
String& operator<< (String& _rString, const BitVector<FEATURES>& _rFeatures)
{
	for (size_t i = 0; i < FEATURES; ++ i)
		_rString << _rFeatures [i] << ' ';
	return _rString;
}




#endif
