#ifndef __RLL_DOCUMENT__
#define __RLL_DOCUMENT__

#include <vector>
#include <map>
#include <set>
#include <nlp_string.h>
#include "Annotation.h"
using namespace std;


class TraceInfo;
class EnhancedState;
class Sentence;
class Document;
class EpisodeStep;

typedef set<char>							char_set_t;
typedef vector<String>						String_vec_t;
typedef vector<int>							int_vec_t;
typedef vector<Sentence*>					Sentence_vec_t;
typedef vector <Document*>					Document_vec_t;
typedef map <String, Document*>				IdToDocument_map_t;
typedef map <String, String_dq_t*>			StartingStateToIds_map_t;
typedef map <EnhancedState*, TraceInfo>		VisitedStates_map_t;
typedef deque <EpisodeStep*>			EpisodeStep_dq_t;


enum DocumentType_e
{
	dt_training,
	dt_development,
	dt_test
};

std::ostream& operator<< (std::ostream& _rStream, const DocumentType_e _eType);


// ------------------------------------------
class TraceInfo
{
	public:
		AnnotatedAction_dq_t	dq_Actions;
		float					f_Reward;

		~TraceInfo (void);
		void RememberActions (EpisodeStep_dq_t& _rdqSteps, int _iEnd);
};


// ------------------------------------------
class WordLocation
{
	public:
		int	i_SentenceIndex;
		int	i_WordIndex;
};


// ------------------------------------------
class Sentence
{
	public:
		int_vec_t		vec_WordIndices;
		String_vec_t	vec_Words;
		int				i_Length;
};

std::ostream& operator<< (std::ostream& _rStream, const Sentence& _rSentence);


// ------------------------------------------
class Document
{
	private:
		static IdToDocument_map_t			map_IdToDocument;
		static StartingStateToIds_map_t		map_StartingStateToIds;
		static Document_vec_t				vec_Documents;

		static void SplitToWords (String _sLine, String_dq_t* _pdqWords);
		static bool LoadHeldoutIds (void);
		static bool LoadDocuments (void);


		WordLocation*		parr_LocationToWord;

		void ComputeWordIndices (void);


	public:
		static String_dq_t	dq_StartingStates;

		String				s_Id;
		String				s_RequiredStartingState;
		DocumentType_e		e_Type;
		Sentence_vec_t		vec_Sentences;
		EnhancedState*		p_StartingEnhancedState;
		EnhancedState*		p_TrainingEnhancedState;
		EnhancedState*		p_TestEnhancedState;
		VisitedStates_map_t	map_VisitedStates;
		int					i_WordCount;


		Document (void);
		~Document (void);
		void SetType (DocumentType_e _eType);

		String sGetWord (int _iLocation)
		{
			if (_iLocation == i_WordCount)
				return "";
			WordLocation& rLoc = parr_LocationToWord [_iLocation];
			return vec_Sentences [rLoc.i_SentenceIndex]->vec_Words [rLoc.i_WordIndex];
		};

		int iGetWord (int _iLocation)
		{
			if (_iLocation == i_WordCount)
				return -1;
			WordLocation& rLoc = parr_LocationToWord [_iLocation];
			return vec_Sentences [rLoc.i_SentenceIndex]->vec_WordIndices [rLoc.i_WordIndex];
		};

		int GetSentenceIndex (int _iLocation)
		{
			if (_iLocation == i_WordCount)
				return -1;
			WordLocation& rLoc = parr_LocationToWord [_iLocation];
			return rLoc.i_SentenceIndex;
		};

		void RememberStates (EpisodeStep_dq_t& _dqSteps);
		AnnotatedAction_dq_t* PickStartState (void);


		static bool Load (void);
		static void Clear (void);

		static void ForgetTestStateGraph (void);
		static void SwitchToTrainingStateGraph (void);
		static void SwitchToTestStateGraph (void);
		static void DumpRewards (int _iIteration);

		static Document* Find (const String& _sId);
		static String_dq_t* GetIdsForStartingState (String& _sStartingState);
		static int DocumentCount (void)
			{ return vec_Documents.size (); };
		static Document_vec_t* GetDocuments (void)
			{ return &vec_Documents; };
};

std::ostream& operator<< (std::ostream& _rStream, const Document& _rDocument);





// ------------------------------------------
class TextInterface
{
	private:
		Document*					p_Document;
		int							i_WordCount;
		char_set_t					set_UsedWords;
		char						i_LastLocation_WU;
		char						i_LastLocation_WC;
		char						i_LastLocation_WP;

	public:
		TextInterface (void);
		~TextInterface (void);

		bool StartEpisode (String& _rDocumentId);
		void EndEpisode (void);

		void UseWords (char _lwc, char _lwu, char _lwp);
		void UseWordSpan (char _lwc, char _lwu, char _lwp);

		inline Document* GetCurrentDocument (void)
			{ return p_Document; };
		inline int GetWordCount (void)
			{ return i_WordCount; };
		inline int LastUsedCommandWord (void)
			{ return i_LastLocation_WC; };
		inline int LastUsedObjectWord (void)
			{ return i_LastLocation_WU; };
		inline int LastUsedParameterWord (void)
			{ return i_LastLocation_WP; };
		inline void GetUsedWords (char_set_t* _psetUsedWords)
			{ *_psetUsedWords = set_UsedWords; };
};


#include "State.h"

#endif
