#include "Learner.h"
#include <nlp_macros.h>
#include <nlp_config.h>
#include "Annotation.h"
#include "Sample.h"


//															
Episode::Episode (Learner* _pLearner, String& _rDocumentId)
{
	p_Learner = _pLearner;
	s_DocumentId = _rDocumentId;
	p_LastStep = NULL;
	p_CurrentEnhancedState = NULL;
	b_WeightUpdatesComputed = false;
	b_OnAnnotatedPath = true;
	i_LastActionSentence = -1;
}



//															
Episode::~Episode (void)
{
	assert ((Learner*) 0x1 != p_Learner);
	p_Learner = (Learner*) 0x1;

	ITERATE (EpisodeStep_dq_t, dq_Steps, ite)
		delete *ite;
	dq_Steps.clear ();
}



//															
EpisodeState_e Episode::ObserveStateAndAct (ExperienceInterface* _pExperienceInterface,
											TextInterface* _pTextInterface,
											bool _bAllowWordlessAction,
											bool _bFromNextSentence,
											TrainTest_e _eTrainTestMode,
											AnnotatedAction* _pAnnotatedAction)
{
	// we can't be in test mode, and have an annotation given to us...	
	// assert ((tt_test_mode != _eTrainTestMode) || (NULL == _pAnnotatedAction));

	if (p_Learner->i_Verbose > 1)
		cout << "Episode::ObserveStateAndAct" << endl;
	
	// observe current world state ...			
	bool bFirstAction = (true == dq_Steps.empty ());
	EpisodeStep* pStep = new EpisodeStep (p_Learner);
	EnvResponse_e eObserve = pStep->ObserveState (p_CurrentEnhancedState, 
													_pExperienceInterface,
													_pTextInterface);
	if (er_ok != eObserve)
	{
		if ((er_no_object_words != eObserve) || (true != _bAllowWordlessAction))
		{
			dq_Steps.push_back (pStep);
			p_LastStep = pStep;
		}

		if ((er_no_object_words == eObserve) && (false == _bAllowWordlessAction))
			return es_no_words_left;
		else if (er_no_visible_objects == eObserve)
			return es_no_object_targets_left;
		else if (er_env_error == eObserve)
			return es_error_in_apply_action;
	}


	// set sentence to select actions from...			
	{
		if (true == _bFromNextSentence)
		{
			assert (NULL != p_LastStep);

			int iLastActionObjectWord = p_LastStep->o_RLAction.i_SelectedObjectWordLocation;
			if (p_LastStep->o_RLState.TextLength () != iLastActionObjectWord)
			{
				i_LastActionSentence = p_LastStep->o_RLState.GetSentenceIndex (iLastActionObjectWord);
				assert (-1 != i_LastActionSentence);
			}
			pStep->i_FirstCandidateActionSentence = i_LastActionSentence + 1;
		}
		else if (NULL != p_LastStep)
		{
			int iLastActionObjectWord = p_LastStep->o_RLAction.i_SelectedObjectWordLocation;
			if (p_LastStep->o_RLState.TextLength () != iLastActionObjectWord)
			{
				i_LastActionSentence = p_LastStep->o_RLState.GetSentenceIndex (iLastActionObjectWord);
				assert (-1 != i_LastActionSentence);
			}
			pStep->i_FirstCandidateActionSentence  = i_LastActionSentence;
		}
	}

	// check if we're in the last sentence, and if so	
	// don't allow wordless actions...					
	{
		int iLastSentence = pStep->o_RLState.GetSentenceIndex (pStep->o_RLState.TextLength () - 1);
		if (pStep->i_FirstCandidateActionSentence == iLastSentence)
			_bAllowWordlessAction = false;
	}

	// compute Q values, or lookup if already computed.
	EpisodeStepState_e eState = ess_unknown;
	eState = pStep->ComputeProbs (bFirstAction, _bAllowWordlessAction);
	if (ess_no_words_left == eState)
	{
		if ((false == _bFromNextSentence) &&
			(false == EpisodeStep::b_LookAtNextSentenceAlso))
		{
			if (p_LastStep == pStep)
				dq_Steps.pop_back ();
			delete pStep;
			return es_try_next_sentence;
		}
		dq_Steps.push_back (pStep);
		p_LastStep = pStep;
		return es_no_words_left;
	}
	else if (ess_no_object_targets_left == eState)
	{
		if ((false == _bFromNextSentence) &&
			(false == EpisodeStep::b_LookAtNextSentenceAlso))
		{
			if (p_LastStep == pStep)
				dq_Steps.pop_back ();
			delete pStep;
			return es_try_next_sentence;
		}
		dq_Steps.push_back (pStep);
		p_LastStep = pStep;
		return es_no_object_targets_left;
	}


	// select action based on Q values ...		
	if ((true == p_Learner->b_RandomlyStrayFromAnnotations) &&
		(true == b_OnAnnotatedPath))
	{
		if (true == Sample::SampleCoinflip (p_Learner->d_StrayProbability))
		{
			cout << " [--] Straying..." << endl;
			b_OnAnnotatedPath = false;
		}
	}

	eState = ess_unknown;
	if (NULL == _pAnnotatedAction)
		eState = pStep->SelectAction (_eTrainTestMode, p_Learner->d_ExplorationRate);
	else
	{
		if (true == b_OnAnnotatedPath)
			eState = pStep->WalkAction (_pAnnotatedAction);
		else
			eState = pStep->SelectAction (_eTrainTestMode, p_Learner->d_ExplorationRate);
	}


	// check if we need to retry with the next	
	// sentence ...								
	if (ess_all_actions_zero_prob == eState)
	{
		if (true == _bFromNextSentence)
		{
			dq_Steps.push_back (pStep);
			p_LastStep = pStep;
		}
		else
		{
			if (p_LastStep == pStep)
				dq_Steps.pop_back ();
			delete pStep;
		}

		// this should normally not happen...		
		if (NULL == p_LastStep)
			return es_error_in_select_action;

		return es_try_next_sentence;
	}


	//											
	dq_Steps.push_back (pStep);
	p_LastStep = pStep;


	// check for error conditions ...			
	if ((ess_ok != eState) &&
		 (ess_action_without_words != eState))
	{
		cout << "   error in select action" << endl;
		Learner::ofs_Log << "   [ERROR] error in select action" << endl;
		if ((NULL != _pAnnotatedAction) && (true == b_OnAnnotatedPath))
			Learner::ofs_Log << "   was following trace." << endl;
		return es_error_in_select_action;
	}

	// cout << *pStep << endl;

	// execute action in the world ...			
	EnvResponseToAction_e eResponse = pStep->ApplySelectedAction (_pExperienceInterface, 
																  _pTextInterface,
																  _eTrainTestMode);
	if (er_error == eResponse)
	{
		cout << "   error in apply action" << endl;
		Learner::ofs_Log << "   [ERROR] error in apply action" << endl;
		return es_error_in_apply_action;
	}

	pStep->b_ActionExecutedWithoutError = true;
	UpdateEnhancedStateGraph (pStep->o_RLAction.GetAsLong ());

	if (er_env_unchanged == eResponse)
	{
		if (ess_action_without_words == eState)
			return es_env_unchanged_action_without_words;
		return es_env_unchanged;
	}

	if (ess_action_without_words == eState)
		return es_action_without_words_ok;
	return es_ok;
}



//															
void Episode::UpdateEnhancedStateGraph (long _lAction)
{
	EnhancedState* pNextState = p_CurrentEnhancedState->GetNextState (_lAction);
	if (NULL == pNextState)
	{
		pNextState = new EnhancedState;
		p_CurrentEnhancedState->AddTransition (_lAction, pNextState);
	}
	
	p_CurrentEnhancedState = pNextState;
}



//															
void Episode::PrintFeedback (ofstream& _ofsLog)
{
	if (true == dq_Steps.empty ())
		return;

	cout << "   [feedback] ";
	_ofsLog << "   [feedback] ";
	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		cout << dq_Steps [t]->e_Feedback << ", ";
		_ofsLog << dq_Steps [t]->e_Feedback << ", ";
	}
	cout << endl;
	_ofsLog << endl;
}



//															
void Episode::PrintReward (ofstream& _ofsLog)
{
	if (true == dq_Steps.empty ())
		return;

	cout << "   [reward] ";
	_ofsLog << "   [reward] ";
	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		cout << dq_Steps [t]->o_RLState.GetReward () << ", ";
		_ofsLog << dq_Steps [t]->o_RLState.GetReward () << ", ";
	}
	cout << endl;
	_ofsLog << endl;
}



//															
bool Episode::ComputeTrace (ExperienceInterface* _pExperienceInterface,
							TextInterface* _pTextInterface,
							TrainTest_e _eTrainTestMode,
							AnnotatedAction_dq_t* _pdqActionsToStartState)
{
	Document* pDocument = Document::Find (s_DocumentId);

	bool bOk = true;
	int iWordlessActionCount = 0;
	p_CurrentEnhancedState = pDocument->p_StartingEnhancedState;

	assert (((tt_test_mode == _eTrainTestMode) && (p_CurrentEnhancedState == pDocument->p_TestEnhancedState)) ||
			((tt_training_mode == _eTrainTestMode) && (p_CurrentEnhancedState == pDocument->p_TrainingEnhancedState)));


	AnnotatedAction* pActionToStartState = NULL;
	AnnotatedAction_dq_t::iterator	iteAction;
	if (NULL != _pdqActionsToStartState)
		iteAction = _pdqActionsToStartState->begin ();

	bool bForceMoveToNextSentence = false;
	while (true)
	{
		pActionToStartState = NULL;
		if (NULL != _pdqActionsToStartState)
		{
			if (_pdqActionsToStartState->end () != iteAction)
			{
				pActionToStartState = *iteAction;
				++ iteAction;
			}
		}

		bool bAllowWordlessAction = (iWordlessActionCount < p_Learner->i_MaxWordlessActions);
		EpisodeState_e eState = Episode::ObserveStateAndAct (_pExperienceInterface,
															 _pTextInterface,
															 bAllowWordlessAction,
															 bForceMoveToNextSentence,
															 _eTrainTestMode,
															 pActionToStartState);

		// the actions taken to set the learner into 	
		// a randomly selected starting state should 	
		// not be used during the parameter updates,	
		// so we need to remember these actions...		
		if ((NULL != pActionToStartState) && (NULL != p_LastStep))
			p_LastStep->b_StartStatePrefix = true;

		//													
		if (es_try_next_sentence == eState)
		{
			// check if we forced a sentence move on the
			// last step.  If so, we're at a deadend...	
			if (true == bForceMoveToNextSentence)
			{
				p_LastStep->e_Feedback = ft_deadend;
				break;
			}

			assert (NULL != p_LastStep);
			cout << "   [--] retrying with the next sentence..." << endl;
			bForceMoveToNextSentence = true;
			continue;
		}


		//													
		bForceMoveToNextSentence = false;

		if ((es_no_words_left == eState) ||
			(es_no_object_targets_left == eState))
		{
			p_LastStep->e_Feedback = ft_no_words_objects;
			break;
		}

		else if (es_env_unchanged == eState)
		{
			p_LastStep->e_Feedback = ft_env_unchanged;
		}

		else if (es_env_unchanged_action_without_words == eState)
		{
			++ iWordlessActionCount;
			assert (iWordlessActionCount <= p_Learner->i_MaxWordlessActions);
			p_LastStep->e_Feedback = ft_env_unchanged;
		}

		else if (es_ok == eState)
		{
			p_LastStep->e_Feedback = ft_none;
			iWordlessActionCount = 0;
		}

		else if (es_action_without_words_ok == eState)
		{
			++ iWordlessActionCount;
			assert (iWordlessActionCount <= p_Learner->i_MaxWordlessActions);
			p_LastStep->e_Feedback = ft_none;
		}

		else if ((es_error_in_observe_state == eState) ||
				(es_error_in_select_action == eState) ||
				(es_error_in_apply_action == eState))
		{
			p_LastStep->e_Feedback = ft_state_action_error;
			bOk = false;
			break;
		}

		else
		{
			p_LastStep->e_Feedback = ft_comm_error;
			bOk = false;
			break;
		}
	}

	return bOk;
}



//															
bool Episode::WalkReplayTrace (ExperienceInterface* _pExperienceInterface,
								TextInterface* _pTextInterface,
								ReplayTrace* _pReplayTrace)
{
	Document* pDocument = Document::Find (s_DocumentId);

	bool bOk = true;
	int iWordlessActionCount = 0;
	p_CurrentEnhancedState = pDocument->p_StartingEnhancedState;

	ReplayTraceSentence_dq_t::iterator	iteReplaySentence;
	iteReplaySentence = _pReplayTrace->dq_Sentence.begin ();
	if (_pReplayTrace->dq_Sentence.end () == iteReplaySentence)
		return false;

	ReplayTraceSentence* pReplaySentence = *iteReplaySentence;

	AnnotatedAction_dq_t::iterator	iteReplayAction;
	iteReplayAction = pReplaySentence->dq_Action.begin ();
	assert (pReplaySentence->dq_Action.end () != iteReplayAction);
	AnnotatedAction* pReplayAction = *iteReplayAction;


	bool bForceMoveToNextSentence = false;
	while (true)
	{
		bool bAllowWordlessAction = (iWordlessActionCount < p_Learner->i_MaxWordlessActions);
		EpisodeState_e eState = Episode::ObserveStateAndAct (_pExperienceInterface,
															 _pTextInterface,
															 bAllowWordlessAction,
															 bForceMoveToNextSentence,
															 tt_training_mode,
															 pReplayAction);


		//													
		if (es_try_next_sentence == eState)
		{
			// check if we forced a sentence move on the
			// last step.  If so, we're at a deadend...	
			if (true == bForceMoveToNextSentence)
			{
				p_LastStep->e_Feedback = ft_deadend;
				break;
			}

			cout << "   [--] retrying with the next sentence..." << endl;
			bForceMoveToNextSentence = true;
			continue;
		}


		//													
		bForceMoveToNextSentence = false;

		if ((es_no_words_left == eState) ||
			(es_no_object_targets_left == eState))
		{
			p_LastStep->e_Feedback = ft_no_words_objects;
			break;
		}

		else if (es_env_unchanged == eState)
		{
			p_LastStep->e_Feedback = ft_env_unchanged;
		}

		else if (es_env_unchanged_action_without_words == eState)
		{
			++ iWordlessActionCount;
			assert (iWordlessActionCount <= p_Learner->i_MaxWordlessActions);
			p_LastStep->e_Feedback = ft_env_unchanged;
		}

		else if (es_ok == eState)
		{
			p_LastStep->e_Feedback = ft_none;
			iWordlessActionCount = 0;
		}

		else if (es_action_without_words_ok == eState)
		{
			++ iWordlessActionCount;
			assert (iWordlessActionCount <= p_Learner->i_MaxWordlessActions);
			p_LastStep->e_Feedback = ft_none;
		}

		else if ((es_error_in_observe_state == eState) ||
				(es_error_in_select_action == eState) ||
				(es_error_in_apply_action == eState))
		{
			p_LastStep->e_Feedback = ft_state_action_error;
			bOk = false;
			break;
		}

		else
		{
			p_LastStep->e_Feedback = ft_comm_error;
			bOk = false;
			break;
		}


		// pick next action annotation...		
		bool bStopBeforeNextAction = false;
		++ iteReplayAction;
		if (pReplaySentence->dq_Action.end () == iteReplayAction)
		{
			++ iteReplaySentence;
			if (_pReplayTrace->dq_Sentence.end () == iteReplaySentence)
				bStopBeforeNextAction = true;
			else
			{
				pReplaySentence = *iteReplaySentence;
				iteReplayAction = pReplaySentence->dq_Action.begin ();
				if (pReplaySentence->dq_Action.end () == iteReplayAction)
					cout << "[ERROR]  Replay actions not present for sentence?  Document [id="
						 << s_DocumentId << ']' << endl;
				assert (pReplaySentence->dq_Action.end () != iteReplayAction);
			}
		}
		if (false == bStopBeforeNextAction)
			pReplayAction = *iteReplayAction;
	}

	return bOk;
}



//															
bool Episode::WalkAnnotationTrace (ExperienceInterface* _pExperienceInterface,
								   TextInterface* _pTextInterface)
{
	Document* pDocument = Document::Find (s_DocumentId);
	Annotation* pAnnotation = Annotation::Find (s_DocumentId);

	bool bOk = true;
	int iWordlessActionCount = 0;


	AnnotatedSentence_dq_t::iterator	iteAnnotatedSentence;
	iteAnnotatedSentence = pAnnotation->dq_Sentence.begin ();
	assert (pAnnotation->dq_Sentence.end () != iteAnnotatedSentence);
	AnnotatedSentence* pAnnotatedSentence = *iteAnnotatedSentence;

	AnnotatedAction_dq_t::iterator		iteAnnotatedAction;
	iteAnnotatedAction = pAnnotatedSentence->dq_Action.begin ();
	assert (pAnnotatedSentence->dq_Action.end () != iteAnnotatedAction);
	AnnotatedAction* pAnnotatedAction = *iteAnnotatedAction;

	p_CurrentEnhancedState = pDocument->p_StartingEnhancedState;

	assert (p_CurrentEnhancedState == pDocument->p_TrainingEnhancedState);

	bool bForceMoveToNextSentence = false;
	while (true)
	{
		bool bAllowWordlessAction = (iWordlessActionCount < p_Learner->i_MaxWordlessActions);
		EpisodeState_e eState = Episode::ObserveStateAndAct (_pExperienceInterface,
															 _pTextInterface,
															 bAllowWordlessAction,
															 bForceMoveToNextSentence,
															 tt_training_mode,
															 pAnnotatedAction);


		// We would get an es_try_next_sentence only if	
		// walk-action has switched to epsilon-greedy	
		// exploration...								
		if (es_try_next_sentence == eState)
		{
			// check if we forced a sentence move on the
			// last step.  If so, we're at a deadend...	
			if (true == bForceMoveToNextSentence)
			{
				p_LastStep->e_Feedback = ft_deadend;
				break;
			}

			assert (NULL != p_LastStep);
			cout << "   [--] retrying with the next sentence..." << endl;
			bForceMoveToNextSentence = true;
			bOk = false;
			continue;
		}

		bForceMoveToNextSentence = false;


		//													
		if (es_ok == eState)
			p_LastStep->e_Feedback = ft_none;

		else if (es_action_without_words_ok == eState)
		{
			++ iWordlessActionCount;
			p_LastStep->e_Feedback = ft_none_wordless;
		}

		else if (es_env_unchanged == eState)
			p_LastStep->e_Feedback = ft_env_unchanged;

		else if (es_env_unchanged_action_without_words == eState)
		{
			++ iWordlessActionCount;
			p_LastStep->e_Feedback = ft_env_unchanged_wordless;
		}

		else
		{
			cout << "[ERROR] Couldn't walk the annotations!  Document [id=" 
				 << s_DocumentId << "]" << endl;
			if (es_no_words_left == eState)
			{
				p_LastStep->e_Feedback = ft_no_words_objects;
				cout << "        All words consumed." << endl;
			}
			else if (es_no_object_targets_left == eState)
			{
				p_LastStep->e_Feedback = ft_no_words_objects;
				cout << "        No objects with text overlap." << endl;
			}
			else if (es_error_in_observe_state == eState)
			{
				p_LastStep->e_Feedback = ft_state_action_error;
				cout << "        Error in observe state." << endl;
			}
			else if (es_error_in_select_action == eState)
			{
				p_LastStep->e_Feedback = ft_state_action_error;
				cout << "        Error in select action." << endl;
			}
			else if (es_error_in_apply_action == eState)
			{
				p_LastStep->e_Feedback = ft_state_action_error;
				cout << "        Error in applying selected action." << endl;
			}
			else
			{
				cout << "        Unknown error : " << eState << endl;
				assert (false);
			}
			bOk = false;
			break;
		}


		// pick next action annotation...		
		++ iteAnnotatedAction;
		if (pAnnotatedSentence->dq_Action.end () == iteAnnotatedAction)
		{
			++ iteAnnotatedSentence;
			if (pAnnotation->dq_Sentence.end () == iteAnnotatedSentence)
				break;
			pAnnotatedSentence = *iteAnnotatedSentence;
			iteAnnotatedAction = pAnnotatedSentence->dq_Action.begin ();
			if (pAnnotatedSentence->dq_Action.end () == iteAnnotatedAction)
				cout << "[ERROR]  Annotations not present for sentence?  Document [id="
					 << s_DocumentId << ']' << endl;
			assert (pAnnotatedSentence->dq_Action.end () != iteAnnotatedAction);
		}
		pAnnotatedAction = *iteAnnotatedAction;
	}

	return bOk;
}



//															
bool Episode::ComputeReward (String& _sId, TrainTest_e _eTrainTestMode)
{
	Document* pDocument = Document::Find (_sId);
	assert (NULL != pDocument);


	// If in test mode, we can't use annotations ...			
	if (tt_test_mode == _eTrainTestMode)
		return Episode::ComputeRewardUsingEnvironment ();

	// Learning from annotations only...						
	else if (lm_annotation_only == p_Learner->e_LearningMode)
		return Episode::ComputeRewardUsingAnnotation (pDocument);

	// Learning from the environment only...					
	else if (lm_environment_only == p_Learner->e_LearningMode)
		return Episode::ComputeRewardUsingEnvironment ();

	// Learning from both the environment and the annotations.	
	else if (lm_mixed_mode == p_Learner->e_LearningMode)
	{
		Annotation* pAnnotation = Annotation::Find (pDocument->s_Id);
		if (NULL != pAnnotation)
			return Episode::ComputeRewardUsingAnnotation (pDocument);
		else
			return Episode::ComputeRewardUsingEnvironment ();
	}
	else
		cerr << "[ERROR] Learning mode has not been set in the configuration file, or is incorrect."
			 << endl;

	return false;
}


//															
bool Episode::ComputeRewardUsingEnvironment (void)
{
	String sRewardType = (config)"reward_type";
	if ("1" == sRewardType)
		return ComputeRewardUsingEnvironment_v1 ();
	else if ("1.5" == sRewardType)
		return ComputeRewardUsingEnvironment_v1_5 ();
	else if ("2" == sRewardType)
		return ComputeRewardUsingEnvironment_v2 ();
	else if ("3" == sRewardType)
		return ComputeRewardUsingEnvironment_v3 ();
	else if ("4" == sRewardType)
		return ComputeRewardUsingEnvironment_v4 ();
	else
		abort ();
}


//															
bool Episode::ComputeRewardUsingEnvironment_v4 (void)
{
	if (true == dq_Steps.empty ())
		return false;

	Document* pDocument = Document::Find (s_DocumentId);
	int iSentenceCount = 1 + pDocument->GetSentenceIndex (pDocument->i_WordCount -1);


	// find last action sentence...
	int iLastActionSentence = -1;
	Feedback_e eLastFeedback = ft_unknown;
	int_set_t setSentencesUsed;
	bool bHasNoneTypingAction = false;
	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		EpisodeStep* pStep = dq_Steps [t];
		RLAction& rAction = pStep->o_RLAction;
		eLastFeedback = pStep->e_Feedback;

		int lwu = rAction.i_SelectedObjectWordLocation;
		if (lwu == pDocument->i_WordCount)
		{
			assert (p_Learner->i_MaxWordlessActions > 0);
			continue;
		}

		if ((ft_no_words_objects != pStep->e_Feedback) &&
			(ft_deadend != pStep->e_Feedback) &&
			(ft_state_action_error != pStep->e_Feedback))
		{
			iLastActionSentence = pDocument->GetSentenceIndex (lwu);
			setSentencesUsed.insert (iLastActionSentence);

			String sCommand = Command::Get (rAction.i_SelectedCommandIndex);
			if (false == sCommand.Has ("type"))
				bHasNoneTypingAction = true;
		}
	}


	// compute rewards ...						
	double dPositive = Learner::d_PositiveRewardMultiplier;
	double dNegative = Learner::d_NegativeRewardMultiplier;

	double dRewardType = -1 * dNegative;
	if ((ft_deadend != eLastFeedback) && 
		(ft_state_action_error != eLastFeedback) &&
		(ft_comm_error != eLastFeedback))
	{
		if ((setSentencesUsed.size () == (size_t)iSentenceCount) &&
			(true == bHasNoneTypingAction))
			dRewardType = dPositive;
	}

	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		EpisodeStep* pStep = dq_Steps [t];
		bool bLastStep = (t == dq_Steps.size () -1);
		
		if (ft_none == pStep->e_Feedback)
		{
			if (true == bLastStep)
				pStep->o_RLState.SetReward (dRewardType);
			else
				pStep->o_RLState.SetReward (0.0001 * dNegative);
		}

		else if (ft_none_wordless == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-0.9 * dNegative);

		else if (ft_no_words_objects == pStep->e_Feedback)
		{
			if (true == bLastStep)
				pStep->o_RLState.SetReward (dRewardType);
			else
				pStep->o_RLState.SetReward (0.0001 * dNegative);
		}

		else if (ft_repetition == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-0.9 * dNegative);

		else if (ft_env_unchanged == pStep->e_Feedback)
		{
			if (true == bLastStep)
				pStep->o_RLState.SetReward (dRewardType);
			else
				pStep->o_RLState.SetReward (0.0001 * dNegative);
		}

		else if (ft_env_unchanged_wordless == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-0.9 * dNegative);


		// special cases...									
		else if (ft_deadend == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-1 * dNegative);

		else if (ft_state_action_error == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-1 * dNegative);

		else if (ft_comm_error == pStep->e_Feedback)
			pStep->o_RLState.SetReward (0);

		else
			abort ();
	}

	return true;
}



//															
bool Episode::ComputeRewardUsingEnvironment_v3 (void)
{
	if (true == dq_Steps.empty ())
		return false;

	Document* pDocument = Document::Find (s_DocumentId);
	size_t iSentenceCount = 1 + pDocument->GetSentenceIndex (pDocument->i_WordCount -1);


	// find last action sentence...
	int iLastActionSentence = -1;
	Feedback_e eLastFeedback = ft_unknown;
	int_set_t setSentencesUsed;
	bool bHasNoneTypingAction = false;
	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		EpisodeStep* pStep = dq_Steps [t];
		RLAction& rAction = pStep->o_RLAction;
		eLastFeedback = pStep->e_Feedback;

		int lwu = rAction.i_SelectedObjectWordLocation;
		if (lwu == pDocument->i_WordCount)
		{
			assert (p_Learner->i_MaxWordlessActions > 0);
			continue;
		}

		if ((ft_no_words_objects != pStep->e_Feedback) &&
			(ft_deadend != pStep->e_Feedback) &&
			(ft_state_action_error != pStep->e_Feedback))
		{
			iLastActionSentence = pDocument->GetSentenceIndex (lwu);
			setSentencesUsed.insert (iLastActionSentence);

			String sCommand = Command::Get (rAction.i_SelectedCommandIndex);
			if (false == sCommand.Has ("type"))
				bHasNoneTypingAction = true;
		}
	}


	// compute rewards ...						
	double dPositive = Learner::d_PositiveRewardMultiplier;
	double dNegative = Learner::d_NegativeRewardMultiplier;

	double dRewardType = -1 * dNegative;
	if ((ft_deadend != eLastFeedback) && 
		(ft_state_action_error != eLastFeedback) &&
		(ft_comm_error != eLastFeedback))
	{
		if ((setSentencesUsed.size () == iSentenceCount) &&
			(true == bHasNoneTypingAction))
			dRewardType = dPositive;
	}

	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		EpisodeStep* pStep = dq_Steps [t];
		bool bWordless = false;
		
		
		if (ft_none == pStep->e_Feedback)
		{
			if (false == bWordless)
				pStep->o_RLState.SetReward (dRewardType);
			else
				pStep->o_RLState.SetReward (-0.49 * dNegative);
		}

		else if (ft_none_wordless == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-0.99 * dNegative);

		else if (ft_no_words_objects == pStep->e_Feedback)
			pStep->o_RLState.SetReward (dRewardType);

		else if (ft_repetition == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-0.9 * dNegative);

		else if (ft_env_unchanged == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-0.99 * dNegative);

		else if (ft_env_unchanged_wordless == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-0.99 * dNegative);


		// special cases...									
		else if (ft_deadend == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-1 * dNegative);

		else if (ft_state_action_error == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-1 * dNegative);

		else if (ft_comm_error == pStep->e_Feedback)
			pStep->o_RLState.SetReward (0);

		else
			abort ();
	}

	if (dRewardType < 0)
	{
		for (size_t t = 0; t < dq_Steps.size () - 1; ++ t)
		{
			EpisodeStep* pStep = dq_Steps [t];
			pStep->o_RLState.SetReward (0.01 * dNegative);
		}
	}

	return true;
}



//															
bool Episode::ComputeRewardUsingEnvironment_v2 (void)
{
	if (true == dq_Steps.empty ())
		return false;

	Document* pDocument = Document::Find (s_DocumentId);
	size_t iSentenceCount = 1 + pDocument->GetSentenceIndex (pDocument->i_WordCount -1);


	// compute length of each sentence ...		
	Vector<int> vecSentenceWords;
	vecSentenceWords.Create (iSentenceCount);
	vecSentenceWords.Memset (0);

	Vector<int> vecSentenceLength;
	vecSentenceLength.Create (iSentenceCount);
	vecSentenceLength.Memset (0);

	for (int w = 0; w < pDocument->i_WordCount; ++ w)
	{
		int iSentence = pDocument->GetSentenceIndex (w);
		++ vecSentenceWords [iSentence];
		vecSentenceLength [iSentence] += pDocument->sGetWord (w).length ();
	}


	// compute potential goal count...			
	GraphCache* pCache = p_Learner->p_ExperienceInterface->GetCache ();

	int iPotentialGoals = 0;
	int_set_t setGoalWordLocations;
	for (int i = 0; i < pDocument->i_WordCount; ++ i)
	{
		int iWord = pDocument->iGetWord (i);
		String sWord = pDocument->sGetWord (i);
		if (NULL != pCache->GetObjectsForWordStrict (iWord))
		{
			++ iPotentialGoals;
			setGoalWordLocations.insert (i);
		}
	}

	

	// compute word consumption & action counts	
	// for each sentence ...					
	Vector<int> vecSentenceWordsConsumed;
	vecSentenceWordsConsumed.Create (iSentenceCount);
	vecSentenceWordsConsumed.Memset (0);

	Vector<int> vecSentenceActions;
	vecSentenceActions.Create (iSentenceCount);
	vecSentenceActions.Memset (0);

	Vector<int> vecActionIncreasedObjects;
	vecActionIncreasedObjects.Create (dq_Steps.size ());
	vecActionIncreasedObjects.Memset (0);

	int iPotentialGoalsHit = 0;
	int_set_t setSentencesUsed;
	bool bHasNoneTypingAction = false;
	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		EpisodeStep* pStep = dq_Steps [t];
		RLAction& rAction = pStep->o_RLAction;

		// check if we hit a potential goal...	
		if (setGoalWordLocations.end () 
				!= setGoalWordLocations.find (rAction.i_SelectedObjectWordLocation))
		{
			assert (NULL != rAction.p_SelectedObject);

			String sGoalWord = pDocument->sGetWord (rAction.i_SelectedObjectWordLocation);
			if (true == RLState::WordMaybeObjectStrict (sGoalWord, rAction.p_SelectedObject->s_CleanName))
				++ iPotentialGoalsHit;
		}

		//										
		RLState& rState = pStep->o_RLState;
		RLState* pNextState = NULL;
		if (t + 1 < dq_Steps.size ())
		{
			pNextState = &(dq_Steps [t+1]->o_RLState);

			char* pUnusedWordLocations = rState.UnusedWordLocations ();
			int iUnusedWordCount = rState.UnusedWordCount ();

			// low precision text-object overlap...		
			{
				int_set_t setStateObjects 
							= rState.PotentialObjectWordsCoveredByState (pUnusedWordLocations,
																		 iUnusedWordCount);
				int_set_t setNextStateObjects
							= pNextState->PotentialObjectWordsCoveredByState (pUnusedWordLocations,
																			  iUnusedWordCount);

				ITERATE (int_set_t, setNextStateObjects, ite)
				{
					if (setStateObjects.end () != setStateObjects.find (*ite))
						continue;
					vecActionIncreasedObjects [t] = 1;
					break;
				}
			}

			// high precision text-object overlap...	
			{
				int_set_t setStateObjects 
							= rState.PotentialObjectWordsCoveredByStateStrict (pUnusedWordLocations,
																			 iUnusedWordCount);
				int_set_t setNextStateObjects
							= pNextState->PotentialObjectWordsCoveredByStateStrict (pUnusedWordLocations,
																			 iUnusedWordCount);

				ITERATE (int_set_t, setNextStateObjects, ite)
				{
					if (setStateObjects.end () != setStateObjects.find (*ite))
						continue;
					vecActionIncreasedObjects [t] = 2;
					break;
				}
			}
		}

		if (-1 != rAction.i_SelectedCommandIndex)
		{
			String sCommand = Command::Get (rAction.i_SelectedCommandIndex);
			if (false == sCommand.Has ("type"))
				bHasNoneTypingAction = true;
		}

		int lwc = rAction.i_SelectedCommandWordLocation;
		int lwu = rAction.i_SelectedObjectWordLocation;
		int lwp = rAction.i_SelectedParameterWordLocation;

		if ((-1 == lwu) || (lwu == pDocument->i_WordCount))
			continue;

		int iSentence = pDocument->GetSentenceIndex (lwu);
		++ vecSentenceActions [iSentence];
		setSentencesUsed.insert (iSentence);

		vecSentenceWordsConsumed [iSentence] += pDocument->sGetWord (lwc).length ();
		vecSentenceWordsConsumed [iSentence] += pDocument->sGetWord (lwu).length ();
		if (lwp < pDocument->i_WordCount)
			vecSentenceWordsConsumed [iSentence] += pDocument->sGetWord (lwp).length ();
	}


	// compute rewards ...						
	double dPositive = Learner::d_PositiveRewardMultiplier;
	double dNegative = Learner::d_NegativeRewardMultiplier;

	float fReward = NAN;
	float fPositiveReward = 0;
	int iSentence = -1;
	int iPreviousSentence = -1;
	int iLastActionIndex = -1;
	EpisodeStep* pStep = NULL;
	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		pStep = dq_Steps [t];
		RLAction& rAction = pStep->o_RLAction;
		

		int lwu = rAction.i_SelectedObjectWordLocation;
		if (lwu == pDocument->i_WordCount)
		{
			assert (p_Learner->i_MaxWordlessActions > 0);

			pStep->o_RLState.SetReward (fReward);
			fReward = 0.0001 * dNegative;
			continue;
		}


		if ((ft_no_words_objects != pStep->e_Feedback) &&
			(ft_deadend != pStep->e_Feedback) &&
			(ft_state_action_error != pStep->e_Feedback))
		{
			iSentence = pDocument->GetSentenceIndex (lwu);
		}

		if ((-1 != iPreviousSentence) && (iSentence != iPreviousSentence))
		{
			assert (false == isnan (fReward));
			fReward = fPositiveReward * dPositive;
		}
		iPreviousSentence = iSentence;



		int iWordsConsumed = vecSentenceWordsConsumed [iSentence];
		int iSentenceLength = vecSentenceLength [iSentence];

		float dWordRatio = iWordsConsumed / (float) iSentenceLength;
		dWordRatio = 1;

		fPositiveReward = 1;
		float fActionIncreasedCandidateObjects = vecActionIncreasedObjects [t] / (float) 2;

		if (ft_none == pStep->e_Feedback)
		{
			// reward carried over from action that	
			// got us to this state...				
			if (false == isnan (fReward))
				pStep->o_RLState.SetReward (fReward);

			if (true == bHasNoneTypingAction)
				fReward = 0.0001 * dNegative + 0.1 * fPositiveReward * dPositive * fActionIncreasedCandidateObjects;
			else
				fReward = 0.0001 * dNegative;
			// reward for ft_none needs to be		
			// assigned to the resulting state...	
		}

		else if (ft_none_wordless == pStep->e_Feedback)
		{
			// reward carried over from action that	
			// got us to this state...				
			if (false == isnan (fReward))
				pStep->o_RLState.SetReward (fReward);

			fReward = -10 * dNegative;
		}

		else if (ft_no_words_objects == pStep->e_Feedback)
		{
			if ((setSentencesUsed.size ()  == iSentenceCount) && (true == bHasNoneTypingAction))
				fReward = fPositiveReward * dPositive;
			else
				fReward = -1 * dNegative;

			pStep->o_RLState.SetReward (fReward);
			iLastActionIndex = t;
			fReward = NAN;
		}

		else if (ft_repetition == pStep->e_Feedback)
		{
			// reward carried over from action that	
			// got us to this state...				
			if (false == isnan (fReward))
				pStep->o_RLState.SetReward (fReward);

			fReward = -0.5 * dNegative;
			// reward for repetition needs to be	
			// assigned to the resulting state...	
		}

		else if (ft_env_unchanged == pStep->e_Feedback)
		{
			// reward carried over from action that	
			// got us to this state...				
			if (false == isnan (fReward))
				pStep->o_RLState.SetReward (fReward);

			fReward = -1 * dNegative;
			// reward for env_unchanged needs to be	
			// assigned to the resulting state...	
		}

		else if (ft_env_unchanged_wordless == pStep->e_Feedback)
		{
			// reward carried over from action that	
			// got us to this state...				
			if (false == isnan (fReward))
				pStep->o_RLState.SetReward (fReward);

			fReward = -10 * dNegative;
		}

		else if (ft_deadend == pStep->e_Feedback)
		{
			fReward = -1 * dNegative;
			pStep->o_RLState.SetReward (fReward);
			iLastActionIndex = t;
			fReward = NAN;
		}

		else if (ft_state_action_error == pStep->e_Feedback)
		{
			fReward = -0.9 * dNegative;
			pStep->o_RLState.SetReward (fReward);
			iLastActionIndex = t;
			fReward = NAN;
		}

		else if (ft_comm_error == pStep->e_Feedback)
		{
			fReward = 0;
			pStep->o_RLState.SetReward (fReward);
			iLastActionIndex = t;
			fReward = NAN;
		}

		else
			abort ();
	}
	if (true != isnan (fReward))
		abort ();

	// if we didn't get to the last sentence of the 
	// document, then the sentence we hit a deadend	
	// on possibly has bad actions...				
	int iLastActionSentence = iSentence;
	if (iLastActionSentence < (int)iSentenceCount - 1)
	{
		for (int t = 0; t < iLastActionIndex; ++ t)
		{
			EpisodeStep* pStep = dq_Steps [t];
			RLAction& rAction = pStep->o_RLAction;
			int lwu = rAction.i_SelectedObjectWordLocation;
			if (lwu == pDocument->i_WordCount)
				continue;

			if ((ft_no_words_objects != pStep->e_Feedback) &&
				(ft_deadend != pStep->e_Feedback) &&
				(ft_state_action_error != pStep->e_Feedback))
			{
				iSentence = pDocument->GetSentenceIndex (lwu);
			}

			if (iSentence == iLastActionSentence)
				pStep->o_RLState.SetReward (0.0001 * dNegative);
		}
	}
	
	return true;
}



//															
bool Episode::ComputeRewardUsingEnvironment_v1_5 (void)
{
	if (true == dq_Steps.empty ())
		return false;

	Document* pDocument = Document::Find (s_DocumentId);
	int iSentenceCount = 1 + pDocument->GetSentenceIndex (pDocument->i_WordCount -1);


	// compute length of each sentence ...		
	Vector<int> vecSentenceWords;
	vecSentenceWords.Create (iSentenceCount);
	vecSentenceWords.Memset (0);

	Vector<int> vecSentenceLength;
	vecSentenceLength.Create (iSentenceCount);
	vecSentenceLength.Memset (0);

	for (int w = 0; w < pDocument->i_WordCount; ++ w)
	{
		int iSentence = pDocument->GetSentenceIndex (w);
		++ vecSentenceWords [iSentence];
		vecSentenceLength [iSentence] += pDocument->sGetWord (w).length ();
	}


	// compute word consumption & action counts	
	// for each sentence ...					
	Vector<int> vecSentenceWordsConsumed;
	vecSentenceWordsConsumed.Create (iSentenceCount);
	vecSentenceWordsConsumed.Memset (0);

	Vector<int> vecSentenceActions;
	vecSentenceActions.Create (iSentenceCount);
	vecSentenceActions.Memset (0);

	Vector<size_t> vecSentenceLastAction;
	vecSentenceLastAction.Create (iSentenceCount);
	vecSentenceLastAction.Initialize (-1);

	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		EpisodeStep* pStep = dq_Steps [t];
		RLAction& rAction = pStep->o_RLAction;

		int lwc = rAction.i_SelectedCommandWordLocation;
		int lwu = rAction.i_SelectedObjectWordLocation;
		int lwp = rAction.i_SelectedParameterWordLocation;

		if ((-1 == lwu) || (lwu == pDocument->i_WordCount))
			continue;

		int iSentence = pDocument->GetSentenceIndex (lwu);
		++ vecSentenceActions [iSentence];
		vecSentenceLastAction [iSentence] = t;

		vecSentenceWordsConsumed [iSentence] += pDocument->sGetWord (lwc).length ();
		vecSentenceWordsConsumed [iSentence] += pDocument->sGetWord (lwu).length ();
		if (lwp < pDocument->i_WordCount)
			vecSentenceWordsConsumed [iSentence] += pDocument->sGetWord (lwp).length ();
	}


	// compute rewards ...						
	double dPositive = Learner::d_PositiveRewardMultiplier;
	double dNegative = Learner::d_NegativeRewardMultiplier;


	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		EpisodeStep* pStep = dq_Steps [t];
		RLAction& rAction = pStep->o_RLAction;
		EpisodeStep* pNextStep = NULL;
		if (t < dq_Steps.size () - 1)
			pNextStep = dq_Steps [t + 1];
		
		int lwu = rAction.i_SelectedObjectWordLocation;
		if (lwu == pDocument->i_WordCount)
		{
			assert (p_Learner->i_MaxWordlessActions > 0);
			pStep->o_RLState.SetReward (0.000001 * dNegative);
			continue;
		}

		int iSentence = -1;
		if ((ft_no_words_objects != pStep->e_Feedback) &&
			(ft_deadend != pStep->e_Feedback) &&
			(ft_state_action_error != pStep->e_Feedback))
		{
			iSentence = pDocument->GetSentenceIndex (lwu);
		}


		// compute word usage ratios...				
		int iWordsConsumed = vecSentenceWordsConsumed [iSentence];
		int iSentenceWords = vecSentenceWords [iSentence];
		int iSentenceLength = vecSentenceLength [iSentence];
		int iSentenceActions = vecSentenceActions [iSentence];

		float dWordRatio = iWordsConsumed / (float) iSentenceLength;
		float dActionRatio = (iSentenceWords - iSentenceActions) 
								/ (float) iSentenceWords;
		float fPositiveReward = dWordRatio * dActionRatio;

		bool bNextActionNotInError = false;
		if (NULL != pNextStep)
		{
			if ((ft_none == pNextStep->e_Feedback) ||
				(ft_repetition == pNextStep->e_Feedback) ||
				(ft_env_unchanged == pNextStep->e_Feedback))
				bNextActionNotInError = true;
		}


		// assign reward...							
		bool bLastActionForSentence = (vecSentenceLastAction [iSentence] == t);
		bool bIsLastSentence = (iSentence == iSentenceCount - 1);
		if (ft_none == pStep->e_Feedback)
		{
			if ((true == bLastActionForSentence) && (true == bNextActionNotInError))
				pStep->o_RLState.SetReward (fPositiveReward * dPositive);
			else
				// the reward below is theoretically 0, 
				// but set to a small positive value	
				// to avoid this state getting caught 	
				// to negative-state-avoidance.			
				pStep->o_RLState.SetReward (0.000001 * dNegative);
		}

		else if (ft_no_words_objects == pStep->e_Feedback)
		{
			if (true == bIsLastSentence)
				pStep->o_RLState.SetReward (fPositiveReward * dPositive);
			else
				pStep->o_RLState.SetReward (-1 * dNegative);
		}

		else if (ft_repetition == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-0.501 * dNegative);

		else if (ft_env_unchanged == pStep->e_Feedback)
			pStep->o_RLState.SetReward (0.000001 * dNegative);

		else if (ft_deadend == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-1 * dNegative);

		else if (ft_state_action_error == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-1 * dNegative);

		else if (ft_comm_error == pStep->e_Feedback)
			pStep->o_RLState.SetReward (-1 * dNegative);

		else
			abort ();
	}

	return true;
}



//															
bool Episode::ComputeRewardUsingEnvironment_v1 (void)
{
	if (true == dq_Steps.empty ())
		return false;

	Document* pDocument = Document::Find (s_DocumentId);
	int iSentenceCount = 1 + pDocument->GetSentenceIndex (pDocument->i_WordCount -1);


	// compute length of each sentence ...		
	Vector<int> vecSentenceWords;
	vecSentenceWords.Create (iSentenceCount);
	vecSentenceWords.Memset (0);

	Vector<int> vecSentenceLength;
	vecSentenceLength.Create (iSentenceCount);
	vecSentenceLength.Memset (0);

	for (int w = 0; w < pDocument->i_WordCount; ++ w)
	{
		int iSentence = pDocument->GetSentenceIndex (w);
		++ vecSentenceWords [iSentence];
		vecSentenceLength [iSentence] += pDocument->sGetWord (w).length ();
	}


	// compute word consumption & action counts	
	// for each sentence ...					
	Vector<int> vecSentenceWordsConsumed;
	vecSentenceWordsConsumed.Create (iSentenceCount);
	vecSentenceWordsConsumed.Memset (0);

	Vector<int> vecSentenceActions;
	vecSentenceActions.Create (iSentenceCount);
	vecSentenceActions.Memset (0);

	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		EpisodeStep* pStep = dq_Steps [t];
		RLAction& rAction = pStep->o_RLAction;

		int lwc = rAction.i_SelectedCommandWordLocation;
		int lwu = rAction.i_SelectedObjectWordLocation;
		int lwp = rAction.i_SelectedParameterWordLocation;

		if ((-1 == lwu) || (lwu == pDocument->i_WordCount))
			continue;

		int iSentence = pDocument->GetSentenceIndex (lwu);
		++ vecSentenceActions [iSentence];

		vecSentenceWordsConsumed [iSentence] += pDocument->sGetWord (lwc).length ();
		vecSentenceWordsConsumed [iSentence] += pDocument->sGetWord (lwu).length ();
		if (lwp < pDocument->i_WordCount)
			vecSentenceWordsConsumed [iSentence] += pDocument->sGetWord (lwp).length ();
	}


	// compute rewards ...						
	double dPositive = Learner::d_PositiveRewardMultiplier;
	double dNegative = Learner::d_NegativeRewardMultiplier;

	float fReward = NAN;
	float fPositiveReward = 0;
	int iSentence = -1;
	int iPreviousSentence = -1;
	EpisodeStep* pStep = NULL;
	for (size_t t = 0; t < dq_Steps.size (); ++ t)
	{
		pStep = dq_Steps [t];
		RLAction& rAction = pStep->o_RLAction;
		

		int lwu = rAction.i_SelectedObjectWordLocation;
		if (lwu == pDocument->i_WordCount)
		{
			assert (p_Learner->i_MaxWordlessActions > 0);

			pStep->o_RLState.SetReward (fReward);
			fReward = 0;
			continue;
		}


		if ((ft_no_words_objects != pStep->e_Feedback) &&
			(ft_deadend != pStep->e_Feedback) &&
			(ft_state_action_error != pStep->e_Feedback))
		{
			iSentence = pDocument->GetSentenceIndex (lwu);
		}

		if ((-1 != iPreviousSentence) && (iSentence != iPreviousSentence))
		{
			assert (false == isnan (fReward));
			fReward = fPositiveReward * dPositive;
		}
		iPreviousSentence = iSentence;



		int iWordsConsumed = vecSentenceWordsConsumed [iSentence];
		int iSentenceWords = vecSentenceWords [iSentence];
		int iSentenceLength = vecSentenceLength [iSentence];
		int iSentenceActions = vecSentenceActions [iSentence];

		float dWordRatio = iWordsConsumed / (float) iSentenceLength;
		float dActionRatio = (iSentenceWords - iSentenceActions) 
								/ (float) iSentenceWords;
		fPositiveReward = dWordRatio * dActionRatio;


		if (ft_none == pStep->e_Feedback)
		{
			// reward carried over from action that	
			// got us to this state...				
			if (false == isnan (fReward))
				pStep->o_RLState.SetReward (fReward);

			fReward = 0;
			// reward for ft_none needs to be		
			// assigned to the resulting state...	
		}

		else if (ft_no_words_objects == pStep->e_Feedback)
		{
			if (iSentence + 1 >= iSentenceCount)
				fReward = fPositiveReward * dPositive;
			else
				fReward = -1 * dNegative;

			pStep->o_RLState.SetReward (fReward);
			fReward = NAN;
		}

		else if (ft_repetition == pStep->e_Feedback)
		{
			// reward carried over from action that	
			// got us to this state...				
			if (false == isnan (fReward))
				pStep->o_RLState.SetReward (fReward);

			fReward = -0.1 * dNegative;
			if ((-1 != iPreviousSentence) && (iSentence != iPreviousSentence))
				fReward = fPositiveReward * dPositive;

			// reward for repetition needs to be	
			// assigned to the resulting state...	
		}

		else if (ft_env_unchanged == pStep->e_Feedback)
		{
			// reward carried over from action that	
			// got us to this state...				
			if (false == isnan (fReward))
				pStep->o_RLState.SetReward (fReward);

			fReward = -0.1 * dNegative;
			if ((-1 != iPreviousSentence) && (iSentence != iPreviousSentence))
				fReward = fPositiveReward * dPositive;

			// reward for env_unchanged needs to be	
			// assigned to the resulting state...	
		}

		else if (ft_deadend == pStep->e_Feedback)
		{
			fReward = -1 * dNegative;
			pStep->o_RLState.SetReward (fReward);
			fReward = NAN;
		}

		else if (ft_state_action_error == pStep->e_Feedback)
		{
			fReward = -1 * dNegative;
			pStep->o_RLState.SetReward (fReward);
			fReward = NAN;
		}

		else if (ft_comm_error == pStep->e_Feedback)
		{
			fReward = 0;
			pStep->o_RLState.SetReward (fReward);
			fReward = NAN;
		}

		else
			abort ();
	}
	assert (true == isnan (fReward));

	return true;
}



//															
bool Episode::ComputeRewardUsingAnnotation (Document* _pDocument)
{
	if (true == dq_Steps.empty ())
		return false;

	Annotation* pAnnotation = Annotation::Find (_pDocument->s_Id);
	if (NULL == pAnnotation)
	{
		cout << "[ERROR]  Annotation not found for document [id=" 
			 << _pDocument->s_Id << "]" << endl;
		assert (false);
	}

	AnnotatedSentence_dq_t::iterator	iteSentenceAnnotation;
	AnnotatedAction_dq_t::iterator	iteActionAnnotation;
	AnnotatedSentence* pSentenceAnnotation;

	iteSentenceAnnotation = pAnnotation->dq_Sentence.begin ();
	pSentenceAnnotation = *iteSentenceAnnotation;
	iteActionAnnotation = pSentenceAnnotation->dq_Action.begin ();

	EpisodeStep* pLastStep = NULL;
	bool bHasIncorrectAction = false;
	int iCorrectActions = 0;
	double dReward = 0;

	size_t t;
	for (t = 0; t < dq_Steps.size (); ++ t)
	{
		EpisodeStep* pStep = dq_Steps [t];
		pLastStep = pStep;
		RLAction& rAction = pStep->o_RLAction;
		AnnotatedAction* pActionAnnotation = *iteActionAnnotation;

		// The code below to set reward might look a little	
		// unusual, but the reward for an action needs to 	
		// be set to the resulting state, that's why the	
		// reward is being carried forward and set here.	
		pStep->o_RLState.SetReward (dReward);

		if (true == bHasIncorrectAction)
			break;

		bool bActionOk = false;
		if ((ft_state_action_error != pStep->e_Feedback) &&
			(ft_comm_error != pStep->e_Feedback) &&
			(ft_deadend != pStep->e_Feedback) &&
			(ft_no_words_objects != pStep->e_Feedback))
		{
			// compute value ...							
			int iWC = rAction.i_SelectedCommandWordLocation;
			int iWU = rAction.i_SelectedObjectWordLocation;
			int iWP = rAction.i_SelectedParameterWordLocation;
			int iC = rAction.i_SelectedCommandIndex;

			iWC = (iWC < pStep->o_RLState.TextLength ())? iWC : -1;
			iWU = (iWU < pStep->o_RLState.TextLength ())? iWU : -1;
			iWP = (iWP < pStep->o_RLState.TextLength ())? iWP : -1;

			String sObjectName = rAction.p_SelectedObject->s_CleanName;
			String sObjectClass = ObjectClass::Get (rAction.p_SelectedObject->i_ObjectClass);

			bool bCommandOk = (pActionAnnotation->s_Command == Command::Get (iC));
			bool bObjectOk = (pActionAnnotation->s_ObjectName == sObjectName) &&
							 (pActionAnnotation->s_ObjectType == sObjectClass);
			bool bCommandWordOk = (pActionAnnotation->i_CommandWord == iWC);
			bool bObjectWordOk = (pActionAnnotation->i_ObjectWord == iWU);
			bool bParamWordOk = (pActionAnnotation->i_Parameter == iWP);

			if (p_Learner->i_Verbose > 1)
			{
				cout << "      [cr]  " << pActionAnnotation->s_Command 
					 << (bCommandOk? " = " : " <> ") << Command::Get (iC) << endl;
				cout << "      [cr]  " << pActionAnnotation->s_ObjectName 
					 << (bObjectOk? " = " : " <> ") << sObjectName << endl;
				cout << "      [cr]  " << pActionAnnotation->s_ObjectType 
					 << (bObjectOk? " = " : " <> ") << sObjectClass << endl;
				cout << "      [cr]  " << pActionAnnotation->i_CommandWord 
					 << (bCommandWordOk? " = " : " <> ") << iWC;
				cout << "  " << pActionAnnotation->i_ObjectWord 
					 << (bObjectWordOk? " = " : " <> ") << iWU;
				cout << "  " << pActionAnnotation->i_Parameter 
					 << (bParamWordOk? " = " : " <> ") << iWP << endl;
			}


			bActionOk = (bCommandOk && bObjectOk && bCommandWordOk
							  && bObjectWordOk && bParamWordOk);
		}

		if (true == bActionOk)
		{
			++ iCorrectActions;
			dReward = Learner::d_PositiveRewardMultiplier;
		}
		else
		{
			bHasIncorrectAction = true;
			dReward = -1 * Learner::d_NegativeRewardMultiplier;
		}


		// increment annotation...						
		++ iteActionAnnotation;
		if (pSentenceAnnotation->dq_Action.end () == iteActionAnnotation)
		{
			++ iteSentenceAnnotation;
			if (pAnnotation->dq_Sentence.end () == iteSentenceAnnotation)
				break;

			else
			{
				pSentenceAnnotation = *iteSentenceAnnotation;
				iteActionAnnotation = pSentenceAnnotation->dq_Action.begin ();
			}
		}
	}


	// double dLastReward = pLastStep->o_RLState.GetReward ();

	t ++;
	for (; t < dq_Steps.size (); ++ t)
	{
		EpisodeStep* pStep = dq_Steps [t];
		assert (pLastStep != pStep);

		pStep->o_RLState.SetReward (0);
	}

	return true;
}



//															
void Episode::AccumulateTrace_WC_WU_WP (int _iTimesteps)
{
	for (int t = _iTimesteps - 1; t > 0; --t)
		dq_Steps [t - 1]->vec_Trace_WC_WU_WP  +=  dq_Steps [t]->vec_Trace_WC_WU_WP;
}


//															
void Episode::AccumulateTrace_U (int _iTimesteps)
{
	for (int t = _iTimesteps - 1; t > 0; --t)
		dq_Steps [t - 1]->vec_Trace_U  +=  dq_Steps [t]->vec_Trace_U;
}


//															
void Episode::AccumulateTrace_C (int _iTimesteps)
{
	for (int t = _iTimesteps - 1; t > 0; --t)
		dq_Steps [t - 1]->vec_Trace_C  +=  dq_Steps [t]->vec_Trace_C;
}


//															
void Episode::AccumulateTrace_A (int _iTimesteps)
{
	for (int t = _iTimesteps - 1; t > 0; --t)
		dq_Steps [t - 1]->vec_Trace_A  +=  dq_Steps [t]->vec_Trace_A;
}






//															
void Episode::ComputeValidActionWordRanges (WordSet_vec_t& _rvecWordSet)
{
	size_t iSteps = dq_Steps.size ();
	_rvecWordSet.clear ();
	_rvecWordSet.resize (iSteps);


	// collect all used word locations...		
	char_set_t setUsedWords;
	ITERATE (EpisodeStep_dq_t, dq_Steps, ite)
	{
		EpisodeStep* pStep = *ite;
		if (false == (*ite)->b_ActionExecutedWithoutError)
			break;

		RLAction& rAction = pStep->o_RLAction;
		RLState& rState = pStep->o_RLState;
		if (rAction.i_SelectedObjectWordLocation == rState.TextLength ())
			continue;

		setUsedWords.insert (rAction.i_SelectedCommandWordLocation);
		setUsedWords.insert (rAction.i_SelectedObjectWordLocation);
		if (rAction.i_SelectedParameterWordLocation < rState.TextLength ())
			setUsedWords.insert (rAction.i_SelectedParameterWordLocation);
	}


	// compute useable command word ranges...	
	for (size_t i = 0; i < iSteps; ++ i)
	{
		EpisodeStep* pStep = dq_Steps [i];
		if (false == pStep->b_ActionExecutedWithoutError)
			break;

		RLAction& rAction = pStep->o_RLAction;
		RLState& rState = pStep->o_RLState;
		if (rAction.i_SelectedObjectWordLocation == rState.TextLength ())
		{
			// if the action was a wordless one, the 'none'	
			// command word (the one that was originally	
			// selected) is the only available option.		
			_rvecWordSet [i].insert (rAction.i_SelectedCommandWordLocation);
			continue;
		}

		int iSentence = rState.GetSentenceIndex (rAction.i_SelectedObjectWordLocation);
		char iMin = rAction.i_SelectedObjectWordLocation;
		char iMax = rAction.i_SelectedObjectWordLocation;
		if (iMin > rAction.i_SelectedCommandWordLocation)
			iMin = rAction.i_SelectedCommandWordLocation;
		if (iMax < rAction.i_SelectedCommandWordLocation)
			iMax = rAction.i_SelectedCommandWordLocation;
		if (rAction.i_SelectedParameterWordLocation < rState.TextLength ())
		{
			if (iMin > rAction.i_SelectedParameterWordLocation)
				iMin = rAction.i_SelectedParameterWordLocation;
			if (iMax < rAction.i_SelectedParameterWordLocation)
				iMax = rAction.i_SelectedParameterWordLocation;
		}

		char_set_t::iterator	ite;
		char iRangeStart = iMin;
		ite = setUsedWords.find (iMin);
		if (setUsedWords.begin () != ite)
		{
			-- ite;
			iRangeStart = *ite + 1;
		}

		ite = setUsedWords.find (iMax);
		++ ite;
		char iRangeEnd = (setUsedWords.end () != ite)? *ite - 1 : iMax;

		for (char c = iRangeStart; c <= iRangeEnd; ++ c)
		{
			if (rState.GetSentenceIndex (c) != iSentence)
				continue;
			_rvecWordSet [i].insert (c);
		}

		_rvecWordSet [i].erase (rAction.i_SelectedObjectWordLocation);
		if (rAction.i_SelectedParameterWordLocation < rState.TextLength ())
			_rvecWordSet [i].erase (rAction.i_SelectedParameterWordLocation);
	}
}






//															
void Episode::ComputeWeightUpdates (WordSet_vec_t& _rvecWordSet)
{
	assert (false == b_WeightUpdatesComputed);

	size_t iRewards = 0;
	size_t iSteps = dq_Steps.size ();
	for (size_t i = 0; i < iSteps; ++ i)
	{
		EpisodeStep* pStep = dq_Steps [i];
		if (false == pStep->b_ActionExecutedWithoutError)
			break;
		pStep->UpdateTraces (_rvecWordSet [i]);
		++ iRewards;
	}
	if (0 == iRewards)
		return;

	// The rewards are assigned to the state 	
	// resulting from the action, but during	
	// the parameter update computation, should	
	// be used with the state from which the	
	// action was selected...					
	if (iRewards == dq_Steps.size ())
		-- iRewards;

	// collect rewards ...				
	double_vec_t vecReward;
	vecReward.Create (iRewards);
	vecReward.Memset (0);

	for (size_t i = 0; i < iRewards; ++ i)
		vecReward [i] = dq_Steps [i + 1]->o_RLState.GetReward ();


	// compute cumulative reward ...	
	for (size_t i = iRewards - 1; i > 0; -- i)
		vecReward [i - 1] += vecReward [i];

	cout << "   [cumulative reward] ";
	for (size_t i = 0; i < iRewards; ++ i)
		cout << vecReward [i] << ", ";
	cout << endl;

	// add the cumulative reward back	
	// to the environment state...		
	for (size_t i = 0; i < iRewards; ++ i)
		dq_Steps [i + 1]->o_RLState.AddToEnvironmentStateReward (vecReward [i]);

	// update parameters ...			
	ComputeWeightUpdate_WC_WU_WP (vecReward, p_Learner->vec_WeightUpdate_WC_WU_WP);
	ComputeWeightUpdate_U (vecReward, p_Learner->vec_WeightUpdate_U);
	ComputeWeightUpdate_C (vecReward, p_Learner->vec_WeightUpdate_C);
	ComputeWeightUpdate_A (vecReward, p_Learner->vec_WeightUpdate_A);

	b_WeightUpdatesComputed = true;
}



//															
void Episode::ComputeCumulativeRewardAndSetToEnvironment (void)
{
	size_t iRewards = 0;
	size_t iSteps = dq_Steps.size ();
	for (size_t i = 0; i < iSteps; ++ i)
	{
		EpisodeStep* pStep = dq_Steps [i];
		if (false == pStep->b_ActionExecutedWithoutError)
			break;
		++ iRewards;
	}
	if (0 == iRewards)
		return;

	// The rewards are assigned to the state 	
	// resulting from the action, but during	
	// the parameter update computation, should	
	// be used with the state from which the	
	// action was selected...					
	if (iRewards == dq_Steps.size ())
		-- iRewards;

	// collect rewards ...				
	double_vec_t vecReward;
	vecReward.Create (iRewards);
	vecReward.Memset (0);

	for (size_t i = 0; i < iRewards; ++ i)
		vecReward [i] = dq_Steps [i + 1]->o_RLState.GetReward ();


	// compute cumulative reward ...	
	for (size_t i = iRewards - 1; i > 0; -- i)
		vecReward [i - 1] += vecReward [i];

	cout << "   [cumulative reward] ";
	for (size_t i = 0; i < iRewards; ++ i)
		cout << vecReward [i] << ", ";
	cout << endl;

	// add the cumulative reward back	
	// to the environment state...		
	for (size_t i = 0; i < iRewards; ++ i)
		dq_Steps [i + 1]->o_RLState.AddToEnvironmentStateReward (vecReward [i]);
}



//															
void Episode::ComputeWeightUpdate_WC_WU_WP (double_vec_t& _vecReward, double_vec_t& _rvecWeightUpdate)
{
	int iTimesteps = _vecReward.Size ();
	size_t iFeatureCount = _rvecWeightUpdate.Size ();
	AccumulateTrace_WC_WU_WP (iTimesteps);

	// compute updates to individual feature weights	
	double_vec_t vecUpdates (iFeatureCount);
	vecUpdates.Memset (0);

	for (int t = 0; t < iTimesteps; ++t)
	{
		double dReward = _vecReward [t];
		double_vec_t& vecTrace = dq_Steps [t]->vec_Trace_WC_WU_WP;
		if (true == vecTrace.HasNan ())
		{
			cout << "[ERROR] Nan detected in vec_Trace_WC_WU_WP" << endl;
			abort ();
		}
		for (size_t f = 0; f < iFeatureCount; ++f)
			vecUpdates [f] += dReward * vecTrace [f];
	}

	// compute actual update weighted by learning rate & apply	
	for (size_t f = 0; f < iFeatureCount; ++f)
		_rvecWeightUpdate [f] += vecUpdates [f];

	if (true == _rvecWeightUpdate.HasNan ())
	{
		cout << "[ERROR] Nan detected in weight_update_wc_wu_wp" << endl;
		abort ();
	}
}



//															
void Episode::ComputeWeightUpdate_U (double_vec_t& _vecReward, double_vec_t& _rvecWeightUpdate)
{
	int iTimesteps = _vecReward.Size ();
	size_t iFeatureCount = _rvecWeightUpdate.Size ();
	AccumulateTrace_U (iTimesteps);

	// compute updates to individual feature weights	
	double_vec_t vecUpdates (iFeatureCount);
	vecUpdates.Memset (0);

	for (int t = 0; t < iTimesteps; ++t)
	{
		double dReward = _vecReward [t];
		double_vec_t& vecTrace = dq_Steps [t]->vec_Trace_U;
		for (size_t f = 0; f < iFeatureCount; ++f)
			vecUpdates [f] += dReward * vecTrace [f];
	}

	// compute actual update weighted by learning rate & apply	
	for (size_t f = 0; f < iFeatureCount; ++f)
		_rvecWeightUpdate [f] += vecUpdates [f];

	if (true == _rvecWeightUpdate.HasNan ())
	{
		cout << "[ERROR] Nan detected in weight_update_u" << endl;
		abort ();
	}
}



//															
void Episode::ComputeWeightUpdate_C (double_vec_t& _vecReward, double_vec_t& _rvecWeightUpdate)
{
	int iTimesteps = _vecReward.Size ();
	size_t iFeatureCount = _rvecWeightUpdate.Size ();
	AccumulateTrace_C (iTimesteps);

	// compute updates to individual feature weights	
	double_vec_t vecUpdates (iFeatureCount);
	vecUpdates.Memset (0);

	for (int t = 0; t < iTimesteps; ++t)
	{
		double dReward = _vecReward [t];
		double_vec_t& vecTrace = dq_Steps [t]->vec_Trace_C;
		for (size_t f = 0; f < iFeatureCount; ++f)
			vecUpdates [f] += dReward * vecTrace [f];
	}

	// compute actual update weighted by learning rate & apply	
	for (size_t f = 0; f < iFeatureCount; ++f)
		_rvecWeightUpdate [f] += vecUpdates [f];

	if (true == _rvecWeightUpdate.HasNan ())
	{
		cout << "[ERROR] Nan detected in weight_update_c" << endl;
		abort ();
	}
}



//															
void Episode::ComputeWeightUpdate_A (double_vec_t& _vecReward, double_vec_t& _rvecWeightUpdate)
{
	int iTimesteps = _vecReward.Size ();
	size_t iFeatureCount = _rvecWeightUpdate.Size ();
	AccumulateTrace_A (iTimesteps);

	// compute updates to individual feature weights	
	double_vec_t vecUpdates (iFeatureCount);
	vecUpdates.Memset (0);

	for (int t = 0; t < iTimesteps; ++t)
	{
		double dReward = _vecReward [t];
		double_vec_t& vecTrace = dq_Steps [t]->vec_Trace_A;
		for (size_t f = 0; f < iFeatureCount; ++f)
			vecUpdates [f] += dReward * vecTrace [f];
	}

	// compute actual update weighted by learning rate & apply	
	for (size_t f = 0; f < iFeatureCount; ++f)
		_rvecWeightUpdate [f] += vecUpdates [f];

	if (true == _rvecWeightUpdate.HasNan ())
		cout << "[ERROR] Nan detected in weight_update_a" << endl;
}




//															
std::ostream& operator<< (std::ostream& _rStream, const Episode& _rEpisode)
{
	ITERATE (EpisodeStep_dq_t, ((Episode&)_rEpisode).dq_Steps, ite)
	{
		EpisodeStep* pStep = *ite;
		if ((ft_no_words_objects == pStep->e_Feedback) ||
			(ft_deadend == pStep->e_Feedback) ||
			(ft_state_action_error == pStep->e_Feedback))
			continue;

		_rStream << *pStep << endl;
	}
	return _rStream;
}



//															
std::ostream& operator<< (std::ostream& _rStream, const Feedback_e _eFeedback)
{
	if (ft_unknown == _eFeedback)
		_rStream << "unknown";
	else if (ft_none == _eFeedback)
		_rStream << "none";
	else if (ft_no_words_objects == _eFeedback)
		_rStream << "no_words/objects";
	else if (ft_deadend == _eFeedback)
		_rStream << "deadend";
	else if (ft_env_unchanged == _eFeedback)
		_rStream << "nochange";
	else if (ft_repetition == _eFeedback)
		_rStream << "repeat";
	else if (ft_state_action_error == _eFeedback)
		_rStream << "sa_err";
	else if (ft_comm_error == _eFeedback)
		_rStream << "com_err";
	else
		assert (false);

	return _rStream;
}





