#include <math.h>
#include <nlp_macros.h>
#include "Learner.h"
#include "Sample.h"


double EpisodeStep::d_ExplorationEpsilon = 0.1;
double EpisodeStep::d_NegativeRewardAvoidanceProb = 0;
bool EpisodeStep::b_ExploreStochastically = false;
bool EpisodeStep::b_UseEpsilonGreedy = false;
bool EpisodeStep::b_UseSoftmax = false;
bool EpisodeStep::b_UseSmartSoftmax = false;
bool EpisodeStep::b_PrintSelectedValue = true;
bool EpisodeStep::b_AvoidNegativeRewardStates = false;


//														
EpisodeStep::EpisodeStep (Learner* _pLearner)
{
	pvec_Weights_WC_WU_WP = &_pLearner->vec_Weights_WC_WU_WP;
	pvec_Weights_U = &_pLearner->vec_Weights_U;
	pvec_Weights_C = &_pLearner->vec_Weights_C;
	pvec_Weights_A = &_pLearner->vec_Weights_A;

	e_Feedback = ft_unknown;
	d_SelectedActionLogProb = -100000000;

	p_ExperienceInterface = NULL;
	p_CurrentState = NULL;

	l_CandidateCount_WC_WU_WP = 0;
	l_CandidateCount_U = 0;
	l_CandidateCount_C = 0;

	i_FirstCandidateActionSentence = -1;
	b_ActionExecutedWithoutError = false;
	b_StartStatePrefix = false;
}


EpisodeStep::~EpisodeStep (void)
{
	assert ((double_vec_t*)0x1 != pvec_Weights_WC_WU_WP);
	pvec_Weights_WC_WU_WP = (double_vec_t*)0x1;
	
	pvec_Weights_U = NULL;
	pvec_Weights_C = NULL;
}


void EpisodeStep::UpdateTraces (char_set_t& _rsetUseableCommandWords)
{
	// compute normalization factor for alternate	
	// word selections ...							
	#ifndef NDEBUG
	int swc = o_RLAction.i_SelectedCommandWordLocation;
	assert (_rsetUseableCommandWords.end () != _rsetUseableCommandWords.find (swc));
	#endif
	int swu = o_RLAction.i_SelectedObjectWordLocation;
	int swp = o_RLAction.i_SelectedParameterWordLocation;
	long siu = o_RLAction.i_SelectedObjectIndex;
	int sic = o_RLAction.i_SelectedCommandIndex;


	double dMaxLogProb = NAN;
	ITERATE (char_set_t, _rsetUseableCommandWords, iteC)
	{
		double dLogProbWCWUWP = mtx_LogProb_WC_WU_WP (*iteC, swu, swp);
		if (true == isnan (dLogProbWCWUWP))
			continue;
		double dLogProbC = mtx_LogProb_C (*iteC, siu, sic);
		if (true == isnan (dLogProbC))
			continue;
		double dLogProb = dLogProbWCWUWP + dLogProbC;
		if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
			dMaxLogProb = dLogProb;
	}

	double dNormalization = 0;
	ITERATE (char_set_t, _rsetUseableCommandWords, iteC)
	{
		double dLogProbWCWUWP = mtx_LogProb_WC_WU_WP (*iteC, swu, swp);
		if (true == isnan (dLogProbWCWUWP))
			continue;
		double dLogProbC = mtx_LogProb_C (*iteC, siu, sic);
		if (true == isnan (dLogProbC))
			continue;
		double dLogProb = dLogProbWCWUWP + dLogProbC - dMaxLogProb;
		dNormalization += exp (dLogProb);
	}
	double dLogNormalization = log (dNormalization) + dMaxLogProb;


	// -- TEST CODE --------------------------------
	// _rsetUseableCommandWords.clear ();
	// _rsetUseableCommandWords.insert (swc);
	// dLogNormalization = 0;
	// -- TEST CODE --------------------------------


	// update traces ...							
	UpdateTrace_WC_WU_WP (_rsetUseableCommandWords, dLogNormalization);
	UpdateTrace_U ();
	UpdateTrace_C (_rsetUseableCommandWords, dLogNormalization);
	UpdateTrace_A ();

	mxf_Features_WC_WU_WP.Destroy ();
	mxf_Features_U.Destroy ();
	mxf_Features_C.Destroy ();
	mxf_Features_A.Destroy ();

	mtx_LogProb_WC_WU_WP.Destroy ();
	mtx_LogProb_U.Destroy ();
	mtx_LogProb_C.Destroy ();
	mtx_LogProb_A.Destroy ();
}




//														
EpisodeStepState_e EpisodeStep::ComputeProbs (bool _bFirstAction,
											  bool _bAllowWordlessAction)
{
	if (false == ComputeProb_WC_WU_WP (_bAllowWordlessAction))
		return ess_no_words_left;

	if (false == ComputeProb_U (_bFirstAction, _bAllowWordlessAction))
		return ess_no_object_targets_left;

	ComputeProb_C (_bAllowWordlessAction);
	ComputeProb_A (_bAllowWordlessAction);
	return ess_ok;
}



//														
EpisodeStepState_e EpisodeStep::WalkAction (AnnotatedAction* _pAnnotation)
{
	int swc = _pAnnotation->i_CommandWord;
	int swu = _pAnnotation->i_ObjectWord;
	int swp = _pAnnotation->i_Parameter;
	if (-1 == swp)
		swp = o_RLState.TextLength ();

	int sic = 0;
	while (sic < Command::i_Count)
	{
		if (Command::Get (sic) == _pAnnotation->s_Command)
			break;
		++ sic;
	}
	int siu = -1;
	int iMatchingObjects = 0;
	String sMultipleObjectMatches;
	for (int iu = 0; iu < o_RLState.EnvironmentObjectCount (); ++ iu)
	{
		UIObject* pObject = (UIObject*) o_RLState.GetObject (iu);
		if ((_pAnnotation->s_TemporallyUniqueObjectId
					== pObject->s_TemporallyUniqueId) && 
			(_pAnnotation->s_ObjectName == pObject->s_CleanName))
		{
			if (0 == iMatchingObjects)
				siu = iu;
			++ iMatchingObjects;
			sMultipleObjectMatches << pObject->s_TemporallyUniqueId << '\n';
		}
	}

	if (-1 == siu)
	{
		Learner::ofs_Log << "[ERROR] Failed to find matching object in environment" << endl
						 << "        Object : " << _pAnnotation->s_ObjectType 
						 << " : " << _pAnnotation->s_ObjectName << endl << endl
						 << " : " << _pAnnotation->s_TemporallyUniqueObjectId << endl << endl
						 << "-------------------------------------------------" << endl
						 << o_RLState.GetEnvironmentState () << endl
						 << "-------------------------------------------------" << endl;
		cout << "[ERROR] Failed to find matching object in environment" << endl
			 << "        Object : " << _pAnnotation->s_ObjectType 
			 << " : " << _pAnnotation->s_ObjectName << endl << endl
			 << " : " << _pAnnotation->s_TemporallyUniqueObjectId << endl << endl
			 << "-------------------------------------------------" << endl
			 << o_RLState.GetEnvironmentState () << endl
			 << "-------------------------------------------------" << endl;
		return ess_apply_action_failed;
	}

	UIObject* pObject = (UIObject*) o_RLState.GetObject (siu);
	if (false == RLState::WordMaybeObject (o_RLState.sGetWord (swu),
										   pObject->s_CleanName))
	{
		Learner::ofs_Log << "[WARNING] Object word does not overlap with Object's name" << '\n'
						 << "          object: " << pObject->s_CleanName << '\n'
						 << "            word: " << swu << "  " << o_RLState.sGetWord (swu)
						 << "             cmd: " << _pAnnotation->s_Command << "  " 
						 << pObject->s_CleanName << endl;
		cout << "[WARNING] Object word does not overlap with Object's name" << '\n'
			 << "          object: " << pObject->s_CleanName << '\n'
			 << "            word: " << swu << "  " << o_RLState.sGetWord (swu)
			 << "             cmd: " << _pAnnotation->s_Command << "  " 
			 << pObject->s_CleanName << endl;
	}
	

	//												
	size_t iMaxRequirement = l_CandidateCount_WC_WU_WP
							 * l_CandidateCount_U
							 * l_CandidateCount_C;
	Vector <double> vecProb (iMaxRequirement);
	Vector <long> vecIndex (iMaxRequirement);

	double dMinProb = ComputeJointProbs (vecProb, vecIndex, false);
	if (true == isnan (dMinProb))
	{
		Learner::ofs_Log << "[ERROR] Annotated action not possible under current values! All actions are invalid\n";
		cerr << "[ERROR] Annotated action not possible under current values!\n";
		return ess_no_words_left;
	}


	long lActionInfo = RLAction::ActionToLong (swu, swc, swp, sic, siu);
	size_t iSample;
	bool bActionFound = false;
	for (iSample = 0; iSample < vecIndex.Size (); ++ iSample)
	{
		if (lActionInfo == vecIndex [iSample])
		{
			bActionFound = true;
			break;
		}
	}

	if (false == bActionFound)
	{
		SetSelectedAction (swc, swu, swp, sic, siu);

		Learner::ofs_Log << "[ERROR] Annotated action not possible under current values! Annotated action is invalid\n";
		Learner::ofs_Log << *this << endl;
		cerr << "[ERROR] Annotated action not possible under current values!\n";
		cerr << *this << endl;
		return ess_no_words_left;
	}

	d_SelectedActionLogProb = vecProb [iSample];
	assert (false == isnan (d_SelectedActionLogProb));


	//												
	SetSelectedAction (swc, swu, swp, sic, siu);

	return ess_ok;
}



//														
EpisodeStepState_e EpisodeStep::SelectAction (TrainTest_e _eTrainTestMode, double _dExplorationRate)
{
	if (false == Sample_WC_WU_WP_U_C (_eTrainTestMode, _dExplorationRate))
		return ess_all_actions_zero_prob;


	//												
	if (o_RLAction.i_SelectedCommandWordLocation == o_RLState.TextLength ())
	{
		assert (o_RLAction.i_SelectedObjectWordLocation == o_RLState.TextLength ());
		assert (o_RLAction.i_SelectedParameterWordLocation == o_RLState.TextLength ());
		return ess_action_without_words;
	}

	return ess_ok;
}



//														
double EpisodeStep::ComputeJointProbs (Vector <double>& _rvecLogProb,
									   Vector <long>& _rvecIndex,
									   bool _bAvoidNegativeRewardStates)
{
	int iRangeWC = o_RLState.TextLength () + 1;
	int iRangeWU = o_RLState.TextLength () + 1;
	int iRangeWP = o_RLState.TextLength () + 1;

	double dMinLogProb = NAN;
	double dMaxLogProb = NAN;

	size_t i = 0;
	for (int lwc = 0; lwc < iRangeWC; ++ lwc)
	{
		bool bNoneWC = (o_RLState.TextLength () == lwc);

		for (int lwu = 0; lwu < iRangeWU; ++ lwu)
		{
			bool bNoneWU = (o_RLState.TextLength () == lwu);

			if (((true == bNoneWC) && (false == bNoneWU)) ||
				((false == bNoneWC) && (true == bNoneWU)))
				continue;
			if ((lwc == lwu) && (false == bNoneWU))
				continue;

			for (int lwp = 0; lwp < iRangeWP; ++ lwp)
			{
				if ((lwp == lwc) && (false == bNoneWU))
					continue;
				if ((lwp == lwu) && (false == bNoneWU))
					continue;
				bool bNoneWP = (o_RLState.TextLength () == lwp);
				bool bHasParam = !bNoneWP;

				if ((true == bNoneWC) && (false == bNoneWP))
					continue;

				double dLogProb_WC_WU_WP = mtx_LogProb_WC_WU_WP (lwc, lwu, lwp);
				if (true == isnan (dLogProb_WC_WU_WP))
					continue;

				for (int iu = 0; iu < o_RLState.EnvironmentObjectCount (); ++ iu)
				{
					double dLogProb_U = mtx_LogProb_U (lwu, iu);
					if (true == isnan (dLogProb_U))
						continue;

					for (int ic = 0; ic < Command::i_Count; ++ ic)
					{
						double dLogProb_C = NAN;
						bool bNeedsParam = Command::NeedsParam (ic);
						if (((true == bNeedsParam) && (true == bHasParam)) ||
							 ((false == bNeedsParam) && (false == bHasParam)))
							dLogProb_C = mtx_LogProb_C (lwc, iu, ic);

						if (true == isnan (dLogProb_C))
							continue;

						double dLogProb_A = mtx_LogProb_A (lwp, iu, ic);
						assert (false == isnan (dLogProb_A));
						double dLogProb = dLogProb_WC_WU_WP + dLogProb_U + dLogProb_C + dLogProb_A;


						// translate the indices to a compound long ...		
						long lActionIndex = RLAction::ActionToLong (lwu, lwc, lwp, ic, iu);

						if (true == _bAvoidNegativeRewardStates)
						{
							EnhancedState* pNextState = o_RLState.p_EnhancedState->GetNextState (lActionIndex);
							if (NULL != pNextState)
							{
								if (pNextState->GetReward () == -1 * Learner::d_NegativeRewardMultiplier)
									continue;
							}
						}

						// remember minimum value for later use...			
						if ((true == isnan (dMinLogProb)) || (dMinLogProb > dLogProb))
							dMinLogProb = dLogProb;
						if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
							dMaxLogProb = dLogProb;

						// set values ...									
						_rvecLogProb [i] = dLogProb;
						_rvecIndex [i] = lActionIndex;
						++ i;
					}
				}
			}
		}
	}

	_rvecLogProb.SetSize (i);
	_rvecIndex.SetSize (i);

	// rescale _rvecLogProb to ensure it does not	
	// overflow when exponentiated...				
	for (size_t j = 0; j < i; ++ j)
		_rvecLogProb [j] -= dMaxLogProb;
	dMinLogProb -= dMaxLogProb;

	
	if (0 == i)
	{
		// cout << "[ERROR] pdf is all zeros" << endl;
		return NAN;
	}
	return dMinLogProb;
}


//														
bool EpisodeStep::Sample_WC_WU_WP_U_C (TrainTest_e _eTrainTestMode, 
									   double _dExplorationRate)
{
	size_t iMaxRequirement = l_CandidateCount_WC_WU_WP
							 * l_CandidateCount_U
							 * l_CandidateCount_C;
	Vector <double> vecLogProb (iMaxRequirement);
	Vector <long> vecIndex (iMaxRequirement);

	bool bAvoidNegativeRewardStates = false;
	if (tt_test_mode == _eTrainTestMode)
		bAvoidNegativeRewardStates = true;
	else if (true == b_AvoidNegativeRewardStates)
	{
		bAvoidNegativeRewardStates 
			= Sample::SampleCoinflip (EpisodeStep::d_NegativeRewardAvoidanceProb);
	}

	double dMinLogProb = ComputeJointProbs (vecLogProb, 
											vecIndex, 
											bAvoidNegativeRewardStates);
	// dMinLogProb is NAN only if there are no	
	// valid actions in vecLogProb...			
	if (true == isnan (dMinLogProb))
	{
		Learner::ofs_Log << "[ERROR] No valid actions present" << endl;
		return false;
	}


	size_t iSample = -1;
	// test mode...								
	if (tt_test_mode == _eTrainTestMode)
		iSample = Sample::SelectMax (vecLogProb);


	// act greedily...							
	else if (false == EpisodeStep::b_ExploreStochastically)
		iSample = Sample::SelectMax (vecLogProb);
	

	// explore using epsilon greedy...			
	else if (true == EpisodeStep::b_UseEpsilonGreedy)
	{
		if (false == Sample::SampleCoinflip (EpisodeStep::d_ExplorationEpsilon))
			iSample = Sample::SelectMax (vecLogProb);

		else
		{
			if (true == EpisodeStep::b_UseSoftmax)
			{
				Vector <double> vecSoftmax (vecLogProb.Size ());
				for (size_t i = 0; i < vecLogProb.Size (); ++ i)
				{
					double dSoftmax = exp (_dExplorationRate * vecLogProb [i]);
					vecSoftmax [i] = dSoftmax;
				}

				// SampleFromPDF doesn't need to have the pdf normalized.		
				iSample = Sample::SampleFromPDF (vecSoftmax);
			}
			else if (true == EpisodeStep::b_UseSmartSoftmax)
			{
				double dMaxProb = NAN;
				double dMinProb = NAN;
				Vector <double> vecSoftmax (vecLogProb.Size ());
				for (size_t i = 0; i < vecLogProb.Size (); ++ i)
				{
					double dSoftmax = exp (_dExplorationRate * vecLogProb [i]);
					vecSoftmax [i] = dSoftmax;
					if ((true == isnan (dMaxProb)) || (dMaxProb < dSoftmax))
						dMaxProb = dSoftmax;
					if ((true == isnan (dMinProb)) || (dMinProb > dSoftmax))
						dMinProb = dSoftmax;
				}
				cout << "      [**] smart softmax ratio : " << dMaxProb / dMinProb << endl;
				if (1000 * dMinProb < dMaxProb)
				{
					for (size_t i = 0; i < vecLogProb.Size (); ++ i)
						vecSoftmax [i] += dMaxProb / 1000;
				}

				// SampleFromPDF doesn't need to have the pdf normalized.		
				iSample = Sample::SampleFromPDF (vecSoftmax);
			}
			else
			{
				Vector <double> vecUniform (vecLogProb.Size ());
				vecUniform.Initialize (1);
				iSample = Sample::SampleFromPDF (vecUniform);
			}
		}
	}

	// explore using softmax...					
	else if (true == EpisodeStep::b_UseSoftmax)
	{
		Vector <double> vecSoftmax (vecLogProb.Size ());
		for (size_t i = 0; i < vecLogProb.Size (); ++ i)
		{
			double dSoftmax = exp (_dExplorationRate * vecLogProb [i]);
			vecSoftmax [i] = dSoftmax;
		}

		// SampleFromPDF doesn't need to have the pdf normalized.		
		iSample = Sample::SampleFromPDF (vecSoftmax);
	}

	else
		assert (false);

	d_SelectedActionLogProb = vecLogProb [iSample];


	// Get the action corresponding to the value we sampled above...	
	ActionInfo_u uAction;
	uAction.l_ActionInfo = vecIndex [iSample];

	#ifndef NDEBUG
	int iRangeWC = o_RLState.TextLength () + 1;
	int iRangeWU = o_RLState.TextLength () + 1;
	int iRangeWP = o_RLState.TextLength () + 1;
	#endif
	assert (iSample < vecLogProb.Size ());
	assert (uAction.o_ActionInfo.i_SelectedCommandWordLocation < iRangeWC);
	assert (uAction.o_ActionInfo.i_SelectedObjectWordLocation < iRangeWU);
	assert (uAction.o_ActionInfo.i_SelectedParameterWordLocation < iRangeWP);
	assert (uAction.o_ActionInfo.i_SelectedObjectIndex < o_RLState.EnvironmentObjectCount ());
	assert (uAction.o_ActionInfo.i_SelectedCommandIndex < Command::i_Count);


	SetSelectedAction (uAction.o_ActionInfo.i_SelectedCommandWordLocation,
					   uAction.o_ActionInfo.i_SelectedObjectWordLocation,
					   uAction.o_ActionInfo.i_SelectedParameterWordLocation,
					   uAction.o_ActionInfo.i_SelectedCommandIndex,
					   uAction.o_ActionInfo.i_SelectedObjectIndex);

	return true;
}



//														
void EpisodeStep::SetSelectedAction (int _iLocationWC, 
									int _iLocationWU,
									int _iLocationWP,
									int _iCommand,
									int _iObject)
{
	o_RLAction.i_SelectedCommandWordLocation = _iLocationWC;
	o_RLAction.i_SelectedObjectWordLocation = _iLocationWU;
	o_RLAction.i_SelectedParameterWordLocation = _iLocationWP;

	o_RLAction.i_SelectedCommandIndex = _iCommand;
	o_RLAction.i_SelectedObjectIndex = _iObject;
	if (_iLocationWP != o_RLState.TextLength ())
		o_RLAction.s_SelectedParameter = o_RLState.sGetWord (_iLocationWP);
	else
		o_RLAction.s_SelectedParameter = "";

	o_RLAction.p_SelectedObject = (UIObject*) o_RLState.GetObject (o_RLAction.i_SelectedObjectIndex);
}



//														
std::ostream& operator<< (std::ostream& _rStream, const EpisodeStep& _rStep)
{
	RLAction& _rAction = ((EpisodeStep&)_rStep).o_RLAction;
	RLState& _rState = ((EpisodeStep&)_rStep).o_RLState;

	String sCommandType = Command::Get (_rAction.i_SelectedCommandIndex);
	if ("[none]" == sCommandType)
	{
		_rStream << "         [null action]" << endl;
		return _rStream;
	}

	// sentence ...
	_rStream << "         [sentence] \x01";
	if (_rAction.i_SelectedObjectWordLocation < _rState.TextLength ())
		_rStream << _rStep.o_RLState.GetSentenceIndex (_rAction.i_SelectedObjectWordLocation)
				 << endl;
	else
		_rStream << "(none)" << endl;


	// words selected ...
	_rStream << "         [textaction] \x01";

	if (_rAction.i_SelectedCommandWordLocation < _rState.TextLength ())
		_rStream << "wc|" << (int) _rAction.i_SelectedCommandWordLocation << '|' 
				 << _rState.sGetWord (_rAction.i_SelectedCommandWordLocation) << " \x01 ";
	else
		_rStream << "wc|-1|(none) \x01 ";

	if (_rAction.i_SelectedObjectWordLocation < _rState.TextLength ())
		_rStream << "wu|" << (int) _rAction.i_SelectedObjectWordLocation << '|' 
				 << _rState.sGetWord (_rAction.i_SelectedObjectWordLocation) << " \x01 ";
	else
		_rStream << "wu|-1|(none) \x01 ";

	if (_rAction.i_SelectedParameterWordLocation < _rState.TextLength ())
		_rStream << "wp|" << (int) _rAction.i_SelectedParameterWordLocation << '|' 
				 << _rState.sGetWord (_rAction.i_SelectedParameterWordLocation);
	else
		_rStream << "wp|-1|(none)";
	_rStream << endl;

	String sSelectedCommand = Command::Get (_rAction.i_SelectedCommandIndex);	
	// command & object ...
	_rStream << "         [worldaction] \x01";
	_rStream << "c|" << (int) _rAction.i_SelectedCommandIndex << '|' 
			 << sSelectedCommand << " \x01 ";
	_rStream << "u|" << _rAction.p_SelectedObject->z_ID << '|' 
			 << _rAction.p_SelectedObject->s_Name << '|'
			 << ObjectClass::Get (_rAction.p_SelectedObject->i_ObjectClass) << " \x01 ";
	_rStream << _rAction.p_SelectedObject->s_TemporallyUniqueId << " \x01";
	_rStream << endl;

	// actual command ...
	String sCommand;
	sCommand << '!' << _rAction.p_SelectedObject->z_ID << ':' << sSelectedCommand;
	if (_rAction.i_SelectedParameterWordLocation < _rState.TextLength ())
		sCommand << '|' << _rState.sGetWord (_rAction.i_SelectedParameterWordLocation);

	_rStream << "         [command] \x01" << sCommand << endl;

	return _rStream;
}


