#include <math.h>
#include <nlp_macros.h>
#include <nlp_time.h>
#include "Learner.h"
#include "FieldId.h"
#include "Vocabulary.h"

#define COUT cout


bool EpisodeStep::b_LookAtNextSentenceAlso = false;


//														
bool EpisodeStep::ComputeProb_WC_WU_WP (bool _bAllowWordlessAction)
{
	// p(wc,wu|.)
	i_ActionsStartAtSentence = i_FirstCandidateActionSentence;

	if (0 == o_RLState.UnusedWordCount ())
		return false;

	mxf_Features_WC_WU_WP.Create (o_RLState.TextLength (), 
								  o_RLState.TextLength (),
								  o_RLState.TextLength () + 1,
								  Learner::FeatureCount_WC_WU_WP ());
	mtx_LogProb_WC_WU_WP.Create (o_RLState.TextLength () + 1,
								 o_RLState.TextLength () + 1,
								 o_RLState.TextLength () + 1);
	mtx_LogProb_WC_WU_WP.Initialize (NAN);

	//											
	Vector <float> vecFeatures;
	double dMaxLogProb = NAN;
	l_CandidateCount_WC_WU_WP = 0;

	int iCommaIndex = Vocabulary::Find (",");
	assert (-1 != iCommaIndex);



	//											
	for (int c = 0; c < o_RLState.UnusedWordCount (); ++ c)
	{
		char lwc = o_RLState.UnusedWordLocation (c);
		if (iCommaIndex == o_RLState.iGetWord (lwc))
			continue;

		int iSentenceWC = o_RLState.GetSentenceIndex (lwc);
		if (iSentenceWC < i_FirstCandidateActionSentence)
			continue;

		if (iSentenceWC > i_FirstCandidateActionSentence + 1)
			break;

		if ((false == EpisodeStep::b_LookAtNextSentenceAlso) &&
			(-1 != i_FirstCandidateActionSentence) && 
			(iSentenceWC > i_FirstCandidateActionSentence))
			break;

		if (iSentenceWC < o_RLState.GetSentenceIndex (o_RLState.PreviousLocation_WC ()))
			continue;

		for (int u = 0; u < o_RLState.i_UnusedPotentialObjectWordCount; ++ u)
		{
			char lwu = o_RLState.p_UnusedPotentialObjectWords [u];
			if (lwc == lwu)
				continue;

			int iSentenceWU = o_RLState.GetSentenceIndex (lwu);
			if (iSentenceWU > iSentenceWC)
				break;
			if (iSentenceWU != iSentenceWC)
				continue;

			if (iCommaIndex == o_RLState.iGetWord (lwu))
				continue;

			for (int p = 0; p < o_RLState.UnusedWordCount (); ++ p)
			{
				char lwp = o_RLState.UnusedWordLocation (p);
				if ((lwc == lwp) || (lwu == lwp))
					continue;

				int iSentenceWP = o_RLState.GetSentenceIndex (lwp);
				if (iSentenceWP > iSentenceWC)
					break;
				if (false == o_RLState.WordLocationsAreValid (lwc, lwu, lwp))
					continue;

				if (iCommaIndex == o_RLState.iGetWord (lwp))
					continue;


				++ l_CandidateCount_WC_WU_WP;
				mxf_Features_WC_WU_WP.GetSlice (vecFeatures, lwc, lwu, lwp);
				vecFeatures.Memset (0);
				double dLogProb = ComputeScore_WC_WU_WP (lwc, lwu, lwp, vecFeatures);
				mtx_LogProb_WC_WU_WP (lwc, lwu, lwp) = dLogProb;

				if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
					dMaxLogProb = dLogProb;
			}


			++ l_CandidateCount_WC_WU_WP;
			// this is the special case of no-param		
			int lwp = o_RLState.TextLength ();
			if (false == o_RLState.WordLocationsAreValid (lwc, lwu, lwp))
				continue;

			mxf_Features_WC_WU_WP.GetSlice (vecFeatures, lwc, lwu, lwp);
			vecFeatures.Memset (0);
			double dLogProb = ComputeScore_WC_WU_WP (lwc, lwu, lwp, vecFeatures);
			mtx_LogProb_WC_WU_WP (lwc, lwu, lwp) = dLogProb;

			if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
				dMaxLogProb = dLogProb;
		}
	}


	// action without object word ...					
	if (true == _bAllowWordlessAction)
	{
		int lwc = o_RLState.TextLength ();
		int lwu = o_RLState.TextLength ();
		int lwp = o_RLState.TextLength ();

		++ l_CandidateCount_WC_WU_WP;
		mxf_Features_WC_WU_WP.GetSlice (vecFeatures, lwc, lwu, lwp);
		vecFeatures.Memset (0);
		double dLogProb = ComputeScore_WC_WU_WP (lwc, lwu, lwp, vecFeatures);
		mtx_LogProb_WC_WU_WP (lwc, lwu, lwp) = dLogProb;

		if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
			dMaxLogProb = dLogProb;
	}



	// before normalization, we need to re-scale 	
	// the log probs to avoid over/underflow		
	double dNormalization = 0;
	for (int c = 0; c < o_RLState.UnusedWordCount (); ++ c)
	{
		char lwc = o_RLState.UnusedWordLocation (c);
		int iSentenceWC = o_RLState.GetSentenceIndex (lwc);

		for (int u = 0; u < o_RLState.i_UnusedPotentialObjectWordCount; ++ u)
		{
			char lwu = o_RLState.p_UnusedPotentialObjectWords [u];
			if (lwc == lwu)
				continue;
			int iSentenceWU = o_RLState.GetSentenceIndex (lwu);
			if (iSentenceWU > iSentenceWC)
				break;
			if (iSentenceWU != iSentenceWC)
				continue;

			for (int p = 0; p < o_RLState.UnusedWordCount (); ++ p)
			{
				char lwp = o_RLState.UnusedWordLocation (p);
				double dLogProb = mtx_LogProb_WC_WU_WP (lwc, lwu, lwp);
				if (false == isnan (dLogProb))
				{
					dLogProb -= dMaxLogProb;
					mtx_LogProb_WC_WU_WP (lwc, lwu, lwp) = dLogProb;
					dNormalization += exp (dLogProb);
				}
			}

			// this is the special case of no-param		
			char lwp = o_RLState.TextLength ();
			double dLogProb = mtx_LogProb_WC_WU_WP (lwc, lwu, lwp);
			if (false == isnan (dLogProb))
			{
				dLogProb -= dMaxLogProb;
				mtx_LogProb_WC_WU_WP (lwc, lwu, lwp) = dLogProb;
				dNormalization += exp (dLogProb);
			}
		}
	}

	if (true == _bAllowWordlessAction)
	{
		int lwc = o_RLState.TextLength ();
		int lwu = o_RLState.TextLength ();
		int lwp = o_RLState.TextLength ();

		double dLogProb = mtx_LogProb_WC_WU_WP (lwc, lwu, lwp);
		dLogProb -= dMaxLogProb;
		mtx_LogProb_WC_WU_WP (lwc, lwu, lwp) = dLogProb;
		dNormalization += exp (dLogProb);
	}



	// normalize probabilities...				
	double dLogNormalization = log (dNormalization);

	for (int c = 0; c < o_RLState.UnusedWordCount (); ++ c)
	{
		char lwc = o_RLState.UnusedWordLocation (c);
		int iSentenceWC = o_RLState.GetSentenceIndex (lwc);

		for (int u = 0; u < o_RLState.i_UnusedPotentialObjectWordCount; ++ u)
		{
			char lwu = o_RLState.p_UnusedPotentialObjectWords [u];
			if (lwc == lwu)
				continue;
			int iSentenceWU = o_RLState.GetSentenceIndex (lwu);
			if (iSentenceWU > iSentenceWC)
				break;
			if (iSentenceWU != iSentenceWC)
				continue;

			for (int p = 0; p < o_RLState.UnusedWordCount (); ++ p)
			{
				char lwp = o_RLState.UnusedWordLocation (p);
				if (false == isnan (mtx_LogProb_WC_WU_WP (lwc, lwu, lwp)))
					mtx_LogProb_WC_WU_WP (lwc, lwu, lwp) -= dLogNormalization;
			}

			// this is the special case of no-param		
			int lwp = o_RLState.TextLength ();
			if (false == isnan (mtx_LogProb_WC_WU_WP (lwc, lwu, lwp)))
				mtx_LogProb_WC_WU_WP (lwc, lwu, lwp) -= dLogNormalization;
		}
	}

	if (true == _bAllowWordlessAction)
	{
		int lwc = o_RLState.TextLength ();
		int lwu = o_RLState.TextLength ();
		int lwp = o_RLState.TextLength ();

		mtx_LogProb_WC_WU_WP (lwc, lwu, lwp) -= dLogNormalization;
	}



	//											
	if (0 == o_RLState.UnusedWordCount ())
		cout << "   [ERROR] unused word set is empty" << endl;
	if (0 == l_CandidateCount_WC_WU_WP)
		return false;

	return true;
}



//														
bool EpisodeStep::ComputeProb_U (bool _bFirstAction, bool _bAllowWordlessAction)
{
	// p(u|wu, .)
	if ((false == _bAllowWordlessAction) &&
		(0 == o_RLState.i_UnusedPotentialObjectWordCount))
		return false;


	mxf_Features_U.Create (o_RLState.TextLength () + 1,
						   o_RLState.EnvironmentObjectCount (),
						   Learner::FeatureCount_U ());
	mtx_LogProb_U.Create (o_RLState.TextLength () + 1,
						   o_RLState.EnvironmentObjectCount ());
	mtx_LogProb_U.Initialize (NAN);


	Vector <float>		vecFeatures;
	Matrix <double, 1>	mtxLogProb;
	set_PotentiallyInteractableObjects.clear ();
	l_CandidateCount_U = 0;


	for (int u = 0; u < o_RLState.i_UnusedPotentialObjectWordCount; ++ u)
	{
		char lwu = o_RLState.p_UnusedPotentialObjectWords [u];
		mtxLogProb.GetSlice (mtx_LogProb_U, lwu);

		double dMaxLogProb = NAN;
		UIObject_dq_t& rdqObjects = o_RLState.vec_WordObjects [lwu];
		ITERATE (UIObject_dq_t, rdqObjects, ite)
		{
			UIObject* pObject = *ite;
			assert (0 != pObject->bit_Features [VISIBLE]);

			mxf_Features_U.GetSlice (vecFeatures, lwu, pObject->i_Index);
			vecFeatures.Memset (0);
			double dLogProb = ComputeScore_U (lwu, pObject, _bFirstAction, vecFeatures);
			mtxLogProb (pObject->i_Index) = dLogProb;
			if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
				dMaxLogProb = dLogProb;

			set_PotentiallyInteractableObjects.insert (*ite);

			++ l_CandidateCount_U;
		}


		// scale log probs to avoid over/underflow	
		double dNormalization = 0;
		ITERATE (UIObject_dq_t, rdqObjects, ite)
		{
			UIObject* pObject = *ite;
			assert (0 != pObject->bit_Features [VISIBLE]);

			double dLogProb = mtxLogProb (pObject->i_Index);
			dLogProb -= dMaxLogProb;
			mtxLogProb (pObject->i_Index) = dLogProb;
			dNormalization += exp (dLogProb);
		}


		// normalization ...			
		if (0 != dNormalization)
		{
			double dLogNormalization = log (dNormalization);
			ITERATE (UIObject_dq_t, rdqObjects, ite)
			{
				UIObject* pObject = *ite;
				assert (0 != pObject->bit_Features [VISIBLE]);
				mtxLogProb (pObject->i_Index) -= dLogNormalization;
			}
		}
		else
			cout << "[WARNING]  zero pdf in ComputeProb_U" << endl;
	}

	// action without object word ...					
	if (true == _bAllowWordlessAction)
	{
		char lwu = o_RLState.TextLength ();
		mtxLogProb.GetSlice (mtx_LogProb_U, lwu);

		double dMaxLogProb = NAN;
		for (long u = 0; u < o_RLState.EnvironmentObjectCount (); ++ u)
		{
			UIObject* pObject = o_RLState.GetObject (u);
			if (0 == pObject->bit_Features [VISIBLE])
				continue;
			if ("" == pObject->s_CleanName)
				continue;

			mxf_Features_U.GetSlice (vecFeatures, lwu, pObject->i_Index);
			vecFeatures.Memset (0);
			double dLogProb = ComputeScore_U (lwu, pObject, _bFirstAction, vecFeatures);
			mtxLogProb (pObject->i_Index) = dLogProb;
			if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
				dMaxLogProb = dLogProb;

			++ l_CandidateCount_U;
		}


		// scale log probs to avoid over/underflow	
		double dNormalization = 0;
		for (long u = 0; u < o_RLState.EnvironmentObjectCount (); ++ u)
		{
			UIObject* pObject = o_RLState.GetObject (u);
			if (0 == pObject->bit_Features [VISIBLE])
				continue;
			if ("" == pObject->s_CleanName)
				continue;

			double dLogProb = mtxLogProb (pObject->i_Index);
			dLogProb -= dMaxLogProb;
			mtxLogProb (pObject->i_Index) = dLogProb;
			dNormalization += exp (dLogProb);
		}


		// normalization ...			
		if (0 != dNormalization)
		{
			double dLogNormalization = log (dNormalization);
			for (long u = 0; u < o_RLState.EnvironmentObjectCount (); ++ u)
			{
				UIObject* pObject = o_RLState.GetObject (u);
				if (0 == pObject->bit_Features [VISIBLE])
					continue;
				if ("" == pObject->s_CleanName)
					continue;
				mtxLogProb (pObject->i_Index) -= dLogNormalization;
			}
		}
		else
			cout << "[WARNING]  zero pdf in ComputeProb_U" << endl;
	}



	// the candidate count is only used to compute the	
	// max amount of memory required for the compound	
	// prob vector.  The wc,wu,wp counts are already	
	// accounted for by the other candidate counts, so	
	// this count here only has to count the objects...	
	if (true == _bAllowWordlessAction)
	{
		if (l_CandidateCount_U > o_RLState.EnvironmentObjectCount ())
			l_CandidateCount_U = o_RLState.EnvironmentObjectCount ();
	}
	else
	{
		if (l_CandidateCount_U > (long)set_PotentiallyInteractableObjects.size ())
			l_CandidateCount_U = (long)set_PotentiallyInteractableObjects.size ();
	}

	if (0 == l_CandidateCount_U)
		cout << "   [ERROR] u candidates = 0" << endl;

	return true;
}



//														
bool EpisodeStep::ComputeProb_C (bool _bAllowWordlessAction)
{
	// p(c|wc, u, .) 
	mxf_Features_C.Create (o_RLState.TextLength () + 1,
						   o_RLState.EnvironmentObjectCount (),
						   Command::i_Count);
	mtx_LogProb_C.Create (o_RLState.TextLength () + 1,
						   o_RLState.EnvironmentObjectCount (),
						   Command::i_Count);
	mtx_LogProb_C.Initialize (NAN);


	BitVector <FEATURES_C>*	pvecFeatures;
	Matrix <double, 1>		mtxLogProb;


	for (int c = 0; c < o_RLState.UnusedWordCount (); ++ c)
	{
		char lwc = o_RLState.UnusedWordLocation (c);
		ITERATE (UIObject_set_t, set_PotentiallyInteractableObjects, ite)
		{
			UIObject* pObject = *ite;
			assert (0 != pObject->bit_Features [VISIBLE]);
			mtxLogProb.GetSlice (mtx_LogProb_C, lwc, pObject->i_Index);

			double dMaxLogProb = NAN;
			for (int ic = 0; ic < Command::i_Count; ++ ic)
			{
				mxf_Features_C.GetSlice (&pvecFeatures, lwc, pObject->i_Index, ic);
				pvecFeatures->Memset (0);
				double dLogProb = ComputeScore_C (lwc, pObject, ic, *pvecFeatures);
				mtxLogProb (ic) = dLogProb;
				if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
					dMaxLogProb = dLogProb;
			}


			// scale log probs to avoid over/underflow	
			double dNormalization = 0;
			for (int ic = 0; ic < Command::i_Count; ++ ic)
			{
				double dLogProb = mtxLogProb (ic);
				dLogProb -= dMaxLogProb;
				mtxLogProb (ic) = dLogProb;
				dNormalization += exp (dLogProb);
			}


			// normalization...				
			if (0 != dNormalization)
			{
				double dLogNormalization = log (dNormalization);
				for (int ic = 0; ic < Command::i_Count; ++ ic)
					mtxLogProb (ic) -= dLogNormalization;
			}
			else
				cout << "[WARNING]  zero pdf in ComputeProb_C" << endl;
		}
	}

	// action without consuming words ...					
	if (true == _bAllowWordlessAction)
	{
		char lwc = o_RLState.TextLength ();
		for (long u = 0; u < o_RLState.EnvironmentObjectCount (); ++ u)
		{
			UIObject* pObject = o_RLState.GetObject (u);
			if (0 == pObject->bit_Features [VISIBLE])
				continue;
			if ("" == pObject->s_CleanName)
				continue;
			mtxLogProb.GetSlice (mtx_LogProb_C, lwc, pObject->i_Index);

			double dMaxLogProb = NAN;
			for (int ic = 0; ic < Command::i_Count; ++ ic)
			{
				mxf_Features_C.GetSlice (&pvecFeatures, lwc, pObject->i_Index, ic);
				pvecFeatures->Memset (0);
				double dLogProb = ComputeScore_C (lwc, pObject, ic, *pvecFeatures);
				mtxLogProb (ic) = dLogProb;
				if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
					dMaxLogProb = dLogProb;
			}


			// scale log probs to avoid over/underflow	
			double dNormalization = 0;
			for (int ic = 0; ic < Command::i_Count; ++ ic)
			{
				double dLogProb = mtxLogProb (ic);
				dLogProb -= dMaxLogProb;
				mtxLogProb (ic) = dLogProb;
				dNormalization += exp (dLogProb);
			}


			// normalization...				
			if (0 != dNormalization)
			{
				double dLogNormalization = log (dNormalization);
				for (int ic = 0; ic < Command::i_Count; ++ ic)
					mtxLogProb (ic) -= dLogNormalization;
			}
			else
				cout << "[WARNING]  zero pdf in ComputeProb_C" << endl;
		}
	}

	// the candidate count is only used to compute the	
	// max amount of memory required for the compound	
	// prob vector.  The wc,wu,wp,u counts are already	
	// accounted for by the other candidate counts, so	
	// this count here only has to count the commands...
	l_CandidateCount_C = Command::i_Count;

	return true;
}





//														
bool EpisodeStep::ComputeProb_A (bool _bAllowWordlessAction)
{
	mxf_Features_A.Create (o_RLState.TextLength () + 1,
						   o_RLState.EnvironmentObjectCount (),
						   Command::i_Count,
						   Learner::FeatureCount_A ());

	mtx_LogProb_A.Create (o_RLState.TextLength () + 1,
					    o_RLState.EnvironmentObjectCount (),
					    Command::i_Count);
	mtx_LogProb_A.Initialize (NAN);

	//											
	Vector <float> vecFeatures;

	// we need to compute LogProb_A only for the
	// actions for which the other three values	
	// are valid...								
	long lCandidateCount = 0;
	long lDuplicates = 0;
	Matrix <double, 1>	mtxLogProbU;
	Matrix <double, 1>	mtxLogProbC;

	String_set_t	setMessages;


	double dMaxLogProb = NAN;
	for (char lwc = 0; lwc < o_RLState.TextLength () + 1; ++ lwc)
	{
		bool bNoneWC = (o_RLState.TextLength () == lwc);
		for (int lwu = 0; lwu < o_RLState.TextLength () + 1; ++ lwu)
		{
			bool bNoneWU = (o_RLState.TextLength () == lwu);
			if (((true == bNoneWC) && (false == bNoneWU)) ||
				((false == bNoneWC) && (true == bNoneWU)))
				continue;
			if ((lwc == lwu) && (lwu != o_RLState.TextLength ()))
				continue;
			mtxLogProbU.GetSlice (mtx_LogProb_U, lwu);

			for (int lwp = 0; lwp < o_RLState.TextLength () + 1; ++ lwp)
			{
				if ((lwp == lwc) && (false == bNoneWU))
					continue;
				if ((lwp == lwu) && (false == bNoneWU))
					continue;
				if (true == isnan (mtx_LogProb_WC_WU_WP (lwc, lwu, lwp)))
					continue;
				bool bNoneWP = (o_RLState.TextLength () == lwp);
				bool bHasParam = !bNoneWP;
				if ((true == bNoneWC) && (false == bNoneWP))
					continue;

				ITERATE (UIObject_set_t, set_PotentiallyInteractableObjects, ite)
				{
					UIObject* pObject = *ite;
					if (true == isnan (mtxLogProbU (pObject->i_Index)))
						continue;

					mtxLogProbC.GetSlice (mtx_LogProb_C, lwc, pObject->i_Index);
					for (int ic = 0; ic < Command::i_Count; ++ ic)
					{
						if (true == isnan (mtxLogProbC (ic)))
							continue;

						bool bNeedsParam = Command::NeedsParam (ic);
						bool bCommandParamCombinationOK = false;
						if (((true == bNeedsParam) && (true == bHasParam)) ||
							 ((false == bNeedsParam) && (false == bHasParam)))
							bCommandParamCombinationOK = true;
						if (false == bCommandParamCombinationOK)
							continue;

						// check if we have already computed this A...	
						if (false == isnan (mtx_LogProb_A (lwp, pObject->i_Index, ic)))
						{
							++ lDuplicates;
							continue;
						}

						++ lCandidateCount;
						mxf_Features_A.GetSlice (vecFeatures, lwp, pObject->i_Index, ic);
						vecFeatures.Memset (0);
						double dLogProb = ComputeScore_A (lwp, pObject, ic, vecFeatures);
			
						mtx_LogProb_A (lwp, pObject->i_Index, ic) = dLogProb;

		
						if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
							dMaxLogProb = dLogProb;
					}
				}
			}
		}
	}


	// actions without consuming words ...					
	if (true == _bAllowWordlessAction)
	{
		int lwp = o_RLState.TextLength ();
		for (long u = 0; u < o_RLState.EnvironmentObjectCount (); ++ u)
		{
			UIObject* pObject = o_RLState.GetObject (u);
			if (0 == pObject->bit_Features [VISIBLE])
				continue;
			if ("" == pObject->s_CleanName)
				continue;

			for (int ic = 0; ic < Command::i_Count; ++ ic)
			{
				bool bNeedsParam = Command::NeedsParam (ic);
				if (true == bNeedsParam)
					continue;

				// check if we have already computed this A...	
				if (false == isnan (mtx_LogProb_A (lwp, pObject->i_Index, ic)))
				{
					++ lDuplicates;
					continue;
				}

				++ lCandidateCount;
				mxf_Features_A.GetSlice (vecFeatures, lwp, pObject->i_Index, ic);
				vecFeatures.Memset (0);
				double dLogProb = ComputeScore_A (lwp, pObject, ic, vecFeatures);
				mtx_LogProb_A (lwp, pObject->i_Index, ic) = dLogProb;

				if ((true == isnan (dMaxLogProb)) || (dMaxLogProb < dLogProb))
					dMaxLogProb = dLogProb;
			}
		}
	}


	// scale log probs to avoid over/underflow...	
	double dNormalization = 0;
	for (int lwp = 0; lwp < o_RLState.TextLength () + 1; ++ lwp)
	{
		for (long u = 0; u < o_RLState.EnvironmentObjectCount (); ++ u)
		{
			UIObject* pObject = o_RLState.GetObject (u);
			for (int ic = 0; ic < Command::i_Count; ++ ic)
			{
				if (false == isnan (mtx_LogProb_A (lwp, pObject->i_Index, ic)))
				{
					double dLogProb = mtx_LogProb_A (lwp, pObject->i_Index, ic);
					dLogProb -= dMaxLogProb;
					mtx_LogProb_A (lwp, pObject->i_Index, ic) = dLogProb;
					dNormalization += exp (dLogProb);
				}
			}
		}
	}


	// normalize probs ...					
	double dLogNormalization = log (dNormalization);
	for (int lwp = 0; lwp < o_RLState.TextLength () + 1; ++ lwp)
	{
		for (long u = 0; u < o_RLState.EnvironmentObjectCount (); ++ u)
		{
			UIObject* pObject = o_RLState.GetObject (u);
			for (int ic = 0; ic < Command::i_Count; ++ ic)
			{
				if (false == isnan (mtx_LogProb_A (lwp, pObject->i_Index, ic)))
					mtx_LogProb_A (lwp, pObject->i_Index, ic) -= dLogNormalization;
			}
		}
	}


	return true;
}




