#include <algorithm>
#include "Learner.h"
#include "Vocabulary.h"
#include "FieldId.h"
using namespace std;


int		EpisodeStep::i_NeighbourhoodU;
int		EpisodeStep::i_NeighbourhoodWP;
bool	EpisodeStep::b_UseA;


//																
int EpisodeStep::DistanceToIndex (int _iDistance, int _iRange)
{
	if (_iDistance < 0)
	{
		_iDistance += _iRange;
		if (_iDistance < 0)
			return 0;
	}
	else
	{
		_iDistance += _iRange - 1;
		if (_iDistance > 2 * _iRange - 1)
			return 2 * _iRange - 1;
	}
	return _iDistance;
}



//																
double EpisodeStep::ComputeScore_WC_WU_WP (int _iLocationWC, 
											int _iLocationWU,
											int _iLocationWP,
											Vector<float>& _vecFeatures)
{
	double dScore = 0;
	size_t iOffset = 0;

	// ---------------------------------------------
	// feature for wordless action					
	// ---------------------------------------------
	if (_iLocationWU == o_RLState.TextLength ())
	{
		iOffset = Vocabulary::l_Count + 20 +
				  Vocabulary::l_Count + 2 + 20 +
				  2 * Vocabulary::l_Count + 40 + 
				  Vocabulary::l_Count + 4;

		_vecFeatures (iOffset) = 100;
		dScore += (*pvec_Weights_WC_WU_WP) [iOffset] * 100;
		++ iOffset;

		assert (pvec_Weights_WC_WU_WP->Size () == iOffset);
		return dScore;
	}



	// ---------------------------------------------
	// WC features									
	// ---------------------------------------------
	
	// current word's type							
	int iIndexWC = o_RLState.iGetWord (_iLocationWC);
	_vecFeatures (iIndexWC) = 1;
	dScore += (*pvec_Weights_WC_WU_WP) [iIndexWC];
	iOffset	= Vocabulary::l_Count;


	// distance of current word from last used word	
	{
		int iDistanceToU = _iLocationWC + 1;
		int iDistanceToC = _iLocationWC + 1;
		int iDistanceToP = 0;

		if (-1 != o_RLState.PreviousLocation_WU ())
			iDistanceToU = _iLocationWC - o_RLState.PreviousLocation_WU ();
		if (-1 != o_RLState.PreviousLocation_WC ())
			iDistanceToC = _iLocationWC - o_RLState.PreviousLocation_WC ();
		if (-1 != o_RLState.PreviousLocation_WP ())
			iDistanceToP = _iLocationWC - o_RLState.PreviousLocation_WP ();

		int iDistance = -10000;
		if ((iDistanceToU > 0) && (iDistanceToC > 0))
		{
			iDistance = (iDistanceToU < iDistanceToC)? iDistanceToU : iDistanceToC;
			if (iDistanceToP > 0)
				iDistance = (iDistance < iDistanceToP)? iDistance : iDistanceToP;
		}
		if ((iDistanceToU < 0) && (iDistanceToC < 0))
		{
			iDistance = (iDistanceToU > iDistanceToC)? iDistanceToU : iDistanceToC;
			if (iDistanceToP < 0)
				iDistance = (iDistance > iDistanceToP)? iDistance : iDistanceToP;
		}

		if (-10000 != iDistance)
		{
			int iDistanceIndex = EpisodeStep::DistanceToIndex (iDistance, 10);
			assert (iDistanceIndex < 40);
			_vecFeatures (iOffset + iDistanceIndex) = 1;
			dScore += (*pvec_Weights_WC_WU_WP) [iOffset + iDistanceIndex];

			if (iDistanceIndex > 0)
			{
				_vecFeatures (iOffset + iDistanceIndex - 1) = 0.5;
				dScore += 0.5 * (*pvec_Weights_WC_WU_WP) [iOffset + iDistanceIndex - 1];
			}
			if (iDistanceIndex < 19)
			{
				_vecFeatures (iOffset + iDistanceIndex + 1) = 0.5;
				dScore += 0.5 * (*pvec_Weights_WC_WU_WP) [iOffset + iDistanceIndex + 1];
			}
		}

		iOffset += 20;
	}


	// ---------------------------------------------
	// WU features									
	// ---------------------------------------------
	
	String sWordWU = o_RLState.sGetWord (_iLocationWU);
	int iIndexWU = o_RLState.iGetWord (_iLocationWU);
	_vecFeatures (iOffset + iIndexWU) = 1;
	dScore += (*pvec_Weights_WC_WU_WP) [iOffset + iIndexWU];
	iOffset	+= Vocabulary::l_Count;

	// Presence of objects in s^*_t containing 		
	// current word lowest edit distance in s^*_t	
	// to current word								
	if (true == o_RLState.CurrentStateHasWord (sWordWU))
	{
		_vecFeatures (iOffset) = 1;
		dScore += (*pvec_Weights_WC_WU_WP) [iOffset];
		float fLowestEdit = o_RLState.LowestEditDistanceInState (sWordWU);
		_vecFeatures (iOffset + 1) = fLowestEdit;
		dScore += (*pvec_Weights_WC_WU_WP) [iOffset + 1] * fLowestEdit;
	}
	else
		abort ();

	iOffset += 2;


	// ---------------------------------------------
	// WC - WU joint features						
	// ---------------------------------------------
	
	// distance of WU from WC						
	{
		int iDistance = EpisodeStep::DistanceToIndex (_iLocationWU - _iLocationWC, 10);
		_vecFeatures (iOffset + iDistance) = 1;
		dScore += (*pvec_Weights_WC_WU_WP) [iOffset + iDistance];

		if (iDistance > 0)
		{
			_vecFeatures (iOffset + iDistance - 1) = 0.5;
			dScore += 0.5 * (*pvec_Weights_WC_WU_WP) [iOffset + iDistance - 1];
		}
		if (iDistance < 19)
		{
			_vecFeatures (iOffset + iDistance + 1) = 0.5;
			dScore += 0.5 * (*pvec_Weights_WC_WU_WP) [iOffset + iDistance + 1];
		}

		iOffset += 20;
	}


	// ---------------------------------------------
	// WP features									
	// ---------------------------------------------

	// lwp == TextLength () is the special				
	// condition where no parameter is being selected.	
	if (_iLocationWP != o_RLState.TextLength ())
	{
		// WP											
		{
			int iWP = o_RLState.iGetWord (_iLocationWP);
			_vecFeatures (iOffset + iWP) = 1;
			dScore += (*pvec_Weights_WC_WU_WP) [iOffset + iWP];
			iOffset += Vocabulary::l_Count;
		}

		// Words in the neighbourhood of WP				
		{
			int iLeft = _iLocationWP - i_NeighbourhoodWP;
			if (iLeft < 0)
				iLeft = 0;
			for (int w = iLeft; w < _iLocationWP; ++ w)
			{
				int iNeighbour = o_RLState.iGetWord (w);
				_vecFeatures (iOffset + iNeighbour) = 1;
				dScore += (*pvec_Weights_WC_WU_WP) [iOffset + iNeighbour];
			}
		}

		{
			int iRight = _iLocationWP + i_NeighbourhoodWP;
			if (iRight > o_RLState.TextLength ())
				iRight = o_RLState.TextLength ();
			for (int w = _iLocationWP + 1; w < iRight; ++ w)
			{
				int iNeighbour = o_RLState.iGetWord (w);
				_vecFeatures (iOffset + iNeighbour) = 1;
				dScore += (*pvec_Weights_WC_WU_WP) [iOffset + iNeighbour];
			}
			iOffset += Vocabulary::l_Count;
		}

		// distance of current word from WU				
		{
			int iDistance = EpisodeStep::DistanceToIndex (_iLocationWP - _iLocationWU, 10);
			_vecFeatures (iOffset + iDistance) = 1;
			dScore += (*pvec_Weights_WC_WU_WP) [iOffset + iDistance];
			iOffset += 20;
		}
		
		// distance of current word from WC				
		{
			int iDistance = EpisodeStep::DistanceToIndex (_iLocationWP - _iLocationWC, 10);
			_vecFeatures (iOffset + iDistance) = 1;
			dScore += (*pvec_Weights_WC_WU_WP) [iOffset + iDistance];
			iOffset += 20;
		}
	}
	else
		iOffset += 2 * Vocabulary::l_Count + 40;
	
	// non-action words consumed by this action...	
	if (_iLocationWU != o_RLState.TextLength ())
	{
		int iMin = (_iLocationWC < _iLocationWU)? _iLocationWC : _iLocationWU;
		int iMax = (_iLocationWC > _iLocationWU)? _iLocationWC : _iLocationWU;
		if (_iLocationWP != o_RLState.TextLength ())
		{
			iMin = (_iLocationWP < iMin)? _iLocationWP : iMin;
			iMax = (_iLocationWP > iMax)? _iLocationWP : iMax;
		}
		for (int w = iMin; w <= iMax; ++ w)
		{
			int iW = o_RLState.iGetWord (w);
			_vecFeatures (iOffset + iW) = 1;
			dScore += (*pvec_Weights_WC_WU_WP) [iOffset + iW];
		}
	}
	iOffset += Vocabulary::l_Count;

	// sentence distance ...						
	if (_iLocationWU != o_RLState.TextLength ())
	{
		int iSentencePrevious = 0;
		if (-1 != o_RLState.PreviousLocation_WC ())
			iSentencePrevious = o_RLState.GetSentenceIndex (o_RLState.PreviousLocation_WC ());
		int iSentence = o_RLState.GetSentenceIndex (_iLocationWC);
		int iDistance = EpisodeStep::DistanceToIndex (iSentence - iSentencePrevious, 2);
		_vecFeatures (iOffset + iDistance) = 1;
		dScore += (*pvec_Weights_WC_WU_WP) [iOffset + iDistance];
	}
	iOffset += 4;
	

	// added offset for the wordless action feature	
	++ iOffset;

	//												
	assert (pvec_Weights_WC_WU_WP->Size () == iOffset);
	return dScore;
}



//																
double EpisodeStep::ComputeScore_U (int _iLocationWU,
									UIObject* _pObject,
									bool _bFirstAction,
									Vector<float>& _vecFeatures)
{
	double dScore = 0;
	size_t iOffset = 0;

	// World attributes of object					
	for (size_t i = 0; i < FEATURE_COUNT; ++ i)
	{
		if (NAME_EDIT_DIST == i)
			continue;

		if (true == _bFirstAction)
		{
			// on the first action, the values of	
			// the NEWLY_X features make no sense	
			// since we haven't caused a state 		
			// change yet, so we use the value of	
			// feature X itself in place of NEWLY_X	
			if ((NEWLY_VISIBLE == i) ||
				(NEWLY_SHOWING == i) ||
				(NEWLY_INFOCUS == i) ||
				(NEWLY_FG == i) ||
				(NEWLY_INPUT_ENABLED == i) ||
				(NEWLY_SELECTED == i))
			{
				_vecFeatures (i) = _pObject->bit_Features [i - 1];
				dScore += (*pvec_Weights_U) [i] * _vecFeatures (i);
				continue;
			}
		}

		_vecFeatures (i) = _pObject->bit_Features [i];
		dScore += (*pvec_Weights_U) [i] * _vecFeatures (i);
	}
	iOffset += FEATURE_COUNT;

	// edit distance of WU to object name			
	// (this needs to be recomputed)				
	if (_iLocationWU < o_RLState.TextLength ())
	{
		String sWord = o_RLState.sGetWord (_iLocationWU);
		float fEditDistance
			= RLState::NormalizedEditDistance (_pObject->s_CleanName, sWord);

		_vecFeatures (NAME_EDIT_DIST) = fEditDistance;
		dScore += (*pvec_Weights_U) [NAME_EDIT_DIST] * fEditDistance;
	}
	iOffset += 2;

	// presence of object's class name in 			
	// neighbourhood of WU							
	if (_iLocationWU < o_RLState.TextLength ())
	{
		int iLeft = _iLocationWU - EpisodeStep::i_NeighbourhoodU;
		if (iLeft < 0)
			iLeft = 0;
		int iRight = _iLocationWU + EpisodeStep::i_NeighbourhoodU;
		if (iRight > o_RLState.TextLength ())
			iRight = o_RLState.TextLength ();

		String sObjectClass = ObjectClass::Get (_pObject->i_ObjectClass);
		for (int i = iLeft; i < iRight; ++ i)
		{
			if (o_RLState.sGetWord (i) == sObjectClass)
			{
				_vecFeatures (iOffset) = 1;
				dScore += (*pvec_Weights_U) [iOffset];
				break;
			}
		}
	}
	++ iOffset;

	assert (pvec_Weights_U->Size () == iOffset);
	return dScore;
}



//																
double EpisodeStep::ComputeScore_C (int _iLocationWC,
									UIObject* _pObject, 
									int _iCommand,
									BitVector<FEATURES_C>& _vecFeatures)
{
	double dScore = 0;
	size_t iOffset = 0;

	// WC and neighbouring words					
	if (_iLocationWC < o_RLState.TextLength ())
	{
		int iWC = o_RLState.iGetWord (_iLocationWC);
		// _vecFeatures (iWC + _iCommand * Vocabulary::l_Count) = 1;
		_vecFeatures.SetFeature (iWC + _iCommand * Vocabulary::l_Count, 1);
		dScore += (*pvec_Weights_C) [(size_t)(iWC + _iCommand * Vocabulary::l_Count)];
	}
	iOffset += Command::i_Count * Vocabulary::l_Count;

	if (_iLocationWC >= o_RLState.TextLength ())
	{
		// _vecFeatures.SetFeature (iOffset, 1);
		// dScore += (*pvec_Weights_C) [iOffset];
	}
	++ iOffset;


	// UI Object's class							
	_vecFeatures.SetFeature (iOffset + _pObject->i_ObjectClass + ObjectClass::i_Count * _iCommand, 1);
	dScore += (*pvec_Weights_C) [(size_t)(iOffset + _pObject->i_ObjectClass + ObjectClass::i_Count * _iCommand)];
	iOffset += Command::i_Count * ObjectClass::i_Count;


	//												
	assert (pvec_Weights_C->Size () == iOffset);
	return dScore;
}



//																
double EpisodeStep::ComputeScore_A (int _iLocationWP,
									UIObject* _pObject,
									int _iCommand,
									Vector<float>& _vecFeatures)
{
	if (false == b_UseA)
		return 0;

	double dScore = 0;
	size_t iOffset = 0;

	// Convert action info to action string...			
	RLAction oAction;
	oAction.i_SelectedObjectIndex = _pObject->i_Index;
	oAction.p_SelectedObject = _pObject;
	oAction.i_SelectedCommandIndex = _iCommand;
	oAction.i_SelectedParameterWordLocation = _iLocationWP;

	// Get action's result state from known world...	
	GraphCacheNode* pNextState = NULL;
	if ((false == Command::NeedsParam (_iCommand)) ||
		("" != oAction.s_SelectedParameter))
	{
		String sAction = o_RLState.ConvertToCommandString (oAction);
		pNextState = p_ExperienceInterface->PeekAction (sAction, p_CurrentState);
	}

	if (NULL != pNextState)
	{
		int iUnusedWordCount = o_RLState.UnusedWordCount ();
		int_vec_t vecObjectWordIndicesStrict;
		vecObjectWordIndicesStrict.reserve (iUnusedWordCount);
		int_vec_t vecObjectWordIndicesStrictWithoutCurrentObject;
		vecObjectWordIndicesStrictWithoutCurrentObject.reserve (iUnusedWordCount);

		GraphCache* pCache = p_ExperienceInterface->GetCache ();

		int iPotentialObjectWordsStrict = 0;
		int iPotentialObjectWordsStrictWithoutCurrentObject = 0;
		for (int i = 0; i < iUnusedWordCount; ++ i)
		{
			int iWord = o_RLState.iGetWord (o_RLState.UnusedWordLocation (i));
			String sWord = o_RLState.sGetWord (o_RLState.UnusedWordLocation (i));
			if (sWord.length () < 3)
				continue;

			if (NULL != pCache->GetObjectsForWordStrict (iWord))
			{
				vecObjectWordIndicesStrict.push_back (iWord);
				++ iPotentialObjectWordsStrict;
			}

			if (false == RLState::WordMaybeObjectStrict (sWord, _pObject->s_CleanName))
			{
				vecObjectWordIndicesStrictWithoutCurrentObject.push_back (iWord);
				++ iPotentialObjectWordsStrictWithoutCurrentObject;
			}
		}

		// -----------------------------------------------------
		// Strict look-ahead features...						
		// -----------------------------------------------------

		// ratio of words covered by future states in order.	
		{
			GraphCacheNode_set_t setNodesTraversed;
			Coverage oCoverage = pNextState->GetMaxWordsCoveredByATraceInOrder (vecObjectWordIndicesStrict, 0, 0,
																			   setNodesTraversed,
																			   true);
			double dCoverageRatio = 1;
			if (0 != iPotentialObjectWordsStrict)
				dCoverageRatio = oCoverage.f_WordsCovered / (double) iPotentialObjectWordsStrict;
			double dDistance = oCoverage.f_StepsTaken;

			if ((iPotentialObjectWordsStrict != iPotentialObjectWordsStrictWithoutCurrentObject) &&
				(0 != iPotentialObjectWordsStrictWithoutCurrentObject))
			{
				GraphCacheNode_set_t setNodesTraversed;
				Coverage oCoverageAlt
					= pNextState->GetMaxWordsCoveredByATraceInOrder (vecObjectWordIndicesStrictWithoutCurrentObject,
																	 0, 0, setNodesTraversed, true);

				double dCoverageAlt = (oCoverageAlt.f_WordsCovered + 1) 
										/ (double)(iPotentialObjectWordsStrictWithoutCurrentObject + 1);
				if (dCoverageRatio < dCoverageAlt)
				{
					dCoverageRatio = dCoverageAlt;
					dDistance = oCoverageAlt.f_StepsTaken;
				}
			}

			// basic look-ahead feature ...			
			_vecFeatures (0) = dCoverageRatio;
			dScore += (*pvec_Weights_A) [0] * dCoverageRatio;

			// distance-weighted look-ahead ...		
			// double dDistance = oCoverage.f_WordsCovered / (double) oCoverage.f_StepsTaken;
			// _vecFeatures (1) = dCoverageRatio * dDistance;
			// dScore += (*pvec_Weights_A) [1] * dCoverageRatio * dDistance;
			_vecFeatures (1) = dDistance;
			dScore += (*pvec_Weights_A) [1] * dDistance;

			// reward-weighted look-ahead ...		
			float fAverageReward = pNextState->GetAverageReward ();
			_vecFeatures (2) = dCoverageRatio * fAverageReward;
			dScore += (*pvec_Weights_A) [2] * dCoverageRatio * fAverageReward;
		}
	}
	iOffset += 3;

	assert (pvec_Weights_A->Size () == iOffset);
	return dScore;
}

