#include "Learner.h"
#include <math.h>
#include <nlp_macros.h>
#include <nlp_vector.h>
#include "Vocabulary.h"



//															
void EpisodeStep::UpdateTrace_WC_WU_WP (char_set_t& _rsetUseableCommandWords, double _dLogNormalization)
{
	long lFeatureCount = Learner::FeatureCount_WC_WU_WP ();
	vec_Trace_WC_WU_WP.Create (lFeatureCount);
	vec_Trace_WC_WU_WP.Memset (0);
	Vector <double> vecSum (lFeatureCount);
	vecSum.Memset (0);

	int iCommaIndex = Vocabulary::Find (",");
	assert (-1 != iCommaIndex);
	
	Vector <float> vecFeatures;
	for (int c = 0; c < o_RLState.UnusedWordCount (); ++ c)
	{
		char lwc = o_RLState.UnusedWordLocation (c);
		if (iCommaIndex == o_RLState.iGetWord (lwc))
			continue;

		int iSentenceWC = o_RLState.GetSentenceIndex (lwc);
		if (iSentenceWC < i_ActionsStartAtSentence)
			continue;


		if (iSentenceWC > i_ActionsStartAtSentence + 1)
			break;
		if ((false == EpisodeStep::b_LookAtNextSentenceAlso) &&
			(-1 != i_ActionsStartAtSentence) && 
			(iSentenceWC > i_ActionsStartAtSentence))
			break;

		if (iSentenceWC < o_RLState.GetSentenceIndex (o_RLState.PreviousLocation_WC ()))
			continue;

		for (int u = 0; u < o_RLState.i_UnusedPotentialObjectWordCount; ++ u)
		{
			char lwu = o_RLState.p_UnusedPotentialObjectWords [u];
			if (iCommaIndex == o_RLState.iGetWord (lwu))
				continue;

			int iSentenceWU = o_RLState.GetSentenceIndex (lwu);
			if (iSentenceWU > iSentenceWC)
				break;
			if (lwc == lwu)
				continue;

			for (int p = 0; p < o_RLState.UnusedWordCount (); ++ p)
			{
				char lwp = o_RLState.UnusedWordLocation (p);
				if (iCommaIndex == o_RLState.iGetWord (lwp))
					continue;

				int iSentenceWP = o_RLState.GetSentenceIndex (lwp);
				if (iSentenceWP > iSentenceWC)
					break;
				if ((lwc == lwp) || (lwu == lwp))
					continue;
				if (false == o_RLState.WordLocationsAreValid (lwc, lwu, lwp))
					continue;

				mxf_Features_WC_WU_WP.GetSlice (vecFeatures, lwc, lwu, lwp);
				double dLogProb = mtx_LogProb_WC_WU_WP (lwc, lwu, lwp);
				assert (false == isnan (dLogProb));
				double dProb = exp (dLogProb);
				if (0 == dProb)
					continue;
				for (long k = 0; k < lFeatureCount; ++ k)
					vecSum [k] += vecFeatures (k) * dProb;
			}


			// this is the special case of no-param		
			int lwp = o_RLState.TextLength ();
			if (false == o_RLState.WordLocationsAreValid (lwc, lwu, lwp))
				continue;

			mxf_Features_WC_WU_WP.GetSlice (vecFeatures, lwc, lwu, lwp);
			double dLogProb = mtx_LogProb_WC_WU_WP (lwc, lwu, lwp);
			assert (false == isnan (dLogProb));
			double dProb = exp (dLogProb);
			if (0 == dProb)
				continue;
			for (long k = 0; k < lFeatureCount; ++ k)
				vecSum [k] += vecFeatures (k) * dProb;
		}
	}


	// action without object word ...					
	{
		int lwc = o_RLState.TextLength ();
		int lwu = o_RLState.TextLength ();
		int lwp = o_RLState.TextLength ();

		mxf_Features_WC_WU_WP.GetSlice (vecFeatures, lwc, lwu, lwp);
		double dLogProb = mtx_LogProb_WC_WU_WP (lwc, lwu, lwp);
		if (false == isnan (dLogProb))
		{
			double dProb = exp (dLogProb);
			if (dProb > 0)
			{
				for (long k = 0; k < lFeatureCount; ++ k)
					vecSum [k] += vecFeatures (k) * dProb;
			}
		}
	}



	#ifndef NDEBUG
	int swc = o_RLAction.i_SelectedCommandWordLocation;
	assert (_rsetUseableCommandWords.end () != _rsetUseableCommandWords.find (swc));
	#endif

	int swu = o_RLAction.i_SelectedObjectWordLocation;
	int swp = o_RLAction.i_SelectedParameterWordLocation;
	long siu = o_RLAction.i_SelectedObjectIndex;
	int sic = o_RLAction.i_SelectedCommandIndex;


	double dProbOriginal_WCWUWP = 1;
	ITERATE (char_set_t, _rsetUseableCommandWords, iteC)
	{
		double dLogProbWCWUWP = mtx_LogProb_WC_WU_WP (*iteC, swu, swp);
		if (true == isnan (dLogProbWCWUWP))
			continue;
		double dLogProbC = mtx_LogProb_C (*iteC, siu, sic);
		if (true == isnan (dLogProbC))
			continue;

		double dProbRatioLog = dLogProbWCWUWP + dLogProbC - _dLogNormalization;
		if (dProbRatioLog < -20)
			continue;
		double dProbRatio = exp (dProbRatioLog);
		if (dProbRatio > 1.001)
			cout << "prob ratio : " << dProbRatio << endl;
		assert (dProbRatio <= 1.001);

		mxf_Features_WC_WU_WP.GetSlice (vecFeatures, *iteC, swu, swp);
		assert (false == vecFeatures.HasNan ());
		for (long k = 0; k < lFeatureCount; ++ k)
			vec_Trace_WC_WU_WP [k] += dProbOriginal_WCWUWP * dProbRatio 
									  * (vecFeatures (k) - vecSum [k]);
	}

	assert (false == vec_Trace_WC_WU_WP.HasNan ());
}



//															
void EpisodeStep::UpdateTrace_U (void)
{
	long lFeatureCount = Learner::FeatureCount_U ();
	vec_Trace_U.Create (lFeatureCount);
	vec_Trace_U.Memset (0);
	Vector <double> vecSum (lFeatureCount);
	vecSum.Memset (0);

	int swu = o_RLAction.i_SelectedObjectWordLocation;


	Vector <float> vecFeatures;
	for (long iu = 0; iu < o_RLState.EnvironmentObjectCount (); ++ iu)
	{
		double dLogProb = mtx_LogProb_U (swu, iu);
		if (true == isnan (dLogProb))
			continue;
		mxf_Features_U.GetSlice (vecFeatures, swu, iu);
		for (long k = 0; k < lFeatureCount; ++ k)
			vecSum [k] += vecFeatures (k) * exp (dLogProb);
	}


	int siu = o_RLAction.i_SelectedObjectIndex;


	double dProbOriginal_U = 1;
	mxf_Features_U.GetSlice (vecFeatures, swu, siu);
	for (long k = 0; k < lFeatureCount; ++ k)
		vec_Trace_U [k] += dProbOriginal_U * (vecFeatures (k) - vecSum [k]);
}



//															
void EpisodeStep::UpdateTrace_C (char_set_t& _rsetUseableCommandWords, double _dLogNormalization)
{
	long lFeatureCount = Learner::FeatureCount_C ();
	vec_Trace_C.Create (lFeatureCount);
	vec_Trace_C.Memset (0);
	Vector <double> vecSum (lFeatureCount);
	vecSum.Memset (0);

	#ifndef NDEBUG
	int swc = o_RLAction.i_SelectedCommandWordLocation;
	assert (_rsetUseableCommandWords.end () != _rsetUseableCommandWords.find (swc));
	#endif
	long siu = o_RLAction.i_SelectedObjectIndex;
	int swu = o_RLAction.i_SelectedObjectWordLocation;
	int swp = o_RLAction.i_SelectedParameterWordLocation;
	int sic = o_RLAction.i_SelectedCommandIndex;


	double dProbOriginal_C = 1;
	ITERATE (char_set_t, _rsetUseableCommandWords, iteC)
	{
		double dLogProbWCWUWP = mtx_LogProb_WC_WU_WP (*iteC, swu, swp);
		if (true == isnan (dLogProbWCWUWP))
			continue;
		double dLogProbC = mtx_LogProb_C (*iteC, siu, sic);
		if (true == isnan (dLogProbC))
			continue;

		double dProbRatioLog = dLogProbWCWUWP + dLogProbC - _dLogNormalization;
		if (dProbRatioLog < -20)
			continue;
		double dProbRatio = exp (dProbRatioLog);
		if (dProbRatio > 1.001)
			cout << "prob ratio : " << dProbRatio << endl;
		assert (dProbRatio <= 1.001);

		
		vecSum.Memset (0);
		BitVector <FEATURES_C>* pvecFeatures;
		for (int ic = 0; ic < Command::i_Count; ++ ic)
		{
			double dLogProb = mtx_LogProb_C (*iteC, siu, ic);
			if (true == isnan (dLogProb))
				continue;
			mxf_Features_C.GetSlice (&pvecFeatures, *iteC, siu, ic);
			for (long k = 0; k < lFeatureCount; ++ k)
				vecSum [k] += (*pvecFeatures) [k] * exp (dLogProb);
		}

		mxf_Features_C.GetSlice (&pvecFeatures, *iteC, siu, o_RLAction.i_SelectedCommandIndex);
		for (long k = 0; k < lFeatureCount; ++ k)
			vec_Trace_C [k] += dProbOriginal_C * dProbRatio
							   * ((*pvecFeatures) [k] - vecSum [k]);
	}
}



//															
void EpisodeStep::UpdateTrace_A (void)
{
	long lFeatureCount = Learner::FeatureCount_A ();
	vec_Trace_A.Create (lFeatureCount);
	vec_Trace_A.Memset (0);
	Vector <double> vecSum (lFeatureCount);
	vecSum.Memset (0);

	
	Vector <float> vecFeatures;
	for (int lwp = 0; lwp < o_RLState.TextLength () + 1; ++ lwp)
	{
		for (long u = 0; u < o_RLState.EnvironmentObjectCount (); ++ u)
		{
			UIObject* pObject = o_RLState.GetObject (u);

			for (int ic = 0; ic < Command::i_Count; ++ ic)
			{
				double dLogProb = mtx_LogProb_A (lwp, pObject->i_Index, ic);
				if (true == isnan (dLogProb))
					continue;
				double dProb = exp (dLogProb);
				if (0 == dProb)
					continue;

				mxf_Features_A.GetSlice (vecFeatures, lwp, pObject->i_Index, ic);
				for (long k = 0; k < lFeatureCount; ++ k)
					vecSum [k] += vecFeatures (k) * dProb;
			}
		}
	}

	int swp = o_RLAction.i_SelectedParameterWordLocation;
	int siu = o_RLAction.i_SelectedObjectIndex;
	int sic = o_RLAction.i_SelectedCommandIndex;


	double dProbOriginal_A = 1;
	mxf_Features_A.GetSlice (vecFeatures, swp, siu, sic);
	for (long k = 0; k < lFeatureCount; ++ k)
		vec_Trace_A [k] += dProbOriginal_A * (vecFeatures (k) - vecSum [k]);
}




