#ifndef __FEATURE_MATRIX__
#define __FEATURE_MATRIX__
#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <hash_map>
#include <nlp_vector.h>
#include <nlp_macros.h>
using namespace std;


typedef unsigned long					FeatureKey_t;
typedef hash_map <FeatureKey_t, void*>	KeyToData_hmp_t;

//													
template <class TMPL_CLASS, int TMPL_DIM>
class FeatureMatrix
{

	friend class FeatureMatrix <TMPL_CLASS, 5>;
	friend class FeatureMatrix <TMPL_CLASS, 4>;
	friend class FeatureMatrix <TMPL_CLASS, 3>;
	friend class FeatureMatrix <TMPL_CLASS, 2>;
	friend class FeatureMatrix <TMPL_CLASS, 1>;


	private:
		KeyToData_hmp_t		hmp_KeyToData;
		short int			i_BitOffsets [TMPL_DIM];
		int					i_BlockSize;
	
		TMPL_CLASS* GetData (FeatureKey_t _oKey)
		{
			TMPL_CLASS* pData = NULL;

			KeyToData_hmp_t::iterator	ite;
			ite = hmp_KeyToData.find (_oKey);
			if (hmp_KeyToData.end () == ite)
			{
				pData = (TMPL_CLASS*) malloc (sizeof (TMPL_CLASS) * i_BlockSize);
				hmp_KeyToData.insert (make_pair (_oKey, pData));
			}
			else
				pData = (TMPL_CLASS*) ite->second;

			return pData;
		};


	public:
		FeatureMatrix (void);
		FeatureMatrix (int _i0, int _iBlockSize);
		FeatureMatrix (int _i0, int _i1, int _iBlockSize);
		FeatureMatrix (int _i0, int _i1, int _i2, int _iBlockSize);
		FeatureMatrix (int _i0, int _i1, int _i2, int _i3, int _iBlockSize);

		~FeatureMatrix (void);
		void Destroy (void);

		unsigned short int MinRequiredBits (int _iNumber)
			{ return (unsigned short int) floor (1 + log(_iNumber) / (double) log(2)); };
		bool IsInitialized (void)
			{ return (-1 != i_BlockSize); };

		void Create (int _i0, int _iBlockSize);
		void Create (int _i0, int _i1, int _iBlockSize);
		void Create (int _i0, int _i1, int _i2, int _iBlockSize);
		void Create (int _i0, int _i1, int _i2, int _i3, int _iBlockSize);


		void GetSlice (Vector <TMPL_CLASS>& _rFeatureMatrix, int _i0);
		void GetSlice (Vector <TMPL_CLASS>& _rFeatureMatrix, int _i0, int _i1);
		void GetSlice (Vector <TMPL_CLASS>& _rFeatureMatrix, int _i0, int _i1, int _i2);
		void GetSlice (Vector <TMPL_CLASS>& _rFeatureMatrix, int _i0, int _i1, int _i2, int _i3);

		void CopySlice (Vector <TMPL_CLASS>& _rFeatureMatrix, int _i0);
		void CopySlice (Vector <TMPL_CLASS>& _rFeatureMatrix, int _i0, int _i1);
		void CopySlice (Vector <TMPL_CLASS>& _rFeatureMatrix, int _i0, int _i1, int _i2);
		void CopySlice (Vector <TMPL_CLASS>& _rFeatureMatrix, int _i0, int _i1, int _i2, int _i3);
};


//													
template <class TMPL_CLASS, int TMPL_DIM>
FeatureMatrix<TMPL_CLASS, TMPL_DIM>::FeatureMatrix (void)
{
	i_BlockSize = -1;
	memset (i_BitOffsets, 0, sizeof (short int) * TMPL_DIM);
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
FeatureMatrix<TMPL_CLASS, TMPL_DIM>::~FeatureMatrix (void)
{
	ITERATE (KeyToData_hmp_t, hmp_KeyToData, ite)
		free (ite->second);

	hmp_KeyToData.clear ();
	i_BlockSize = -1;
}



//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::Destroy (void)
{
	ITERATE (KeyToData_hmp_t, hmp_KeyToData, ite)
		free (ite->second);

	hmp_KeyToData.clear ();
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
FeatureMatrix<TMPL_CLASS, TMPL_DIM>::FeatureMatrix (int _i0, int _iBlockSize)
{
	Create (_i0, _iBlockSize);
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
FeatureMatrix<TMPL_CLASS, TMPL_DIM>::FeatureMatrix (int _i0, int _i1, int _iBlockSize)
{
	Create (_i0, _i1, _iBlockSize);
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
FeatureMatrix<TMPL_CLASS, TMPL_DIM>::FeatureMatrix (int _i0, int _i1, int _i2, int _iBlockSize)
{
	Create (_i0, _i1, _i2, _iBlockSize);
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
FeatureMatrix<TMPL_CLASS, TMPL_DIM>::FeatureMatrix (int _i0, int _i1, int _i2, int _i3, int _iBlockSize)
{
	Create (_i0, _i1, _i2, _i3, _iBlockSize);
}






//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::Create (int _i0, int _iBlockSize)
{
	assert (1 == TMPL_DIM);
	assert (_i0 <= 8 * sizeof (FeatureKey_t));
	assert (true == hmp_KeyToData.empty ());

	i_BlockSize = _iBlockSize;
	i_BitOffsets [0] = -1;
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::Create (int _i0, int _i1, int _iBlockSize)
{
	assert (2 == TMPL_DIM);
	assert (8 * sizeof (FeatureKey_t) >= (size_t)(MinRequiredBits (_i0) 
												+ MinRequiredBits (_i1)));
	assert (true == hmp_KeyToData.empty ());

	i_BlockSize = _iBlockSize;
	i_BitOffsets [0] = MinRequiredBits (_i1);
	i_BitOffsets [1] = -1;
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::Create (int _i0, int _i1, int _i2, int _iBlockSize)
{
	assert (3 == TMPL_DIM);
	assert (true == hmp_KeyToData.empty ());
	assert (8 * sizeof (FeatureKey_t) >= (size_t)(MinRequiredBits (_i0) 
												+ MinRequiredBits (_i1)
												+ MinRequiredBits (_i2)));

	i_BlockSize = _iBlockSize;
	i_BitOffsets [0] = MinRequiredBits (_i1) + MinRequiredBits (_i2);
	i_BitOffsets [1] = MinRequiredBits (_i2);
	i_BitOffsets [2] = -1;
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::Create (int _i0, int _i1, int _i2, int _i3, int _iBlockSize)
{
	assert (4 == TMPL_DIM);
	assert (8 * sizeof (FeatureKey_t) >= (size_t)(MinRequiredBits (_i0) 
												+ MinRequiredBits (_i1)
												+ MinRequiredBits (_i2)
												+ MinRequiredBits (_i3)));
	assert (true == hmp_KeyToData.empty ());

	i_BlockSize = _iBlockSize;
	i_BitOffsets [0] = MinRequiredBits (_i1) + MinRequiredBits (_i2) + MinRequiredBits (_i3);
	i_BitOffsets [1] = MinRequiredBits (_i2) + MinRequiredBits (_i3);
	i_BitOffsets [2] = MinRequiredBits (_i3);
	i_BitOffsets [3] = -1;
}






//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::GetSlice (Vector <TMPL_CLASS>& _rVector,
													 int _i0)
{
	assert (1 == TMPL_DIM);
	FeatureKey_t oKey = _i0;

	TMPL_CLASS* pData = GetData (oKey);
	_rVector.SetData (pData, i_BlockSize);
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::GetSlice (Vector <TMPL_CLASS>& _rVector,
													 int _i0, int _i1)
{
	assert (2 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						_i1;

	TMPL_CLASS* pData = GetData (oKey);
	_rVector.SetData (pData, i_BlockSize);
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::GetSlice (Vector <TMPL_CLASS>& _rVector,
													 int _i0, int _i1, int _i2)
{
	assert (3 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						(_i1 << i_BitOffsets [1]) |
						_i2;

	TMPL_CLASS* pData = GetData (oKey);
	_rVector.SetData (pData, i_BlockSize);
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::GetSlice (Vector <TMPL_CLASS>& _rVector,
													 int _i0, int _i1, int _i2, int _i3)
{
	assert (4 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						(_i1 << i_BitOffsets [1]) |
						(_i2 << i_BitOffsets [2]) |
						_i3;

	TMPL_CLASS* pData = GetData (oKey);
	_rVector.SetData (pData, i_BlockSize);
}






//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::CopySlice (Vector <TMPL_CLASS>& _rVector,
													 int _i0)
{
	assert (1 == TMPL_DIM);
	FeatureKey_t oKey = _i0;

	TMPL_CLASS* pData = GetData (oKey);
	_rVector.CopyData (pData, i_BlockSize);
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::CopySlice (Vector <TMPL_CLASS>& _rVector,
													 int _i0, int _i1)
{
	assert (2 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						_i1;

	TMPL_CLASS* pData = GetData (oKey);
	_rVector.CopyData (pData, i_BlockSize);
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::CopySlice (Vector <TMPL_CLASS>& _rVector, 
													 int _i0, int _i1, int _i2)
{
	assert (3 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						(_i1 << i_BitOffsets [1]) |
						_i2;

	TMPL_CLASS* pData = GetData (oKey);
	_rVector.CopyData (pData, i_BlockSize);
}


//													
template <class TMPL_CLASS, int TMPL_DIM>
void FeatureMatrix<TMPL_CLASS, TMPL_DIM>::CopySlice (Vector <TMPL_CLASS>& _rVector,
													 int _i0, int _i1, int _i2, int _i3)
{
	assert (4 == TMPL_DIM);
	FeatureKey_t oKey = (_i0 << i_BitOffsets [0]) |
						(_i1 << i_BitOffsets [1]) |
						(_i2 << i_BitOffsets [2]) |
						_i3;

	TMPL_CLASS* pData = GetData (oKey);
	_rVector.CopyData (pData, i_BlockSize);
}





#endif

