#include <nlp_macros.h>
#include <nlp_time.h>
#include <nlp_config.h>
#include "GraphCache.h"
#include "UIObject.h"
#include "State.h"
#include "Vocabulary.h"
#include "Learner.h"

// #define DEBUG_PRINT_LOG

//														
GraphCacheNode::GraphCacheNode (void)
	: CacheNode ()
{
	c_NodeIndex = -1;
	b_IsState = false;
	f_TotalReward = 0;
	ui_RewardCount = 0;
}


//														
GraphCacheNode::~GraphCacheNode (void)
{
	assert (-2 != c_NodeIndex);

	if (-1 != c_NodeIndex)
		((GraphCache*)p_Cache)->ForgetNode (this);
	c_NodeIndex = -2;
}


//														
void GraphCacheNode::AddReward (float _fReward)
{
	f_TotalReward += _fReward;
	++ ui_RewardCount;
}


//														
float GraphCacheNode::GetAverageReward (void)
{
	if (false == b_IsState)
	{
		ITERATE (ActionToState_hmp_t, hmp_ActionToState, ite)
		{
			GraphCacheNode* pChild = (GraphCacheNode*) ite->second;
			f_TotalReward += pChild->f_TotalReward;
			ui_RewardCount += pChild->ui_RewardCount;
		}
	}

	if (0 == ui_RewardCount)
		return 0;
	return f_TotalReward / (float) ui_RewardCount;
}


//														
void GraphCacheNode::SetState (String& _rState)
{
	CacheNode::SetState (_rState);
	b_IsState = _rState.Has ("\x01");
}


//														
void GraphCacheNode::AddChild (String& _rAction, CacheNode* _pChild)
{
	CacheNode::AddChild (_rAction, _pChild);
}


//														
void GraphCacheNode::DiscardBranch (String& _rAction)
{
	GraphCacheNode* pNextNode = (GraphCacheNode*) FindChild (_rAction);
	EnhancedState::DiscardBranch (this, pNextNode);

	CacheNode::DiscardBranch (_rAction);
}


//																
bool GraphCacheNode::IsPeriphery (void)
{
	return (true == hmp_ActionToState.empty ());
}


//																
void GraphCacheNode::GetReachableNodes (GraphCacheNode_set_t& _rsetNodes)
{
	_rsetNodes.insert (this);

	ITERATE (ActionToState_hmp_t, hmp_ActionToState, ite)
	{
		GraphCacheNode* pNextNode = (GraphCacheNode*) ite->second;
		if (_rsetNodes.end () != _rsetNodes.find (pNextNode))
			continue;
		pNextNode->GetReachableNodes (_rsetNodes);
	}
}


//														
int GraphCacheNode::GetMaxWordsCoveredByATrace (int_vec_t& _rvecWords,
												GraphCacheNode_set_t& _rsetNodesTraversed,
												bool _bStrict)
{
	if (_rsetNodesTraversed.end () != _rsetNodesTraversed.find (this))
		return 0;
	_rsetNodesTraversed.insert (this);

	// check how many words are covered by this state...
	int iWordsCoveredByThisState = 0;
	int_vec_t	vecRemainingWords;

	GraphCache* pGraphCache = (GraphCache*)p_Cache;
	int_set_t* psetStateObjects = pGraphCache->GetObjectsForState (this);
	if (NULL != psetStateObjects)
	{
		vecRemainingWords.reserve (_rvecWords.size ());
		ITERATE (int_vec_t, _rvecWords, ite)
		{
			int_set_t* psetWordObjects = NULL;
			if (true == _bStrict)
				psetWordObjects = pGraphCache->GetObjectsForWordStrict (*ite);
			else
				psetWordObjects = pGraphCache->GetObjectsForWord (*ite);


			if (NULL == psetWordObjects)
			{
				vecRemainingWords.push_back (*ite);
				continue;
			}


			bool bWordIsCoveredByState = false;
			ITERATE (int_set_t, (*psetWordObjects), iteIntersect)
			{
				if (psetStateObjects->end () == psetStateObjects->find (*iteIntersect))
					continue;
				bWordIsCoveredByState = true;
				break;
			}

			if (true == bWordIsCoveredByState)
				++ iWordsCoveredByThisState;
			else
				vecRemainingWords.push_back (*ite);
		}
	}
	else
		vecRemainingWords = _rvecWords;


	// check how many of the remaining words are	
	// covered by some trace from this state 		
	int iMaxWordsCoveredByTrace = 0;
	ITERATE (ActionToState_hmp_t, hmp_ActionToState, ite)
	{
		GraphCacheNode* pChild = (GraphCacheNode*)ite->second;
		int iWordsCovered = pChild->GetMaxWordsCoveredByATrace (vecRemainingWords,
																_rsetNodesTraversed,
																_bStrict);
		if (iMaxWordsCoveredByTrace < iWordsCovered)
			iMaxWordsCoveredByTrace = iWordsCovered;
	}

	return iWordsCoveredByThisState + iMaxWordsCoveredByTrace;
}




//														
Coverage GraphCacheNode::GetMaxWordsCoveredByATraceInOrder (int_vec_t& _rvecWords, 
															int _iDepth,
															int _iWordlessActions,
															GraphCacheNode_set_t& _rsetNodesTraversed,
															bool _bStrict)
{
	Coverage oCoverage;
	if (_rsetNodesTraversed.end () != _rsetNodesTraversed.find (this))
		return oCoverage;
	_rsetNodesTraversed.insert (this);

	// if we've reached the max wordless actions,		
	// return the best we've gotten so far...			
	if (_iWordlessActions > Learner::i_MaxWordlessActions)
		return oCoverage;

	// check how many words are covered by this state...
	float fMaxWordsCoveredFromHere = 0;
	float fStepsTaken = 0;
	float fReward = 0;
	int iRewardCount = 0;

	GraphCache* pGraphCache = (GraphCache*)p_Cache;
	int_set_t* psetStateObjects = pGraphCache->GetObjectsForState (this);
	if (NULL != psetStateObjects)
	{
		size_t iWordCount = _rvecWords.size ();
		for (size_t i = 0; i < iWordCount; ++ i)
		{
			int iWord = _rvecWords [i];
			int_set_t* psetWordObjects = NULL;
			if (true == _bStrict)
				psetWordObjects = pGraphCache->GetObjectsForWordStrict (iWord);
			else
				psetWordObjects = pGraphCache->GetObjectsForWord (iWord);


			if (NULL == psetWordObjects)
				continue;


			bool bWordIsCoveredByState = false;
			int iMatchingObjectNameIndex = -1;
			ITERATE (int_set_t, (*psetWordObjects), iteIntersect)
			{
				if (psetStateObjects->end () == psetStateObjects->find (*iteIntersect))
					continue;
				bWordIsCoveredByState = true;
				iMatchingObjectNameIndex = *iteIntersect;
				break;
			}

			if (true == bWordIsCoveredByState)
			{
				// this word is covered by this state...
				// find max covered by children...		
				fMaxWordsCoveredFromHere = 1;
				fStepsTaken = 1 / (float) (_iDepth + 1);
				fReward = GetAverageReward ();
				iRewardCount = 1;

				#ifdef DEBUG_PRINT_LOG
				String sMatchDebug;
				for (int x = 0; x < _iDepth; ++ x)
					sMatchDebug << ' ';
				sMatchDebug << Vocabulary::vec_IndexToWord [iWord] << " == " << ((GraphCache*)p_Cache)->dq_IndexToObjectName [iMatchingObjectNameIndex];
				cout << sMatchDebug << endl;
				#endif


				// copy remaining words ...				
				int_vec_t vecRemainingWords;
				vecRemainingWords.reserve (_rvecWords.size ());
				for (size_t x = 0; x < iWordCount; ++ x)
				{
					if (x == i)
						continue;
					vecRemainingWords.push_back (_rvecWords [x]);
				}

				// find max child overlap...			
				ITERATE (ActionToState_hmp_t, hmp_ActionToState, ite)
				{
					String sAction = ite->first;
					GraphCacheNode* pChild = (GraphCacheNode*)ite->second;

					String sObjectInteractedWith = DebugFindObject (sAction);
					String sWordOverlappedWith = Vocabulary::vec_IndexToWord [iWord];
					if (true == _bStrict)
					{
						if (false == RLState::WordMaybeObjectStrict (sWordOverlappedWith, sObjectInteractedWith))
							continue;
					}
					else
					{
						if (false == RLState::WordMaybeObject (sWordOverlappedWith, sObjectInteractedWith))
							continue;
					}

					Coverage oChildCoverage = pChild->GetMaxWordsCoveredByATraceInOrder (vecRemainingWords,
																					_iDepth + 1,
																					0,
																					_rsetNodesTraversed,
																					_bStrict);

					float fEditDistance = 1;
					if (fMaxWordsCoveredFromHere < oChildCoverage.f_WordsCovered + fEditDistance)
					{
						#ifdef DEBUG_PRINT_LOG
						if ((oChildCoverage.f_WordsCovered > 0) && (false == sAction.StartsWith ("*")))
						{
							cout << sMatchDebug << endl;
							for (int x = 0; x < _iDepth; ++ x)
								cout << ' ';
							cout << oChildCoverage.f_WordsCovered << " : " 
								 << oChildCoverage.f_StepsTaken << " : " << sAction << " : " 
								 << DebugFindObject (sAction) << endl;
						}
						#endif

						fMaxWordsCoveredFromHere = oChildCoverage.f_WordsCovered + fEditDistance;
						fStepsTaken = 1 / (float)(_iDepth + 1) + oChildCoverage.f_StepsTaken;
						fReward = GetAverageReward () + oChildCoverage.f_Reward;
						iRewardCount = 1 + oChildCoverage.i_RewardCount;
					}
					else if (fMaxWordsCoveredFromHere == oChildCoverage.f_WordsCovered + fEditDistance)
					{
						if (fStepsTaken < 1 / (float)(_iDepth + 1) + oChildCoverage.f_StepsTaken)
						{
							fStepsTaken = 1 / (float)(_iDepth + 1) + oChildCoverage.f_StepsTaken;
							fReward = GetAverageReward () + oChildCoverage.f_Reward;
							iRewardCount = 1 + oChildCoverage.i_RewardCount;
						}
					}
				}
			}
		}
	}

	// every state visited is a step, so we need to 	
	// count the current one regardless of matches...	
	oCoverage.f_WordsCovered = fMaxWordsCoveredFromHere;
	oCoverage.f_StepsTaken = fStepsTaken;
	oCoverage.f_Reward = fReward;
	oCoverage.i_RewardCount = iRewardCount;

	// check how many of the remaining words are	
	// covered by some trace from this state 		
	ITERATE (ActionToState_hmp_t, hmp_ActionToState, ite)
	{
		String sAction = ite->first;
		GraphCacheNode* pChild = (GraphCacheNode*)ite->second;

		int iIncrement = (false == sAction.StartsWith ("*"))? 1 : 0;
		Coverage oChildCoverage = pChild->GetMaxWordsCoveredByATraceInOrder (_rvecWords, 
																		_iDepth + iIncrement,
																		_iWordlessActions + iIncrement,
																		_rsetNodesTraversed,
																		_bStrict);

		if (fMaxWordsCoveredFromHere < oChildCoverage.f_WordsCovered)
		{
			#ifdef DEBUG_PRINT_LOG
			if ((oChildCoverage.f_WordsCovered > 0) && (false == sAction.StartsWith ("*")))
			{
				for (int x = 0; x < _iDepth; ++ x)
					cout << ' ';
				cout << oChildCoverage.f_WordsCovered << " : " 
					 << oChildCoverage.f_StepsTaken << " : " << sAction 
					 << " : " << DebugFindObject (sAction) << endl;
			}
			#endif

			fMaxWordsCoveredFromHere = oChildCoverage.f_WordsCovered;
			fStepsTaken = oChildCoverage.f_StepsTaken;
			fReward = oChildCoverage.f_Reward;
			iRewardCount = oChildCoverage.i_RewardCount;
		}
		else if (fMaxWordsCoveredFromHere == oChildCoverage.f_WordsCovered)
		{
			if (fStepsTaken < oChildCoverage.f_StepsTaken)
			{
				fStepsTaken = oChildCoverage.f_StepsTaken;
				fReward = oChildCoverage.f_Reward;
				iRewardCount = oChildCoverage.i_RewardCount;
			}
		}
	}


	//												
	if (oCoverage.f_WordsCovered < fMaxWordsCoveredFromHere)
	{
		oCoverage.f_WordsCovered = fMaxWordsCoveredFromHere;
		oCoverage.f_StepsTaken = fStepsTaken;
		oCoverage.f_Reward = fReward;
		oCoverage.i_RewardCount = iRewardCount;
	}
	else if (oCoverage.f_WordsCovered == fMaxWordsCoveredFromHere)
	{
		if (oCoverage.f_StepsTaken < fStepsTaken + 1)
		{
			oCoverage.f_StepsTaken = fStepsTaken;
			oCoverage.f_Reward = fReward;
			oCoverage.i_RewardCount = iRewardCount;
		}
	}
	return oCoverage;
}



//														
String GraphCacheNode::DebugFindObject (String& _sAction)
{
	if (true == _sAction.StartsWith ("*"))
		return "";

	String sId = _sAction.substr (1, _sAction.find (":") - 1);

	int iObjects = GetObjectCount ();
	for (int i = 0; i < iObjects; ++ i)
	{
		UIObject* pObject = (UIObject*) GetObject (i);
		if (sId == pObject->z_ID)
			return pObject->s_CleanName;
	}
	return "[object not found]";
}




//														
//														
int GraphCache::AssociateWithWords (String& _rObjectName)
{
	if ("" == _rObjectName)
		return -1;

	int iObjectNameIndex = -1;
	ObjectName_hmp_t::iterator	iteIndex;
	iteIndex = hmp_ObjectNameToIndex.find (_rObjectName);
	if (hmp_ObjectNameToIndex.end () == iteIndex)
	{
		iObjectNameIndex = hmp_ObjectNameToIndex.size ();
		hmp_ObjectNameToIndex.insert (make_pair (_rObjectName, iObjectNameIndex));
		assert (iObjectNameIndex == (int)dq_IndexToObjectName.size ());
		dq_IndexToObjectName.push_back (_rObjectName);
	}
	else
	{
		// we have previously processed this object name...
		return iteIndex->second;
	}

	int iWordIndex = -1;
	ITERATE (String_vec_t, Vocabulary::vec_IndexToWord, ite)
	{
		++ iWordIndex;

		// strict word-object mapping...				
		if (true == RLState::WordMaybeObjectStrict (*ite, _rObjectName))
		{
			int_set_t* psetObjectNames = NULL;
			WordToObjectNames_hmp_t::iterator	iteObjects;
			iteObjects = hmp_WordToObjectNameStrict.find (iWordIndex);
			if (hmp_WordToObjectNameStrict.end () == iteObjects)
			{
				pair <WordToObjectNames_hmp_t::iterator, bool> pairInsert;
				pairInsert = hmp_WordToObjectNameStrict.insert (make_pair (iWordIndex, 
																		   int_set_t ()));
				psetObjectNames = &pairInsert.first->second;
			}
			else
				psetObjectNames = &iteObjects->second;

			psetObjectNames->insert (iObjectNameIndex);
		}

		// non-strict word-object mapping...			
		if (true == RLState::WordMaybeObject (*ite, _rObjectName))
		{
			int_set_t* psetObjectNames = NULL;
			WordToObjectNames_hmp_t::iterator	iteObjects;
			iteObjects = hmp_WordToObjectName.find (iWordIndex);
			if (hmp_WordToObjectName.end () == iteObjects)
			{
				pair <WordToObjectNames_hmp_t::iterator, bool> pairInsert;
				pairInsert = hmp_WordToObjectName.insert (make_pair (iWordIndex, 
																	 int_set_t ()));
				psetObjectNames = &pairInsert.first->second;
			}
			else
				psetObjectNames = &iteObjects->second;

			psetObjectNames->insert (iObjectNameIndex);
		}
	}

	return iObjectNameIndex;
}


//														
void GraphCache::UpdateGoalMaps (GraphCacheNode* _pNode)
{
	StateToObjectNames_hmp_t::iterator	iteState;
	iteState = hmp_StateToObjectNames.find (_pNode);
	if (hmp_StateToObjectNames.end () != iteState)
		return;

	pair <StateToObjectNames_hmp_t::iterator, bool> pairInsert;
	pairInsert = hmp_StateToObjectNames.insert (make_pair (_pNode, int_set_t ()));
	int_set_t& rsetObjectNames = pairInsert.first->second;

	for (int i = 0; i < _pNode->GetObjectCount (); ++ i)
	{
		UIObject* pObject = (UIObject*) _pNode->GetObject (i);
		int iObjectNameIndex = AssociateWithWords (pObject->s_CleanName);
		rsetObjectNames.insert (iObjectNameIndex);
	}
}


//														
GraphCache::GraphCache (void)
{
	b_LoadingCache = false;

	String sEmptyState ("[graph cache root]");
	p_Root = CreateNode (sEmptyState);
	((GraphCacheNode*) p_Root)->c_NodeIndex = -3;
}


//														
GraphCache::~GraphCache (void)
{
	Cache::Clear ();
}


//														
int_set_t* GraphCache::GetObjectsForWord (int _iWordIndex)
{
	WordToObjectNames_hmp_t::iterator	iteObjects;
	iteObjects = hmp_WordToObjectName.find (_iWordIndex);
	if (hmp_WordToObjectName.end () == iteObjects)
		return NULL;

	assert (false == iteObjects->second.empty ());
	return &iteObjects->second;
}


//														
int_set_t* GraphCache::GetObjectsForWordStrict (int _iWordIndex)
{
	WordToObjectNames_hmp_t::iterator	iteObjects;
	iteObjects = hmp_WordToObjectNameStrict.find (_iWordIndex);
	if (hmp_WordToObjectNameStrict.end () == iteObjects)
		return NULL;

	assert (false == iteObjects->second.empty ());
	return &iteObjects->second;
}


//														
int_set_t* GraphCache::GetObjectsForState (GraphCacheNode* _pNode)
{
	StateToObjectNames_hmp_t::iterator	iteObjects;
	iteObjects = hmp_StateToObjectNames.find (_pNode);
	if (hmp_StateToObjectNames.end () == iteObjects)
		return NULL;

	return &iteObjects->second;
}


//														
void GraphCache::UnpackState (String& _rState, void** _ppObjects, int* _pObjectCount)
{
	UIObject::UnpackState (_rState, _ppObjects, _pObjectCount);
}


//														
CacheNode* GraphCache::CreateNode (String& _rState, bool _bSetModified)
{
	GraphCacheNode* pNode = new GraphCacheNode;
	pNode->p_Cache = this;
	pNode->SetState (_rState);

	if ((true == _bSetModified) || (true == b_LoadingCache))
	{
		if (true == _bSetModified)
			b_Modified = true;

		// this is not a temporary node ...
		if (true == pNode->b_IsState)
		{
			pNode->c_NodeIndex = 10;
			UpdateGoalMaps (pNode);
		}
	}

	return pNode;
}


//														
bool GraphCache::LoadRewards (IndexToNode_hmp_t& _hmpIndexToNode)
{
	if (false == Path::Exists ((config)"reward_cache_file_name"))
	{
		cout << "[INFO]  Reward cache file does not exist, will be created anew." << endl;
		return true;
	}

	File file ((config)"reward_cache_file_name");
	String sLine;
	while (true == file.ReadLine (sLine))
	{
		sLine.Strip ();
		if ("" == sLine)
			continue;

		zchar_dq_t dqValues = sLine.DestructiveSplit ('\x01', 2);

		IndexToNode_hmp_t::iterator	ite;
		ite = _hmpIndexToNode.find (dqValues [0]);
		if (_hmpIndexToNode.end () == ite)
		{
			cout << "[WARNING] LoadReward : node [" << dqValues [0] << "] not found." << endl;
			continue;
		}

		GraphCacheNode* pNode = (GraphCacheNode*) ite->second;
		pNode->f_TotalReward = atof (dqValues [1]);
		pNode->ui_RewardCount = atoi (dqValues [2]);
	}

	file.close ();
	return true;
}


//														
bool GraphCache::Load (void)
{
	if (NULL != p_Root)
		delete p_Root;
	p_Root = NULL;

	b_LoadingCache = true;
	LockForWriting ();

	IndexToNode_hmp_t hmpIndexToNode;
	bool bRet = Cache::LoadThreadUnsafe (hmpIndexToNode);
	b_Modified = false;

	Unlock ();

	if (true == bRet)
		bRet = LoadRewards (hmpIndexToNode);

	b_LoadingCache = false;
	return bRet;
}



//																
bool GraphCache::SaveInternal (String& _rFileName)
{
	CacheNodeToIndex_map_t	mapNodeToIndex;
	if (false == Cache::SaveThreadUnsafe (_rFileName, mapNodeToIndex))
		return false;

	return SaveRewards (mapNodeToIndex);
}



//																
bool GraphCache::SaveRewards (CacheNodeToIndex_map_t& _mapNodeToIndex)
{
	File file ((config)"reward_cache_file_name", ios_base::out);
	ITERATE (CacheNodeToIndex_map_t, _mapNodeToIndex, ite)
	{
		long iIndex = ite->second;
		GraphCacheNode* pNode = (GraphCacheNode*) ite->first;

		file << iIndex << '\x01' << pNode->f_TotalReward << '\x01' << pNode->ui_RewardCount << endl;
	}
	file.flush ();
	file.close ();

	return true;
}



//														
void GraphCache::ForgetNode (GraphCacheNode* _pNode)
{
	hmp_StateToObjectNames.erase (_pNode);
	EnhancedState::DiscardCacheNode (_pNode);
}




