#ifndef __LEARNER_GRAPH_CACHE__
#define __LEARNER_GRAPH_CACHE__

#include <set>
#include <hash_map>
#include <ili_cache.h>
#include <nlp_matrix.h>

#define UNKNOWN				(char)'u'
#define NODE_IN_PATH		(char)'t'
#define NODE_NOT_IN_PATH	(char)'f'


class GraphCache;
class GraphCacheNode;
typedef set <GraphCacheNode*>						GraphCacheNode_set_t;
typedef deque <GraphCacheNode*>						GraphCacheNode_dq_t;
typedef Matrix <char, 2>							char_mtx_t;

typedef set <int>									int_set_t;
typedef vector <int>								int_vec_t;
typedef hash_map <int, int_set_t>					WordToObjectNames_hmp_t;
typedef hash_map <void*, int_set_t>					StateToObjectNames_hmp_t;			
typedef hash_map <String, int>						ObjectName_hmp_t;



// -----------------------------------------------------------
class Coverage
{
	public:
		float	f_WordsCovered;
		float 	f_StepsTaken;
		float	f_Reward;
		int		i_RewardCount;

		Coverage (void)
		{
			f_WordsCovered = 0;
			f_StepsTaken = 0;
			f_Reward = 0;
			i_RewardCount = 0;
		};
};


// -----------------------------------------------------------
class GraphCacheNode : public CacheNode
{
	friend class GraphCache;

	public:
		float			f_TotalReward;
		unsigned int	ui_RewardCount;
		char			c_NodeIndex;
		bool			b_IsState;

		GraphCacheNode (void);
		~GraphCacheNode (void);

		void SetState (String& _rState);
		void AddChild (String& _rAction, CacheNode* _pChild);
		void DiscardBranch (String& _rAction);

		bool IsPeriphery (void);
		void GetReachableNodes (GraphCacheNode_set_t& _rsetNodes);
		int GetChildCount (void)
			{ return hmp_ActionToState.size (); };


		int GetMaxWordsCoveredByATrace (int_vec_t& _rvecWords,
										GraphCacheNode_set_t& _rsetNodesTraversed,
										bool _bStrict);
		Coverage GetMaxWordsCoveredByATraceInOrder (int_vec_t& _rvecWords,
													int _iStartAtWord,
													int _iWordlessActions,
													GraphCacheNode_set_t& _rsetNodesTraversed,
													bool _bStrict);

		void AddReward (float _fReward);
		float GetAverageReward (void);
		String DebugFindObject (String& _sAction);
		String DebugFindWordObject (String& _sWord);
};



// -----------------------------------------------------------
class GraphCache : public Cache
{
	friend class GraphCacheNode;

	protected:
		GraphCacheNode_dq_t		dq_GraphCacheNode;
		bool					b_LoadingCache;

		ObjectName_hmp_t			hmp_ObjectNameToIndex;
		String_dq_t					dq_IndexToObjectName;
		StateToObjectNames_hmp_t	hmp_StateToObjectNames;
		WordToObjectNames_hmp_t		hmp_WordToObjectName;
		WordToObjectNames_hmp_t		hmp_WordToObjectNameStrict;
		CacheNode_set_t				set_StatesWithKnownNames;

		int AssociateWithWords (String& _rObjectName);
		void UpdateGoalMaps (GraphCacheNode* _pNode);

		bool LoadRewards (IndexToNode_hmp_t& _hmpIndexToNode);
		virtual bool SaveInternal (String& _rFileName);
		bool SaveRewards (CacheNodeToIndex_map_t& _mapNodeToIndex);

	public:
		GraphCache (void);
		~GraphCache (void);
		bool Load (void);

		int_set_t* GetObjectsForWord (int _iWordIndex);
		int_set_t* GetObjectsForWordStrict (int _iWordIndex);
		int_set_t* GetObjectsForState (GraphCacheNode* _pNode);

		void UnpackState (String& _rState, void** _ppObjects, int* _pObjectCount);
		CacheNode* CreateNode (String& _rState, bool _bSetModified = true);
		void ForgetNode (GraphCacheNode* _pNode);
};



#endif
