extern "C"
{
// #include <cblas.h>
// #include <clapack.h>
}

// #include <omp.h>
#include <nlp_config.h>
#include <nlp_macros.h>
#include <nlp_time.h>
#include "Learner.h"
#include "Vocabulary.h"
#include "FieldId.h"
#include "Sample.h"


double		Learner::d_PositiveRewardMultiplier = 0;
double		Learner::d_NegativeRewardMultiplier = 0;
size_t		Learner::i_Features_WC_WU_WP = -1;
size_t		Learner::i_Features_U = -1;
size_t		Learner::i_Features_C = -1;
size_t		Learner::i_Features_A = -1;
size_t		Learner::i_FeaturesTotal = -1;
int			Learner::i_MaxWordlessActions = 0;
ofstream	Learner::ofs_Log;


//													
Learner::Learner (void)
{
	i_Verbose = 0;

	e_LearningMode = lm_unknown;
	b_WalkAnnotations = false;

	d_LearningRate = 0;
	d_ExplorationRate = 0;
	d_PositiveRewardMultiplier = 0;
	d_NegativeRewardMultiplier = 0;
	i_MaxWordlessActions = 0;

	l_ProcessedDocuments = 0;
	l_DocumentsPerUpdate = 0;

	p_ExperienceInterface = NULL;
	p_TextInterface = NULL;
}


Learner::~Learner (void)
{
}



//													
void Learner::UpdateWeights (void)
{
	for (size_t f = 0; f < i_Features_WC_WU_WP; ++ f)
		vec_Weights_WC_WU_WP [f] += d_LearningRate 
									* (vec_WeightUpdate_WC_WU_WP [f] 
									- d_RegularizationFactor * vec_Weights_WC_WU_WP [f]);

	for (size_t f = 0; f < i_Features_U; ++ f)
		vec_Weights_U [f] += d_LearningRate 
							 * (vec_WeightUpdate_U [f] 
							 - d_RegularizationFactor * vec_Weights_U [f]);

	for (size_t f = 0; f < i_Features_C; ++ f)
		vec_Weights_C [f] += d_LearningRate
							 * (vec_WeightUpdate_C [f] 
							 - d_RegularizationFactor * vec_Weights_C [f]);

	for (size_t f = 0; f < i_Features_A; ++ f)
		vec_Weights_A [f] += d_LearningRate
							 * (vec_WeightUpdate_A [f] 
							 - d_RegularizationFactor * vec_Weights_A [f]);


	vec_WeightUpdate_WC_WU_WP.Memset (0);
	vec_WeightUpdate_U.Memset (0);
	vec_WeightUpdate_C.Memset (0);
	vec_WeightUpdate_A.Memset (0);
}


//													
void Learner::ForgetWeightUpdates (void)
{
	vec_WeightUpdate_WC_WU_WP.Memset (0);
	vec_WeightUpdate_U.Memset (0);
	vec_WeightUpdate_C.Memset (0);
	vec_WeightUpdate_A.Memset (0);
}


//													
bool Learner::Init (ExperienceInterface* _pExperienceInterface,
					TextInterface* _pTextInterface)
{
	cout << "Initializing learner ..." << endl;

	ofs_Log.open ((config)"interaction_log", ios_base::app);
	if (false == ofs_Log.is_open ())
	{
		cout << "[ERROR]  Failed to open log file : " << (config)"interaction_log" << endl;
		return false;
	}

	ObjectClass::Load ();
	Command::Load ();
	RLState::Init ();
	Sample::Init ();

	p_ExperienceInterface = _pExperienceInterface;
	p_TextInterface = _pTextInterface;

	i_Verbose = (config)"verbose";
	b_WalkAnnotations = ("1" == (config)"walk_annotations");
	String sLearningMode ((config)"learning_mode");
	if ("annotation_only" == sLearningMode)
		e_LearningMode = lm_annotation_only;
	else if ("environment_only" == sLearningMode)
		e_LearningMode = lm_environment_only;
	else if ("mixed_mode" == sLearningMode)
		e_LearningMode = lm_mixed_mode;
	else
	{
		cout << "[ERROR]  Unkonwn learningn mode: " << sLearningMode << '\n'
			 << "         Valid values are: annotation_only\n"
			 << "                           environment_only\n"
			 << "                           mixed_mode" << endl;
	}


	d_LearningRate = (config)"learning_rate";
	d_ExplorationRate = (config)"exploration_rate";
	d_RegularizationFactor = (config)"regularization_factor";

	EpisodeStep::d_ExplorationEpsilon = (config)"exploration_epsilon";
	EpisodeStep::b_ExploreStochastically = (1 == (int)(config)"use_stochastic_exploration");
	EpisodeStep::b_UseSoftmax = (1 == (int)(config)"use_softmax_exploration");
	EpisodeStep::b_UseSmartSoftmax = (1 == (int)(config)"use_smart_softmax_exploration");
	EpisodeStep::b_UseEpsilonGreedy = (1 == (int)(config)"use_epsilon_greedy");
	EpisodeStep::b_AvoidNegativeRewardStates = (1 == (int)(config)"avoid_negative_reward_states");
	EpisodeStep::d_NegativeRewardAvoidanceProb = (config)"negative_reward_avoidance_probability";
	i_DontStrayTillIteration = (config)"dont_stray_till";
	b_RandomlyStrayFromAnnotations = (1 == (int)(config)"randomly_stray_from_annotations");
	d_StrayProbability = (config)"straying_probability";
	d_AlternateStartStateProbability = (config)"alternate_start_state_probability";

	d_PositiveRewardMultiplier = (config)"positive_reward_multiplier";
	d_NegativeRewardMultiplier = (config)"negative_reward_multiplier";
	i_MaxWordlessActions = (config)"max_wordless_actions";

	EpisodeStep::i_NeighbourhoodU = (config)"neighbourhood_for_object";
	EpisodeStep::i_NeighbourhoodWP = (config)"neighbourhood_for_parameter";

	i_Features_WC_WU_WP = Vocabulary::l_Count + 20 +
						  Vocabulary::l_Count + 2 + 20 +
						  2 * Vocabulary::l_Count + 40 + 
						  Vocabulary::l_Count + 4 +
						  1;
	i_Features_U = FEATURE_COUNT + 3;
	i_Features_C = (ObjectClass::i_Count + Vocabulary::l_Count) * Command::i_Count + 1;
	assert (FEATURES_C == i_Features_C);

	i_Features_A = 3;
	i_FeaturesTotal = i_Features_WC_WU_WP + i_Features_U + i_Features_C + i_Features_A;

	cout << "features c : " << i_Features_C << endl;

	vec_Weights_WC_WU_WP.Create (i_Features_WC_WU_WP);
	vec_Weights_U.Create (i_Features_U);
	vec_Weights_C.Create (i_Features_C);
	vec_Weights_A.Create (i_Features_A);

	vec_WeightUpdate_WC_WU_WP.Create (i_Features_WC_WU_WP);
	vec_WeightUpdate_U.Create (i_Features_U);
	vec_WeightUpdate_C.Create (i_Features_C);
	vec_WeightUpdate_A.Create (i_Features_A);
	vec_WeightUpdate_WC_WU_WP.Memset (0);
	vec_WeightUpdate_U.Memset (0);
	vec_WeightUpdate_C.Memset (0);
	vec_WeightUpdate_A.Memset (0);

	cout << "   - done" << endl;
	return true;
}



//													
int Learner::Learn (int _iIteration, int _iSubIteration)
{
	Document::SwitchToTrainingStateGraph ();

	cout << "======================================\n"
			" TRAINING MODE \n"
			"======================================\n";
	cout << "   [learning rate] " << d_LearningRate << '\n';
	ofs_Log << "======================================\n"
			   " TRAINING MODE \n"
			   "======================================\n";
	ofs_Log << "   [learning rate] " << d_LearningRate << '\n';

	int iTotalStartingStates = Document::dq_StartingStates.size ();
	int iStartingState = 0;

	if ((true == b_RandomlyStrayFromAnnotations) &&
		(_iIteration > i_DontStrayTillIteration))
		d_StrayProbability = (config)"straying_probability";
	else
		d_StrayProbability = 0;

	bool bReplayTrace = (1 == (int)(config)"replay_run");

	int iPositiveRewardDocuments = 0;
	ITERATE (String_dq_t, Document::dq_StartingStates, iteStateName)
	{
		++ iStartingState;
		cout << "\n\n   [start-state]" << *iteStateName << '\n';
		ofs_Log << "\n\n   [start-state]" << *iteStateName << '\n';

		String_dq_t* pdqDocumentIds = Document::GetIdsForStartingState (*iteStateName);
		int iTotalDocuments = pdqDocumentIds->size ();
		int iDocument = 0;

		ITERATE (String_dq_t, (*pdqDocumentIds), iteId)
		{
			++ iDocument;

			//												
			Document* pDocument = Document::Find (*iteId);
			if (dt_training != pDocument->e_Type)
				continue;

			cout << '\n';
			cout << "   [time] " << Time::sDateTime () << '\n';
			cout << "   [iteration] " << _iIteration << '-' << _iSubIteration << " " 
				 << iStartingState << '/' << iTotalStartingStates << " "
				 << iDocument << '/' << iTotalDocuments << '\n';
			cout << "   [document] " << *iteId << " " 
				 << pDocument->s_RequiredStartingState << " "
				 << pDocument->e_Type << '\n';

			ofs_Log << '\n';
			ofs_Log << "   [iteration] " << _iIteration << " " 
					<< iStartingState << '/' << iTotalStartingStates << " "
					<< iDocument << '/' << iTotalDocuments << '\n';
			ofs_Log << "   [document] " << *iteId << " " 
					<< pDocument->s_RequiredStartingState << " "
					<< pDocument->e_Type << '\n';


			// interpret instructions ...					
			Annotation* pAnnotation = Annotation::Find (*iteId);
			ReplayTrace* pReplayTrace = NULL;
			if (true == bReplayTrace)
				pReplayTrace = ReplayTrace::Find (_iIteration - 1, *iteId);
			if (NULL != pReplayTrace)
			{
				cout << "   [replay]\n";
				ofs_Log << "   [replay]\n";
			}

			// sample starting state...						
			AnnotatedAction_dq_t* pdqStartStateTrace = NULL;
			bool bJumpToState = Sample::SampleCoinflip (d_AlternateStartStateProbability);
			if (true == bJumpToState)
				pdqStartStateTrace = pDocument->PickStartState ();


			// start episode ...							
			p_TextInterface->StartEpisode (*iteId);
			p_ExperienceInterface->StartEpisode (pDocument->s_RequiredStartingState);


			// compute trace of actions...					
			Episode* pEpisode = new Episode (this, *iteId);
			if ((true == b_WalkAnnotations) && (ANNOTATION_FULL == pAnnotation->c_Type))
				// annotated training, walk annotations.	
				pEpisode->WalkAnnotationTrace (p_ExperienceInterface,
											   p_TextInterface);

			else if ((true == bReplayTrace) && (NULL != pReplayTrace))
				// replaying old trace, walk trace.			
				pEpisode->WalkReplayTrace (p_ExperienceInterface,
										   p_TextInterface,
										   pReplayTrace);
			else
				// following policy...						
				pEpisode->ComputeTrace (p_ExperienceInterface, 
									    p_TextInterface, 
									    tt_training_mode,
										pdqStartStateTrace);

			if (i_Verbose > 0)
				cout << *pEpisode;
			ofs_Log << *pEpisode;

			// end episode...								
			p_TextInterface->EndEpisode ();
			bool bValidTrace = p_ExperienceInterface->EndEpisode ();

			// compute rewards...							
			if (true == bValidTrace)
			{
				pEpisode->PrintFeedback (ofs_Log);
				pEpisode->PrintReward (ofs_Log);
				if (true == pEpisode->ComputeReward (pDocument->s_Id, tt_training_mode))
				{
					WordSet_vec_t vecAlternateCommandWords;
					pEpisode->ComputeValidActionWordRanges (vecAlternateCommandWords);
					pEpisode->ComputeWeightUpdates (vecAlternateCommandWords);
					pEpisode->PrintReward (ofs_Log);
					if (false == pEpisode->dq_Steps.empty ())
					{
						int iLastStep = pEpisode->dq_Steps.size () - 1;
						EpisodeStep* pLastStep = pEpisode->dq_Steps [iLastStep];
						if (pLastStep->o_RLState.GetReward () > 0)
							++ iPositiveRewardDocuments;
					}
				}

				pDocument->RememberStates (pEpisode->dq_Steps);
			}

			p_ExperienceInterface->ClearEpisodeData ();
			delete pEpisode;
		}
	}


	p_ExperienceInterface->SaveExperience ();

	ofs_Log << flush;
	return iPositiveRewardDocuments;
}



//													
void Learner::Test (int _iIteration, int _iSubIteration)
{
	Document::SwitchToTestStateGraph ();

	cout << "======================================\n"
			" TEST MODE \n"
			"======================================\n";
	ofs_Log << "======================================\n"
			   " TEST MODE \n"
			   "======================================\n";

	int iTotalStartingStates = Document::dq_StartingStates.size ();
	int iStartingState = 0;

	bool bReplayTrace = (1 == (int)(config)"replay_run");

	ITERATE (String_dq_t, Document::dq_StartingStates, iteStateName)
	{
		++ iStartingState;
		cout << "\n\n   [start-state]" << *iteStateName << '\n';
		ofs_Log << "\n\n   [start-state]" << *iteStateName << '\n';

		String_dq_t* pdqDocumentIds = Document::GetIdsForStartingState (*iteStateName);
		int iTotalDocuments = pdqDocumentIds->size ();
		int iDocument = 0;

		ITERATE (String_dq_t, (*pdqDocumentIds), iteId)
		{
			++ iDocument;

			//												
			Document* pDocument = Document::Find (*iteId);
			cout << '\n';
			cout << "   [time] " << Time::sDateTime () << '\n';
			cout << "   [iteration] " << _iIteration << '-' << _iSubIteration << " " 
				 << iStartingState << '/' << iTotalStartingStates << " "
				 << iDocument << '/' << iTotalDocuments << '\n';
			cout << "   [document] " << *iteId << " " 
				 << pDocument->s_RequiredStartingState << " "
				 << pDocument->e_Type << '\n';

			ofs_Log << '\n';
			ofs_Log << "   [iteration] " << _iIteration << " " 
					<< iStartingState << '/' << iTotalStartingStates << " "
					<< iDocument << '/' << iTotalDocuments << '\n';
			ofs_Log << "   [document] " << *iteId << " " 
					<< pDocument->s_RequiredStartingState << " "
					<< pDocument->e_Type << '\n';


			// check if this is a trace replay...			
			ReplayTrace* pReplayTrace = NULL;
			if (true == bReplayTrace)
				pReplayTrace = ReplayTrace::Find (_iIteration - 1, *iteId);
			if (NULL != pReplayTrace)
			{
				cout << "   [replay]\n";
				ofs_Log << "   [replay]\n";
			}

			// start episode ...							
			p_TextInterface->StartEpisode (*iteId);
			p_ExperienceInterface->StartEpisode (pDocument->s_RequiredStartingState);

			// interpret instructions ...					
			Episode* pEpisode = new Episode (this, *iteId);

			if ((true == bReplayTrace) && (NULL != pReplayTrace))
				// replaying old trace, walk trace.			
				pEpisode->WalkReplayTrace (p_ExperienceInterface,
										   p_TextInterface,
										   pReplayTrace);
			else
				pEpisode->ComputeTrace (p_ExperienceInterface, 
										p_TextInterface,
										tt_test_mode,
										NULL);

			if (i_Verbose > 0)
				cout << *pEpisode;
			ofs_Log << *pEpisode;

			// end episode...								
			p_TextInterface->EndEpisode ();
			bool bValidTrace = p_ExperienceInterface->EndEpisode ();

			// compute rewards...							
			if (true == bValidTrace)
			{
				pEpisode->PrintReward (ofs_Log);
				pEpisode->PrintFeedback (ofs_Log);
				if (true == pEpisode->ComputeReward (pDocument->s_Id, tt_test_mode))
				{
					pEpisode->PrintReward (ofs_Log);
					pEpisode->ComputeCumulativeRewardAndSetToEnvironment ();
				}
			}

			p_ExperienceInterface->ClearEpisodeData ();
			delete pEpisode;
		}
	}

	ofs_Log << flush;
}



//													
bool Learner::LoadWeights (void)
{
	cout << "Loading model parameters ..." << endl;
	bool bWordWeightsLoaded = false;
	bool bObjectWeightsLoaded = false;
	bool bCommandWeightsLoaded = false;
	bool bActionWeightsLoaded = false;

	if (1 != (int)(config)"reset_weights")
	{
		if (true == Path::Exists ((config)"word_weights_file"))
		{
			CsvFile file ((config)"word_weights_file");
			unsigned long ulFeatureCount_WC_WU_WP = vec_Weights_WC_WU_WP.Size ();
			file.ReadLastLine ((double*)vec_Weights_WC_WU_WP, &ulFeatureCount_WC_WU_WP);

			bWordWeightsLoaded = true;
		}


		if (true == Path::Exists ((config)"object_weights_file"))
		{
			CsvFile file ((config)"object_weights_file");
			unsigned long ulFeatureCount_U = vec_Weights_U.Size ();
			file.ReadLastLine ((double*)vec_Weights_U, &ulFeatureCount_U);

			bObjectWeightsLoaded = true;
		}


		if (true == Path::Exists ((config)"command_weights_file"))
		{
			CsvFile file ((config)"command_weights_file");
			unsigned long ulFeatureCount_C = vec_Weights_C.Size ();
			file.ReadLastLine ((double*)vec_Weights_C, &ulFeatureCount_C);

			bCommandWeightsLoaded = true;
		}


		if (true == Path::Exists ((config)"action_weights_file"))
		{
			CsvFile file ((config)"action_weights_file");
			unsigned long ulFeatureCount_A = vec_Weights_A.Size ();
			file.ReadLastLine ((double*)vec_Weights_A, &ulFeatureCount_A);

			bActionWeightsLoaded = true;
		}
	}
	else
	{
		if (true == Path::Exists ((config)"word_weights_file"))
			if (false == Path::RemoveFile ((config)"word_weights_file"))
				return false;
		if (true == Path::Exists ((config)"object_weights_file"))
			if (false == Path::RemoveFile ((config)"object_weights_file"))
				return false;
		if (true == Path::Exists ((config)"command_weights_file"))
			if (false == Path::RemoveFile ((config)"command_weights_file"))
				return false;
		if (true == Path::Exists ((config)"action_weights_file"))
			if (false == Path::RemoveFile ((config)"action_weights_file"))
				return false;
	}


	// all files loaded ...					
	if ((true == bWordWeightsLoaded) &&
		(true == bObjectWeightsLoaded) &&
		(true == bCommandWeightsLoaded) &&
		(true == bActionWeightsLoaded))
	{
		cout << "   - loaded" << endl;
		return true;
	}

	// none of the files were present ...	
	if ((false == bWordWeightsLoaded) &&
		(false == bObjectWeightsLoaded) &&
		(false == bCommandWeightsLoaded) &&
		(false == bActionWeightsLoaded))
	{
		vec_Weights_WC_WU_WP.Memset (0);
		vec_Weights_U.Memset (0);
		vec_Weights_C.Memset (0);
		vec_Weights_A.Memset (0);

		cout << "   - new run, params initialized to 0" << endl;
		return true;
	}

	// Oops some files are missing ??!		
	if (false == bWordWeightsLoaded)
		cout << "   [ERROR]  Word weights file not present or readable! : "
			 << (config)"word_weights_file" << endl;
	if (false == bObjectWeightsLoaded)
		cout << "   [ERROR]  Object weights file not present or readable! : "
			 << (config)"object_weights_file" << endl;
	if (false == bCommandWeightsLoaded)
		cout << "   [ERROR]  Command weights file not present or readable! : "
			 << (config)"command_weights_file" << endl;
	if (false == bActionWeightsLoaded)
		cout << "   [ERROR]  Action weights file not present or readable! : "
			 << (config)"action_weights_file" << endl;

	return false;
}



//													
void Learner::SaveWeights (void)
{
	cout << "\n   [saving weights]   ";
	{
		cout << "wc_wu_wp " << vec_Weights_WC_WU_WP.Sum () << "  ";
		CsvFile file ((config)"word_weights_file", ios_base::app | ios_base::out);
		file.WriteLine ((double*)vec_Weights_WC_WU_WP, vec_Weights_WC_WU_WP.Size ());
		file.Close ();
	}

	{
		cout << "u " << vec_Weights_U.Sum () << "  ";
		CsvFile file ((config)"object_weights_file", ios_base::app | ios_base::out);
		file.WriteLine ((double*)vec_Weights_U, vec_Weights_U.Size ());
		file.Close ();
	}

	{
		cout << "c " << vec_Weights_C.Sum () << "  ";
		CsvFile file ((config)"command_weights_file", ios_base::app | ios_base::out);
		file.WriteLine ((double*)vec_Weights_C, vec_Weights_C.Size ());
		file.Close ();
	}

	{
		cout << "a " << vec_Weights_A.Sum () << "  ";
		CsvFile file ((config)"action_weights_file", ios_base::app | ios_base::out);
		file.WriteLine ((double*)vec_Weights_A, vec_Weights_A.Size ());
		file.Close ();
	}
	cout << endl;
}



