#ifndef __LEARNER__
#define __LEARNER__

#include <set>
#include <deque>
// #include <hash_map>
#include <iostream>
#include <fstream>
#include <nlp_string.h>
#include <nlp_matrix.h>
#include <nlp_vector.h>
#include "State.h"
#include "Action.h"
#include "Annotation.h"
#include "ReplayTrace.h"
#include "FeatureMatrix.h"
#include "BitFeatureMatrix.h"
using namespace std;


// #define FEATURES_C	(13 + 613) * 4 + 1 + 13 * 613 * 4
// #define FEATURES_C	(13 + 613) * 5 + 1
// #define FEATURES_C	(13 + 721) * 5 + 1
#define FEATURES_C	(13 + 739) * 5 + 1

class EpisodeStep;
class Episode;
class Learner;
typedef deque <EpisodeStep*>			EpisodeStep_dq_t;
typedef Vector <double>					double_vec_t;
typedef vector <String>					string_vec_t;
typedef set <UIObject*>					UIObject_set_t;
typedef vector <char_set_t>				WordSet_vec_t;


enum EpisodeStepState_e
{
	ess_unknown,
	ess_ok,
	ess_action_without_words,
	ess_no_words_left,
	ess_no_object_targets_left,
	ess_all_actions_zero_prob,
	ess_apply_action_failed
};

enum EpisodeState_e
{
	es_ok,
	es_action_without_words_ok,
	es_no_words_left,
	es_no_object_targets_left,
	es_try_next_sentence,
	es_error_in_observe_state,
	es_error_in_select_action,
	es_error_in_apply_action,
	es_env_unchanged,
	es_env_unchanged_action_without_words
};

enum Feedback_e
{
	ft_unknown,
	ft_none,
	ft_none_wordless,
	ft_deadend,
	ft_no_words_objects,

	ft_env_unchanged,
	ft_env_unchanged_wordless,
	ft_repetition,

	ft_state_action_error,
	ft_comm_error
};

enum LearningMode_e
{
	lm_unknown,
	lm_annotation_only,
	lm_environment_only,
	lm_mixed_mode
};

enum TrainTest_e
{
	tt_training_mode,
	tt_test_mode
};


// --------------------------------------------------------------
std::ostream& operator<< (std::ostream& _rStream, const Feedback_e _eFeedback);


// --------------------------------------------------------------
class EpisodeStep
{
	friend std::ostream& operator<< (std::ostream& _rStream, const EpisodeStep& _rStep);

	private:
		double_vec_t*	pvec_Weights_WC_WU_WP;
		double_vec_t*	pvec_Weights_U;
		double_vec_t*	pvec_Weights_C;
		double_vec_t*	pvec_Weights_A;
		UIObject_set_t	set_PotentiallyInteractableObjects;
		long			l_CandidateCount_WC_WU_WP;
		long			l_CandidateCount_U;
		long			l_CandidateCount_C;

		// feature computation					
		static int DistanceToIndex (int _iDistance, int _iRange);

		double ComputeScore_WC_WU_WP (int _iLocationWC,
									  int _iLocationWU,
									  int _iLocationWP,
									  Vector<float>& _vecFeatures);
		double ComputeScore_U (int _iLocationWU,
								UIObject* _pObject,
								bool _bFirstAction,
								Vector<float>& _vecFeatures);
		double ComputeScore_C (int _iLocationWC,
								UIObject* _pObject,			
								int _iCommand,
								BitVector<FEATURES_C>& _vecFeatures);
		double ComputeScore_A (int _iLocationWP,
							   UIObject* _pObject,
							   int _iCommand,
							   Vector<float>& _vecFeatures);


		// value computations...				
		bool ComputeProb_WC_WU_WP (bool _bAllowWordlessAction);
		bool ComputeProb_U (bool _bFirstAction, bool _bAllowWordlessAction);
		bool ComputeProb_C (bool _bAllowWordlessAction);
		bool ComputeTraceWalkProb_C (long _lwc, UIObject* _pObject);
		bool ComputeProb_A (bool _bAllowWordlessAction);


		// action sampling						
		double ComputeJointProbs (Vector <double>& _rvecValue,
								  Vector <long>& _rvecIndex,
								  bool _bAvoidNegativeRewardStates);

		// trace maintenance					
		void UpdateTrace_WC_WU_WP (char_set_t& _rsetUseableCommandWords, double _dLogNormalization);
		void UpdateTrace_U (void);
		void UpdateTrace_C (char_set_t& _rsetUseableCommandWords, double _dLogNormalization);
		void UpdateTrace_A (void);



	public:
		static int			i_NeighbourhoodU;
		static int			i_NeighbourhoodWP;

		static double		d_ExplorationEpsilon;
		static double		d_NegativeRewardAvoidanceProb;
		static bool			b_ExploreStochastically;
		static bool			b_UseSoftmax;
		static bool			b_UseSmartSoftmax;
		static bool			b_UseEpsilonGreedy;
		static bool			b_AvoidNegativeRewardStates;
		static bool			b_UseA;
		static bool			b_LookAtNextSentenceAlso;
		static bool			b_PrintSelectedValue;
		static bool			b_Verbose;

		Matrix<double, 3> 	mtx_LogProb_WC_WU_WP;
		Matrix<double, 2> 	mtx_LogProb_U;
		Matrix<double, 3> 	mtx_LogProb_C;
		Matrix<double, 3> 	mtx_LogProb_A;

		FeatureMatrix<float, 3>			mxf_Features_WC_WU_WP;
		FeatureMatrix<float, 2>			mxf_Features_U;
		BitFeatureMatrix<FEATURES_C, 3>	mxf_Features_C;
		FeatureMatrix<float, 3>			mxf_Features_A;

		Vector<double>		vec_Trace_WC_WU_WP;
		Vector<double>		vec_Trace_U;
		Vector<double>		vec_Trace_C;
		Vector<double>		vec_Trace_A;

		RLState			o_RLState;
		RLAction		o_RLAction;
		double			d_SelectedActionLogProb;
		Feedback_e		e_Feedback;
		int				i_FirstCandidateActionSentence;
		int				i_ActionsStartAtSentence;
		bool			b_ActionExecutedWithoutError;
		bool			b_StartStatePrefix;

		ExperienceInterface*	p_ExperienceInterface;
		GraphCacheNode*			p_CurrentState;
	

		EpisodeStep (Learner* _pLearner);
		~EpisodeStep (void);

		EpisodeStepState_e ComputeProbs (bool _bFirstAction,
										 bool _bAllowWordlessAction);

		EpisodeStepState_e WalkAction (AnnotatedAction* _pAnnotation);
		EpisodeStepState_e SelectAction (TrainTest_e _eTrainTestMode,
										 double _dExplorationRate);
		bool Sample_WC_WU_WP_U_C (TrainTest_e _eTrainTestMode,
								  double _dExplorationRate);
		void SetSelectedAction (int _iLocationWC,
								int _iLocationWU,
								int _iLocationWP,
								int _iCommand,
								int _iObject);


		inline EnvResponse_e ObserveState (EnhancedState* _pCurrentState,
											ExperienceInterface* _pExperienceInterface,
											TextInterface* _pTextInterface)
		{
			p_ExperienceInterface = _pExperienceInterface;
			return o_RLState.ObserveState (_pCurrentState, 
										   _pExperienceInterface,
										   _pTextInterface);
		}

		inline EnvResponseToAction_e ApplySelectedAction (ExperienceInterface* _pExperienceInterface,
														  TextInterface* _pTextInterface,
														  TrainTest_e _eTrainTestMode)
		{
			return o_RLState.ApplyAction (_pExperienceInterface,
										  _pTextInterface,
										  o_RLAction,
										  (tt_training_mode == _eTrainTestMode));
		}

		void UpdateTraces (char_set_t& _rsetUseableCommandWords);
};

std::ostream& operator<< (std::ostream& _rStream, const EpisodeStep& _rStep);




// --------------------------------------------------------------
class Episode											
{
	friend class EpisodeStep;
	friend std::ostream& operator<< (std::ostream& _rStream, const Episode& _rEpisode);

	private:
		Learner*					p_Learner;
		String						s_DocumentId;
		EpisodeStep*				p_LastStep;
		EnhancedState*				p_CurrentEnhancedState;
		int							i_LastActionSentence;
		bool						b_WeightUpdatesComputed;
		bool						b_OnAnnotatedPath;

		EpisodeState_e ObserveStateAndAct (ExperienceInterface* _pExperienceInterface,
										   TextInterface* _pTextInterface,
										   bool _bAllowWordlessAction,
										   bool _bFromNextSentence,
										   TrainTest_e _eTrainTestMode,
										   AnnotatedAction* _pAnnotatedAction);

		void AccumulateTrace_WC_WU_WP (int _iTimesteps);
		void AccumulateTrace_U (int _iTimesteps);
		void AccumulateTrace_C (int _iTimesteps);
		void AccumulateTrace_A (int _iTimesteps);
		void ComputeWeightUpdate_WC_WU_WP (double_vec_t& _vecReward, double_vec_t& _rvecWeightUpdates);
		void ComputeWeightUpdate_U (double_vec_t& _vecReward, double_vec_t& _rvecWeightUpdates);
		void ComputeWeightUpdate_C (double_vec_t& _vecReward, double_vec_t& _rvecWeightUpdates);
		void ComputeWeightUpdate_A (double_vec_t& _vecReward, double_vec_t& _rvecWeightUpdates);

		void UpdateEnhancedStateGraph (long _lAction);

		bool ComputeRewardUsingEnvironment (void);
		bool ComputeRewardUsingEnvironment_v1 (void);
		bool ComputeRewardUsingEnvironment_v1_5 (void);
		bool ComputeRewardUsingEnvironment_v2 (void);
		bool ComputeRewardUsingEnvironment_v3 (void);
		bool ComputeRewardUsingEnvironment_v4 (void);
		bool ComputeRewardUsingAnnotation (Document* _pDocument);

		void ComputeAlternateWordSelections (void);

	public:
		EpisodeStep_dq_t			dq_Steps;


		Episode (Learner* _pLearner, String& _rDocumentId);
		~Episode (void);

		bool ComputeTrace (ExperienceInterface* _pExperienceInterface,
						   TextInterface* _pTextInterface,
						   TrainTest_e _eTrainTestMode,
						   AnnotatedAction_dq_t* _pdqActionsToStartState);
		bool WalkReplayTrace (ExperienceInterface* _pExperienceInterface,
							  TextInterface* _pTextInterface,
							  ReplayTrace* _pReplayTrace);
		bool WalkAnnotationTrace (ExperienceInterface* _pExperienceInterface,
								  TextInterface* _pTextInterface);
		void PrintFeedback (ofstream& _ofsLog);
		void PrintReward (ofstream& _ofsLog);

		bool ComputeReward (String& _sId, TrainTest_e _eTrainTestMode);
		void ComputeWeightUpdates (WordSet_vec_t& _rvecWordSet);
		void ComputeValidActionWordRanges (WordSet_vec_t& _rvecWordSet);

		void ComputeCumulativeRewardAndSetToEnvironment (void);
};

std::ostream& operator<< (std::ostream& _rStream, const Episode& _rEpisode);




// --------------------------------------------------------------
class Learner
{
	friend class EpisodeStep;
	friend class Episode;

	private:
		static size_t	i_Features_WC_WU_WP;
		static size_t	i_Features_U;
		static size_t	i_Features_C;
		static size_t	i_Features_A;
		static size_t	i_FeaturesTotal;

		double_vec_t	vec_Weights_WC_WU_WP;
		double_vec_t	vec_Weights_U;
		double_vec_t	vec_Weights_C;
		double_vec_t	vec_Weights_A;
		double_vec_t	vec_WeightUpdate_WC_WU_WP;
		double_vec_t	vec_WeightUpdate_U;
		double_vec_t	vec_WeightUpdate_C;
		double_vec_t	vec_WeightUpdate_A;

		LearningMode_e	e_LearningMode;
		int				i_Verbose;

		double			d_RegularizationFactor;
		bool			b_RandomlyStrayFromAnnotations;
		double			d_StrayProbability;
		int				i_DontStrayTillIteration;

		long			l_ProcessedDocuments;
		long			l_DocumentsPerUpdate;

		ofstream		ofs_Features;

		ExperienceInterface*	p_ExperienceInterface;
		TextInterface*			p_TextInterface;



	public:
		static double	d_PositiveRewardMultiplier;
		static double	d_NegativeRewardMultiplier;

		static ofstream		ofs_Log;
		static int			i_MaxWordlessActions;
		double				d_AlternateStartStateProbability;
		double				d_LearningRate;
		double				d_ExplorationRate;
		bool				b_WalkAnnotations;

		Learner (void);
		~Learner (void);

		bool Init (ExperienceInterface* _pExperienceInterface,
				   TextInterface* _pTextInterface);

		void ForgetWeightUpdates (void);
		void UpdateWeights (void);
		bool LoadWeights (void);
		void SaveWeights (void);

		int Learn (int _iIteration, int _iSubIteration);
		void Test (int _iIteration, int _iSubIteration);

		static inline size_t FeatureCount_WC_WU_WP (void) { return i_Features_WC_WU_WP; };
		static inline size_t FeatureCount_U (void) { return i_Features_U; };
		static inline size_t FeatureCount_C (void) { return i_Features_C; };
		static inline size_t FeatureCount_A (void) { return i_Features_A; };
		static inline size_t FeatureCountTotal (void) { return i_FeaturesTotal; };
};



#endif
