#include "ReplayTrace.h"
#include <nlp_macros.h>
#include <nlp_config.h>
#include <nlp_filesystem.h>
#include <assert.h>


ReplayIterations_dq_t	ReplayTrace::dq_ReplayIterations;


//														
ReplayTraceSentence::~ReplayTraceSentence (void)
{
	ITERATE (AnnotatedAction_dq_t, dq_Action, ite)
		delete *ite;
	dq_Action.clear ();
}


//														
ReplayTrace::ReplayTrace (String& _rID)
{
	s_ID = _rID;
}


//														
ReplayTrace::~ReplayTrace (void)
{
	ITERATE (ReplayTraceSentence_dq_t, dq_Sentence, ite)
		delete *ite;
	dq_Sentence.clear ();
}


//														
ReplayTrace* ReplayTrace::Find (int _iIteration, String& _rID)
{
	if (_iIteration >= (int)dq_ReplayIterations.size ())
		return NULL;

	IdToReplayTrace_hmp_t* phmpIdToReplayTrace;
	phmpIdToReplayTrace = dq_ReplayIterations [_iIteration];

	IdToReplayTrace_hmp_t::iterator	ite;
	ite = phmpIdToReplayTrace->find (_rID);
	if (phmpIdToReplayTrace->end () == ite)
		return NULL;

	return ite->second;
}


//														
bool ReplayTrace::Load (void)
{
	if (0 == (int)(config)"replay_run")
		return true;

	cout << "Loading replay trace ..." << endl;


	// load the replay traces ...				
	{
		File file;
		if (false == file.Open ((config)"replay_trace_file"))
			return false;

		IdToReplayTrace_hmp_t*	phmpIdToReplayTrace = NULL;
		ReplayTrace* pReplayTrace = NULL;
		ReplayTraceSentence* pSentence = NULL;
		AnnotatedAction* pAction = NULL;


		String sLine;
		while (true == file.ReadLine (sLine))
		{
			sLine.Strip ();
			if ("" == sLine)
				continue;

			if (true == sLine.Has ("TRAINING MODE"))
			{
				phmpIdToReplayTrace = new IdToReplayTrace_hmp_t;
				dq_ReplayIterations.push_back (phmpIdToReplayTrace);
				continue;
			}

			if (true == sLine.Has ("TEST MODE"))
			{
				phmpIdToReplayTrace = new IdToReplayTrace_hmp_t;
				dq_ReplayIterations.push_back (phmpIdToReplayTrace);
				continue;
			}


			// document id ...						
			if (true == sLine.Has ("[document]"))
			{
				zchar_dq_t dqValues = sLine.DestructiveSplit (' ', 3);
				String sID (dqValues [1]);

				pReplayTrace = new ReplayTrace (sID);
				phmpIdToReplayTrace->insert (make_pair (sID, pReplayTrace));	
				pSentence = NULL;
				pAction = NULL;
				continue;
			}


			// command words ...					
			if (true == sLine.Has ("[textaction]"))
			{
				zchar_dq_t dqWords = sLine.DestructiveSplit ('\x01', 3);

				pAction = new AnnotatedAction;
				zchar_dq_t dqValues;

				dqValues = String::DestructiveSplit (dqWords [1], '|', 2);
				pAction->i_CommandWord = atoi (dqValues [1]);

				dqValues = String::DestructiveSplit (dqWords [2], '|', 2);
				pAction->i_ObjectWord = atoi (dqValues [1]);

				dqValues = String::DestructiveSplit (dqWords [3], '|', 2);
				pAction->i_Parameter = atoi (dqValues [1]);
				continue;
			}

			// environment command ...				
			if (true == sLine.Has ("[worldaction]"))
			{
				// we sould already have encountered  			
				// a line with the word annotations, so pAction	
				// would already have been created.            	
				assert (NULL != pAction);

				zchar_dq_t dqTokens = sLine.DestructiveSplit ('\x01', 4);
				zchar_dq_t dqValues;

				dqValues = String::DestructiveSplit (dqTokens [1], '|', 2);
				pAction->s_Command = dqValues [2];
				pAction->s_Command.Strip ();

				dqValues = String::DestructiveSplit (dqTokens [2], '|', 3);
				pAction->s_ObjectName = dqValues [2];
				pAction->s_ObjectType = dqValues [3];
				pAction->s_ObjectType.Strip ();
				pAction->s_ObjectType.LowerCase ();

				pAction->s_TemporallyUniqueObjectId = dqTokens [3];
				pAction->s_TemporallyUniqueObjectId.Strip ();

				pSentence->dq_Action.push_back (pAction);
				pAction = NULL;
				continue;
			}


			// sentence text ...					
			if (true == sLine.Has ("[sentence]"))
			{
				sLine.Strip ();

				pSentence = new ReplayTraceSentence;
				pReplayTrace->dq_Sentence.push_back (pSentence);
				continue;
			}
		}
	}

	cout << "   - replay iterations : " << dq_ReplayIterations.size () << endl;
	return true;
}


