#include "Sample.h"
#include <algorithm>
#include <time.h>
using namespace std;

gsl_rng*	Sample::p_GslRng;


//													
void Sample::Init (void)
{
	gsl_rng_env_setup ();
	gsl_rng_default_seed = (unsigned long) time (NULL);
	p_GslRng = gsl_rng_alloc (gsl_rng_default);
}



//													
void Sample::Destroy (void)
{
	assert ((gsl_rng*)0x1 != p_GslRng);
	gsl_rng_free (p_GslRng);
	p_GslRng = (gsl_rng*)0x1;
}



//													
size_t Sample::SampleFromPDF (Vector <double>& _vecPDF)
{
	for (size_t i = 1; i < _vecPDF.Size (); ++ i)
		_vecPDF [i] += _vecPDF [i - 1];
	assert (0 != _vecPDF [_vecPDF.Size () - 1]);

	size_t lSample = SampleFromCDF (_vecPDF);
	if (0 == _vecPDF [lSample])
	{
		cout << "[ERROR]  sum(pdf) = " << _vecPDF.Sum () << endl;
		cout << "             pdf  = " << _vecPDF << endl;
	}
	assert (0 != _vecPDF [lSample]);
	return lSample;
}



//													
size_t Sample::SampleFromCDF (Vector <double>& _vecCDF)
{
	double dMaxCDF = _vecCDF [_vecCDF.Size () - 1];
	double dUniform = dMaxCDF * gsl_rng_uniform (p_GslRng);
	if (dUniform == dMaxCDF)
		return _vecCDF.Size () - 1;

	double* pBegin = (double*) _vecCDF;
	double* pEnd = pBegin + _vecCDF.Size ();
	double* pSample =  std::upper_bound (pBegin, pEnd, dUniform);
	return (pSample - pBegin);
}



//													
size_t Sample::SelectMax (Vector <double>& _vecPDF)
{
	size_t lIndex = 0;
	double dMax = _vecPDF [0];
	bool bManyMaxs = false;
	for (size_t i = 1; i < _vecPDF.Size (); ++ i)
	{
		if (dMax < _vecPDF [i])
		{
			dMax = _vecPDF [i];
			lIndex = i;
			bManyMaxs = false;
		}
		else if (dMax == _vecPDF [i])
			bManyMaxs = true;
	}

	if (false == bManyMaxs)
		return lIndex;

	// if there are many instances of the max,	
	// we need ot sample randomly from them...	
	Vector<double> vecMax;
	vecMax.Create (_vecPDF.Size ());
	for (size_t i = 0; i < _vecPDF.Size (); ++ i)
		vecMax [i] = (dMax == _vecPDF [i])? 1 : 0;
	
	return SampleFromPDF (vecMax);
}



//													
bool Sample::SampleCoinflip (double _dProbTrue)
{
	assert (_dProbTrue <= 1);
	double dUniform = gsl_rng_uniform (p_GslRng);
	return (dUniform <= _dProbTrue);
}



//													
int Sample::SampleUniformCategorical (int _iCategories)
{
	double dUniform = gsl_rng_uniform (p_GslRng) * _iCategories;
	return (int) floor (dUniform);
}



