#include <assert.h>
#include <nlp_config.h>
#include "SimpleEnvInterface.h"

#define RECEIVE_BUFFER	81920


//												
SimpleEnvironmentInterface::SimpleEnvironmentInterface (void)
{
	ClientSocket::SetStandalone (true);
}


//												
SimpleEnvironmentInterface::~SimpleEnvironmentInterface (void)
{
	if (true == ClientSocket::IsConnected ())
	{
		cout << "[WARNING]  destructor called on SimpleEnvironmentInterface "
				" while it is still connected to the environment.  Call "
				" SimpleEnvironmentInterface::Stop () before deleting object."
			 << endl;
		Stop ();
	}

	ClientSocket::Close ();
}


//												
bool SimpleEnvironmentInterface::Start (void)
{
	cout << "Connecting to external world ..." << endl;
	if (false == ClientSocket::ConnectBlocking ((config)"cache_server", 
												(config)"cache_service"))
	{
		cout << "   [ERROR]  Failed to connect to external world." << endl;
		return false;
	}
	Socket::SetNoDelay (true);
	ClearConnection (false);
	cout << "   - connected" << endl;
	return true;
}


//												
bool SimpleEnvironmentInterface::Stop (void)
{
	return ClientSocket::Close ();
}


//												
void SimpleEnvironmentInterface::ClearConnection (bool _bWarn)
{
	char zData [1025];
	long lBytes = ReceiveBlocking (zData, 1024, 100);
	long lTotalBytes = lBytes;

	while (lBytes > 0)
	{
		long lBytes = ReceiveBlocking (zData, 1024, 100);
		if (lBytes > 0)
			lTotalBytes += lBytes;
	}

	if ((true == _bWarn) && (lTotalBytes > 0))
		cout << "[INFO] Discarding " << lTotalBytes << " of unexpected data." << endl;
}


//												
bool SimpleEnvironmentInterface::SendReceive (String& _rMessage, String* _pResponse)
{
	if (false == ClientSocket::IsConnected ())
		return false;

	ClientSocket::SendBlocking ((const char*)_rMessage, _rMessage.length ());
	if (false == ClientSocket::IsConnected ())
		return false;

	String sResponse;
	char zData [RECEIVE_BUFFER + 1];
	zData [0] = '\0';
	while (NULL == strchr (zData, '\x10'))
	{
		long lBytes = ClientSocket::ReceiveBlocking (zData, RECEIVE_BUFFER, 1000);
		if (false == ClientSocket::IsConnected ())
			return false;

		if (lBytes > 0)
		{
			zData [lBytes] = '\0';
			sResponse << zData;
		}
	}

	sResponse.Strip ();
	*_pResponse = sResponse;

	if ((true == sResponse.StartsWith ("[ERROR]")) ||
		(true == sResponse.StartsWith ("[[ERROR]]")))
		return false;

	return true;
}


//												
bool SimpleEnvironmentInterface::StartEpisode (String& _rStartingState)
{
	ClearConnection (true);

	String sMessage;
	sMessage << "> " << _rStartingState << '\n';

	String sResponse;
	if (false == SendReceive (sMessage, &sResponse))
		return false;
	if (true == sResponse.Has ("restart complete"))
		return true;
	return false;
}


//												
void SimpleEnvironmentInterface::EndEpisode (void)
{
	ClientSocket::SendBlocking (".\n", 2);
	ClearConnection (false);
}


//												
String SimpleEnvironmentInterface::GetEnvironmentId (void)
{
	return GetServerId ();
}



//												
void SimpleEnvironmentInterface::OnReceive (const void* _zData, long _lBytes)
{
	if (0 == _lBytes)
		return;

	assert (false);
}


//												
void SimpleEnvironmentInterface::OnDisconnect (void)
{
	cout << "[ERROR] Socket connection to cache server lost!" << endl;
	cout << "        Attempting reconnection..." << endl;

	if (false == ClientSocket::ConnectBlocking ((config)"cache_server", 
												(config)"cache_service"))
	{
		cout << "[ERROR] Reconnect failed. Quitting." << endl;
		assert (false);
	}

	ClientSocket::SetNoDelay (true);
	cout << "[INFO] Reconnected to cache server." << endl;
}


