#include <nlp_config.h>
#include <nlp_macros.h>
#include "UIObject.h"
#include "FieldId.h"


StringToIndex_hmp_t	ObjectClass::hmp_ClassToIndex;
String_vec_t		ObjectClass::vec_IndexToClass;
int					ObjectClass::i_Count = -1;



//												
void ObjectClass::Load (void)
{
	String sClasses = (config)"object_classes";

	String_dq_t dqValues;
	sClasses.Split (dqValues, '|');

	i_Count = dqValues.size ();
	vec_IndexToClass.resize (i_Count);
	for (size_t i = 0; i < dqValues.size (); ++ i)
	{
		dqValues [i].Strip ();
		hmp_ClassToIndex.insert (make_pair (dqValues [i], i));
		vec_IndexToClass [i] = dqValues [i];
	}
};



//												
std::ostream& operator<< (std::ostream& _rStream, const UIObject& _rObject)
{
	UIObject& rObject = (UIObject&) _rObject;

	_rStream << rObject.i_Index << ' ' 
			 << rObject.z_ID << ' '
			 << rObject.s_Name << ' '
			 << rObject.s_CleanName << ' '
			 << ObjectClass::Get ((int)rObject.i_ObjectClass) << ' '
			 << rObject.s_TemporallyUniqueId << " ## ";

	for (size_t i = 0; i < FEATURE_COUNT; ++ i)
		 _rStream << ' ' << rObject.bit_Features [i];
	
	return _rStream;
}



//												
UIObject::UIObject (void)
	: EnvObject ()
{
	i_Index = -1;
	i_ObjectClass = 0;
}


//												
UIObject::UIObject (char* _zData, int _iObjectIndex)
{
	SetData (_zData, _iObjectIndex);
}



//												
UIObject::~UIObject (void)
{
}



//												
void UIObject::SetData (char* _zData, int _iObjectIndex)
{
	i_Index = _iObjectIndex;

	String sCopy (_zData);
    zchar_dq_t dqValues = String::DestructiveSplit (_zData, '\x01', 3);
    String sID (dqValues [raw_OBJECT_ID]);
	sID.Strip ();
	assert (sID.length () < 24);
	strcpy (z_ID, sID);


    // split out feature values ...
    zchar_dq_t dqFeatures = String::DestructiveSplit (dqValues [3], '\x01');
	if (raw_MESSAGE_LENGTH  - 3 != dqFeatures.size ())
	{
		cerr << "[ERROR]  Feature count = " << dqFeatures.size () 
			 << ", should be " << raw_MESSAGE_LENGTH << '\n'
			 << "->" << sCopy << "<-" << endl;
	}
	assert (raw_MESSAGE_LENGTH - 3 == dqFeatures.size ());

	s_TemporallyUniqueId = dqFeatures [raw_TEMPORALLY_UNIQUE_ID - 3];
	s_TemporallyUniqueId = s_TemporallyUniqueId.substr (s_TemporallyUniqueId.Find ("\x04"));
	s_TemporallyUniqueId.Strip ();
	s_Name = dqFeatures [raw_OBJECT_NAME - 3];
	s_Name.Strip ();
	String sObjectClass = dqFeatures [raw_OBJECT_CLASS - 3];
	sObjectClass.Strip ();
	sObjectClass.LowerCase ();
	i_ObjectClass = (char) ObjectClass::Find (sObjectClass);
	s_CleanName = s_Name;
	s_CleanName.Replace ("[ \\t]+", " ");
	s_CleanName.Replace ("[^a-zA-Z0-9]+", " ");


	bit_Features.Clear ();
	bit_Features.SetFeature (FAMILY_ID,	(atoi (dqFeatures [raw_FAMILY_TOUCHED - 3]) > 0)? 1 : 0);
	bit_Features.SetFeature (OBJECT_ID, atoi (dqFeatures [raw_OBJECT_TOUCHED - 3]));
	bit_Features.SetFeature (MAIN_WND_ID, atoi (dqFeatures [raw_MAIN_WND_TOUCHED - 3]));
	bit_Features.SetFeature (ACT_WND_ID, atoi (dqFeatures [raw_ACT_WND_TOUCHED - 3]));

	bit_Features.SetFeature (NAME_CHANGED, atoi (dqFeatures [raw_NAME_CHANGED - 3]));
	bit_Features.SetFeature (VISIBLE, atoi (dqFeatures [raw_VISIBLE - 3]));
	bit_Features.SetFeature (SHOWING, atoi (dqFeatures [raw_SHOWING - 3]));
	bit_Features.SetFeature (INFOCUS, atoi (dqFeatures [raw_INFOCUS - 3]));
	bit_Features.SetFeature (FG, atoi (dqFeatures [raw_FG - 3]));
	bit_Features.SetFeature (INPUT_ENABLED, atoi (dqFeatures [raw_INPUT_ENABLED - 3]));
	bit_Features.SetFeature (SELECTED, atoi (dqFeatures [raw_SELECTED - 3]));
	bit_Features.SetFeature (CHILD, atoi (dqFeatures [raw_CHILD - 3]));
	bit_Features.SetFeature (NEW, atoi (dqFeatures [raw_NEW - 3]));

	bit_Features.SetFeature (NEWLY_VISIBLE, atoi (dqFeatures [raw_NEWLY_VISIBLE - 3]));
	bit_Features.SetFeature (NEWLY_SHOWING, atoi (dqFeatures [raw_NEWLY_SHOWING - 3]));
	bit_Features.SetFeature (NEWLY_INFOCUS, atoi (dqFeatures [raw_NEWLY_INFOCUS - 3]));
	bit_Features.SetFeature (NEWLY_FG, atoi (dqFeatures [raw_NEWLY_FG - 3]));
	bit_Features.SetFeature (NEWLY_INPUT_ENABLED, atoi (dqFeatures [raw_NEWLY_INPUT_ENABLED - 3]));
	bit_Features.SetFeature (NEWLY_SELECTED, atoi (dqFeatures [raw_NEWLY_SELECTED - 3]));
}



//												
void UIObject::DestroyObjects (void* _pObjects)
{
	UIObject* pObjects = (UIObject*) _pObjects;
	delete[] pObjects;
}



//												
void UIObject::UnpackState (String& _rState, void** _ppObjects, int* _pObjectCount)
{
	String sCopy (_rState);
	zchar_dq_t dqRawObjects = sCopy.DestructiveSplit ('\x03');

	int iCount = 0;
	ITERATE (zchar_dq_t, dqRawObjects, ite)
	{
		if ('\0' == **ite)
			continue;
		if (NULL == strchr (*ite, '\x01'))
			continue;
		++ iCount;
	}

	*_pObjectCount = iCount;
	if (0 == iCount)
	{
		*_ppObjects = NULL;
		return;
	}

	UIObject* pObjects = new UIObject [iCount];
	*_ppObjects = pObjects;

	int iObjectIndex = 0;
	ITERATE (zchar_dq_t, dqRawObjects, ite)
	{
		if ('\0' == **ite)
			continue;
		if (NULL == strchr (*ite, '\x01'))
			continue;

		pObjects [iObjectIndex].SetData (*ite, iObjectIndex);
		++ iObjectIndex;
	}
}



